#include <stdio.h>
#include <math.h>
#include "cm_cm01.h"
#include "pcsgbd.h" /* Luciana */

/***********************************************************************************
CM - Software para programao e manipulao de matrizes

Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os 
termos da Licena Pblica Geral GNU conforme publicada pela Free Software 
Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso 
posterior.

Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA 
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes. 

Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este 
programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59 
Temple Street, Suite 330, Boston, MA 02111-1307 USA. 

***********************************************************************************

Parte integrante do SOC - SOftware Cientfico 

Descrio: 	
 
Calcula os autovalores e autovetores de uma matriz  quadrada. Esta funo s pode ser usada com o comando de atribuio (=), com a especificao de dois nomes de variveis separados por vrgula (,). Os resultados que esta funo retorna dependem do tipo de matriz quadrada que  especificada como argumento:
1. Quando a matriz especificada como argumento  simtrica, esta funo retorna para a primeira varivel especificada no comando de atribuio o vetor dos autovalores em ordem crescente e para a segunda varivel a matriz dos autovetores ortonormais correspondentes.
Se B e C representam, respectivamente, essas variveis:
 C'* arg * C = DI  e  C'* C = C * C' = I,
onde 
DI  uma matriz diagonal formada pelos elementos de B (autovalores) 
I  a matriz identidade.
Exemplo:

B,C = AUTO(A1'A1);

B    	     Col1
Lin1	      0.140
Lin2	      4.830
Lin3 	1034.030

C    	   Col1        Col2        Col3
Lin1	0.796      -0.602        0.055
Lin2 	-0.598     -0.770        0.224
Lin3	0.093       0.212        0.973
Para cada i-simo autovalor, dado pela i-sima linha da vetor B, o autovetor ortonormal  dado pela correspondente j-sima coluna da matriz C. 
2. Quando a matriz especificada como argumento no  simtrica, esta funo retorna para a primeira varivel especificada no comando 
de atribuio  uma matriz com duas colunas, onde a primeira coluna identifica a parte real e a segunda coluna identifica a parte imaginria 
dos autovalores. Os pares conjugados dos autovalores aparecem consecutivamente, com o autovalor com a parte imaginria positiva precedendo 
o autovalor com a parte imaginria negativa. Se todos os elementos da segunda coluna so iguais a zero, a matriz especificada como argumento 
possui somente autovalores reais. Para a segunda varivel especificada no comando de atribuio, esta funo retorna uma matriz que contm a parte
 real e imaginria dos autovetores. Se o i-simo auto-valor  real, a correspondente j-sima coluna dessa matriz  um autovetor para  esse autovalor. 
Se o i-simo autovalor  complexo com parte imaginria positiva, as correspondentes j-sima e (j+1)-sima colunas dessa matriz contm a parte real e a parte 
imaginria de um autovetor para esse autovalor. O conjugado desse vetor  um autovetor para o conjugado desse autovalor com parte imaginria positiva.
Exemplo 1:

A = { 2 1 5,6 8 9,1 3 1 };

B,C = AUTO(A);

B         Col1     Col2
Lin1   12.095    0.000
Lin2   -0.547     0.970
Lin3   -0.547    -0.970

C          Col1     Col2     Col3
Lin1    0.235    0.075    0.995 
Lin2    0.964    0.269   -0.189
Lin3    0.282   -0.285   -0.454
Neste exemplo, foram obtidos para a matriz A os autovalores
12.095,	-0.547+0.970i	-0.547-0.970i
e os autovetores correspondentes 
(0.235,0.964,0.282)'

(0.075+0.995i,0.269-0.189i,-0.285-0.454i)'

(0.075-0.995i,0.269+0.189i,-0.285+0.454i)'. 

Exemplo 2:

D = { 1 3 5 2,2 2 1 5,5 4 3 2,1 7 0 8 };

E,F = AUTO(D);

E        Col1    Col2
Lin1   13.090   0.000
Lin2   -2.917   0.000
Lin3    6.086   0.000
Lin4   -2.259   0.000


F         Col1    Col2     Col3     Col4
Lin1    0.487  -0.811   0.577  -0.629
Lin2    0.483   0.472   0.016   1.087
Lin3    0.583   0.444   0.722   0.030
Lin4    0.760  -0.228  -0.362  -0.680

Neste caso a matriz D especificada como argumento tem somente autovalores reais. Para cada i-simo autovalor, 
dado pela i-sima linha da matriz E, um autovetor  dado pela correspondente j-sima coluna da matriz F.
	OBSERVAO: O CM no faz operaes com nmeros complexos, de modo que operaes envolvendo autovalores ou 
	autovetores com a parte imaginria no podem ser especificadas. 

Desenvolvido por: 

	Moacir Pedroso Junior 
	Silvio Evangelista
		
	Informaes do CVS:
	$Source:
	$Revision:
	$Date:

***********************************************************************************/
extern double _toler_;

#define	TOL	_toler_
#define MAEPS	_toler_

void meigen(a,d,z)
Simbolo_t *a;
Simbolo_t *d;
Simbolo_t *z;
{
	int i,j,k,l,n,m;
	double b,c,f,g,h,p,r,s;

	Simbolo_t *e = faztemp(NLINS(a),1);
#ifdef EDIX
	double *dp;
#endif

	if ((n=NLINS(a)) != NCOLS(a)) {

/* *** Padroniz. msg erro - Luciana 23/12/94
		execerro("matriz nao e quadrada",(char *) 0);
*** */
		execerro(PegaMsg(3115), (char *) 0);
	}

	for(i=0; i < n; ++i) {
		for(j=i; j < n; ++j) {
			if (fuzz(DADOij(a,i,j),DADOij(a,j,i)) != 0)
				execerro("matriz nao e simetrica",(char *) 0);
		}
	}

	mtri(a,e,d,z);

	for(i=1; i < n; ++i) {
#ifdef EDIX
		dp = &(DADOij(e,i-1,0));
		*dp = DADOij(e,i,0);
#else
		DADOij(e,i-1,0) = DADOij(e,i,0);
#endif
	}
#ifdef EDIX
 	dp = &(DADOij(e,n-1,0));
	*dp  = b = f =0.0;
#else
 	DADOij(e,n-1,0)  = b = f =0.0;
#endif

	for(l=0; l < n; ++l) {
		j = 0;
		h = MAEPS*(fabs(DADOij(d,l,0)) + fabs(DADOij(e,l,0)));
		if (b < h)
			b = h;
		for(m=l; m < n; ++m) {
			if (fabs(DADOij(e,m,0)) <= b) 
				goto cont; 
		}
	cont:	if ( m == l)
			goto root;
	nextit: if (j == 100)

/* *** Padroniz. msg erro - Luciana 23/12/94
			execerro("nao convergiu",(char *) 0);
*** */
			execerro(PegaMsg(1105),(char *) 0);
		++j;
		
		g = DADOij(d,l,0);
		p = (DADOij(d,l+1,0) - g) / ( 2.0*DADOij(e,l,0) );
		r = sqrt(p*p + 1.0);
#ifdef EDIX
		dp = &(DADOij(d,l,0));
		*dp = DADOij(e,l,0) / ((p < 0.0)? (p-r):(p+r));
#else
		DADOij(d,l,0) = DADOij(e,l,0) / ((p < 0.0)? (p-r):(p+r));
#endif
		h = g - DADOij(d,l,0) ;
		for(i=l+1; i<n; ++i) {
#ifdef EDIX
			dp = &(DADOij(d,i,0));
			*dp =  DADOij(d,i,0) - h;
#else
			DADOij(d,i,0) =  DADOij(d,i,0) - h;
#endif
		}
		f += h;
		p = DADOij(d,m,0);
		c = 1.0;
		s = 0.0;
 		for(i=m-1; i >= l; --i) {
			g = c*DADOij(e,i,0) ;
			h = c*p;
			if (fabs(p) >= fabs(DADOij(e,i,0))) {
				c = DADOij(e,i,0) /p;
				r = sqrt(c*c+1.0);
#ifdef EDIX
				dp = &(DADOij(e,i+1,0));
				*dp = s*p*r;
#else
				DADOij(e,i+1,0) = s*p*r;
#endif
				s = c/r;
				c = 1/r;
			}
			else {
				c = p / DADOij(e,i,0);
				r = sqrt(c*c+1.0);
#ifdef EDIX
				dp = &(DADOij(e,i+1,0));
				*dp = s*DADOij(e,i,0)*r;
#else
				DADOij(e,i+1,0) = s*DADOij(e,i,0)*r;
#endif
				s = 1 / r;
				c = c / r;
			}
			p = c * DADOij(d,i,0) - s* g;
#ifdef EDIX
			dp = &(DADOij(d,i+1,0));
			*dp = h + s * (c*g+s*DADOij(d,i,0));
#else
			DADOij(d,i+1,0) = h + s * (c*g+s*DADOij(d,i,0));
#endif
			for(k=0; k < n; ++k) {
				h = DADOij(z,k,i+1);
#ifdef EDIX
				dp = &(DADOij(z,k,i+1));
				*dp = s * DADOij(z,k,i)+c*h;
				dp = &(DADOij(z,k,i));
				*dp = c * DADOij(z,k,i) -s*h;
#else
				DADOij(z,k,i+1) = s * DADOij(z,k,i)+c*h;
				DADOij(z,k,i) = c * DADOij(z,k,i) -s*h;
#endif
			}
		}
#ifdef EDIX
		dp = &(DADOij(e,l,0));
		*dp = s*p;
		dp = &(DADOij(d,l,0));
		*dp = c*p;
#else
		DADOij(e,l,0) = s*p;
		DADOij(d,l,0) = c*p;
#endif
		if (fabs(DADOij(e,l,0)) > b)
			goto nextit;
root:
#ifdef EDIX
		dp = &(DADOij(d,l,0));
		*dp += f;
#else
		DADOij(d,l,0) += f;
#endif
	}
	for(i=0; i < n; ++i) {
		k = i;
		p = DADOij(d,i,0) ;
		for(j=i+1; j < n; ++j) {
			if (DADOij(d,j,0) < p) {
				k = j;
				p = DADOij(d,j,0) ;
			}
		}
		if (k != i) {
#ifdef EDIX
			dp = &(DADOij(d,k,0));
			*dp = DADOij(d,i,0) ;
			dp = &(DADOij(d,i,0));
			*dp = p;
#else
			DADOij(d,k,0) = DADOij(d,i,0) ;
			DADOij(d,i,0) = p;
#endif
			for(j=0; j < n; ++j) {
				p = DADOij(z,j,i);
#ifdef EDIX
				dp = &(DADOij(z,j,i));
				*dp = DADOij(z,j,k);
				dp = &(DADOij(z,j,k));
				*dp = p;
#else
				DADOij(z,j,i) = DADOij(z,j,k);
				DADOij(z,j,k) = p;
#endif
			}
		}
	}

	tentaLibSim(e);
}

