/***********************************************************************************
 ESTAT - Clculo de Estatsticas Descritivas
 
 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
 termos da Licena Pblica Geral GNU conforme publicada pela Free Software
 Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
 posterior.

 Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
 programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
 Temple Street, Suite 330, Boston, MA 02111-1307 USA.

 **********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

    Mdulo integrante do software ESTAT, que  calcula estatsticas descritivas 
    de variveis  numricas. As  estatsticas  que  podem  ser requeridas so: nmero 
    de observaes, valor mnimo, valor mximo, total, mdia, varincia, desvio padro, 
    erro padro, coeficiente de variao, assimetria, curtose, mediana, estatstica D 
    de Kolmogorov, percentis e modas.
    Neste mdulo encontram-se as rotinas de manipulao do tipo abstrato de dados conjunto,
    utilizado para guardar as estatsticas. 

 Desenvolvido por:

	  Dalton Francisco Andrade
	  Maria Fernanda Moura
	  Maurcio Noronha Festa

 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/

#include "estat.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>

#ifdef MEM_DEBUG
#include "mshell.h"
#endif

/* FUNCAO INS :
	Insere elementos em um conjunto, onde :
		. o conjunto e' uma lista duplamente ligada
		. o elemento e' uma cadeia de caracteres
	A insercao sempre e' feita no final da lista, isto se o elemento nao
	pertencia ao conjunto(lista).
*/
extern char **aux_ins;
extern int aux_insm;
extern int controle;

void ins(nome, elemento)
Lista nome;
char *elemento;
{

#ifndef i486
    int i;
    if (!pert(nome, elemento)) {
	if (controle >= aux_insm) {
	    if ((aux_ins =
		 realloc(aux_ins,
			 (aux_insm + 50) * sizeof(char *))) == NULL) {
		fprintf(stderr, PegaMsg(1039));
		setOSstdin();
#ifdef MEM_DEBUG
		if (Mem_Used() != 0) {
		    printf("***Lista de memoria nao esta vazia ***\n");
		    Mem_Store("estat");
		    exit(1);
		}
#endif
		exit(1);

	    }
	    aux_insm += 50;
	    for (i = controle; i < aux_insm; i++) {
		if ((aux_ins[i] =
		     (char *) calloc(16, sizeof(char))) == NULL) {
		    fprintf(stderr, PegaMsg(1039));
		    setOSstdin();
#ifdef MEM_DEBUG
		    if (Mem_Used() != 0) {
			printf("***Lista de memoria nao esta vazia ***\n");
			Mem_Store("estat");
			exit(1);
		    }
#endif
		    exit(1);
		}
	    }
	}
	strcpy(aux_ins[controle], elemento);
	L_Insert(nome, aux_ins[controle], L_End(nome) + 1);
	controle++;
    }
#else
    char *quem_fica;
    if (!pert(nome, elemento)) {
	quem_fica = (char *) calloc((strlen(elemento) + 1), sizeof(char));
	strcpy(quem_fica, elemento);
	L_Insert(nome, quem_fica, L_End(nome) + 1);
    }
#endif

}

/* FUNCAO RET :
	Retira um elemento de um conjunto, onde :
		. o conjunto e' uma lista duplamente ligada
		. o elemento e' uma cadeia de caracteres
*/

void ret(nome, elemento)
Lista nome;
char *elemento;
{
    if (pert(nome, elemento))
	L_Delete(nome, L_Locate(nome, elemento));
}

/* FUNCAO PERT :
	Verifica se um elemento pertence a  um conjunto, onde :
		. o conjunto e' uma lista duplamente ligada
		. o elemento e' uma cadeia de caracteres
	Se a funcao que localiza elementos na lista devolver um endereco
	invalido e' porque o elemento nao pertencia ao conjunto(lista)
*/
int pert(nome, elemento)
Lista nome;
char *elemento;
{
    int ret;
    if (L_Locate(nome, elemento) != -1)
	ret = 1;
    else
	ret = 0;
    return (ret);
}

/* FUNCAO ELEM :
	Retorna a posicao do elemento no conjunto, onde :
		. o conjunto e' uma lista duplamente ligada
		. o elemento e' uma cadeia de caracteres
	A funcao corresponde, exatamente, a de localizacao de elementos
	na lista.
*/
int elem(nome, elemento)
Lista nome;
char *elemento;
{
    return (L_Locate(nome, elemento));
}

/* FUNCAO CARD :
	Retorna   o numero de elementos do conjunto 
		. o conjunto e' uma lista duplamente ligada
*/
int card(nome)
Lista nome;
{
    int n = 0;
    int i;

    for (i = 1; (L_Retrieve(nome, i)) != (Dado) NULL; i++)
	++n;
    return (n);
}

/* FUNCAO ECONT :
	Retorna 1 (verdadeiro) ou 0 (falso), caso o conjunto1 esteja 
	ou nao contido no conjunto2, onde :
		. o conjunto1 e' uma lista duplamente ligada
		. o conjunto2 e' uma lista duplamente ligada
	A funcao corresponde, a verificar se todo o elemento do con-
	junto1 pertence ao conjunto2.
	Obs : o conjunto vazio esta contido em qualquer outro conjun-
	to, assim como em outro vazio.
*/
int econt(nome1, nome2)
Lista nome1;
Lista nome2;
{
    char *P;
    int i;
    for (i = 1; (P = L_Retrieve(nome1, i)) != (Dado) NULL; i++) {
	if (!pert(nome2, P))
	    return (0);
    }
    return (1);
}

/* FUNCAO INTER :
	Retorna um conjunto que e' o resultado da interseccao do con-
	junto1 com o conjunto2 , onde :
		. o conjunto1 e' uma lista duplamente ligada
		. o conjunto2 e' uma lista duplamente ligada
	A funcao corresponde, exatamente, a verificar se cada elemento
	do conjunto1 pertence ao conjunto2, todo elemento que satisfaz
	esta condicao e' inserido no conjunto retornado.
*/
Lista inter(nome1, nome2)
Lista nome1;
Lista nome2;
{
    Lista nome3;
    char *P;
    int i;

    nome3 = L_Create();
    for (i = 1; (P = L_Retrieve(nome1, i)) != (Dado) NULL; i++)
	if (pert(nome2, P))
	    ins(nome3, P);
    return (nome3);
}
