/***********************************************************************************
 ESTAT - Clculo de Estatsticas Descritivas
 
 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
 termos da Licena Pblica Geral GNU conforme publicada pela Free Software
 Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
 posterior.

 Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
 programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
 Temple Street, Suite 330, Boston, MA 02111-1307 USA.

 **********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

    Mdulo integrante do software ESTAT, que  calcula estatsticas descritivas 
    de variveis  numricas. As  estatsticas  que  podem  ser requeridas so: nmero 
    de observaes, valor mnimo, valor mximo, total, mdia, varincia, desvio padro, 
    erro padro, coeficiente de variao, assimetria, curtose, mediana, estatstica D 
    de Kolmogorov, percentis e modas.
    Neste mdulo encontram-se as rotinas de controle da tabela de hashing:
    	. aloc_hash()	: aloca tabela e nomes	
	. hashcalc()	: calculos para uma tabela 	
	. ini_hash()	: inicializa/reinicializa controles quando do inicio de uma classe
	. le_proxima()	: retorna arquivo e monta tabela para proxima variavel 
	. libera()	: libera tabela e nomes	
	. t_hashing()	: montagem da tabela pra uma variavel 
	. trata_outras(): trata as proximas variaveis que usam hashing 	
	. var_hash()	: seleciona proxima variavel 	

 Desenvolvido por:

	  Dalton Francisco Andrade
	  Maria Fernanda Moura
	  Maurcio Noronha Festa
	  Slvio Roberto Medeiros Evangelista
	  Snia Ternes Frassetto

 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/
#include "estat.h"

#if !defined(VAXVMS) & !defined(SCO) & !defined(i486)
# include <search.h>
#endif

#  include <sys/types.h>

/*..................... Variaveis Externas ................................*/


extern int cpercentis;
extern int tamnome;
extern int tamhash;
extern int ehashing;
extern int ghashing;
extern int v_hash;
extern unsigned int nobs;
extern unsigned int obsant;

/*.........................................................................*/

void aloc_hash()
{
    /* aloca a tabela de hashing */

    tabhash[0] = (struct simb **) aloc(TAMTAB, sizeof(struct simb **));
    tamhash = TAMTAB;

    /* aloca o vetor de observacoes ( nomes ) */

    nomes[0] = (char **) aloc(TAMINI, sizeof(char **));	/*ilegal */
    tamnome = TAMINI;
}

/*.........................................................................*/

void ini_hash()
{
    apnome = 0;
    v_hash = -1;
    var_hash(&v_hash);
    aloc_hash();
    obsant = nobs;
}

/*.........................................................................*/

void le_proxima(num, nobs, obsant)
int num;
unsigned int nobs;
unsigned int obsant;
{
    long prox;

    rew_arq(fd);

    while ((prox = ler_obs(fd, obs)) < (long) obsant);

    do {
	if (strcmp(obs[posvar[num]], ".") == 0)
	    goto outra;
	if (pondera) {
	    if (strcmp(obs[pospeso], ".") == 0)
		goto outra;
	    wt = atof(obs[pospeso]);
	    if (wt <= 0.0)
		wt = 0.0;
	} else
	    wt = 1.0;
	if (nfreq) {
	    if (strcmp(obs[posfreq], ".") == 0)
		goto outra;
	    qfreq = atof(obs[posfreq]);
	    if (qfreq <= 0.0)
		goto outra;
	} else
	    qfreq = 1.0;
	t_hashing(num);
      outra:prox = ler_obs(fd, obs);
    } while ((prox != (long) EOF) && (prox < (long) nobs));
}

/*.........................................................................*/

void libera()
{
    free((char *) nomes[0]);
    free((char *) tabhash[0]);
}

/*.........................................................................*/

void t_hashing(num)
int num;
{
    struct simb *p;
    char ss[128];

    strcpy(ss, obs[posvar[num]]);
    p = pesquisa(ss, tabhash[0]);
    if (p != SNULL)
	p->def += qfreq;
    else {
	p = instala(ss, qfreq, tabhash[0]);
	nomes[0][apnome++] = p->nome;
	if (apnome >= tamnome) {
	    tamnome += TAMINI;
	    nomes[0] = (char **) realoca(nomes[0], sizeof(char **),
					 tamnome);
	}
    }
}

/*.........................................................................*/

void trata_outras(obsant, v_hash, nobs)
unsigned int *obsant;
int *v_hash;
unsigned int *nobs;
{
    double pesoant, freqant;

    pesoant = wt;
    freqant = qfreq;

    hashcalc(v_hash);
    var_hash(v_hash);
    while (*v_hash != -1) {
	libera();
	aloc_hash();
	apnome = 0;
	le_proxima(*v_hash, *nobs, *obsant);
	hashcalc(v_hash);
	var_hash(v_hash);
    }
    libera();

    wt = pesoant;
    qfreq = freqant;
}

/*.........................................................................*/

void var_hash(num)
int *num;
{
    int i;
    char nome[MAXNOME];
    double mmin, mmax;
    char tipo;

    for (i = (*num + 1); i < nvar; i++) {
	if (ehashing) {
	    *num = i;
	    break;
	}
	pega_var(fd, posvar[i], nome, &tipo, (MINVAL *) & mmin,
		 (MAXVAL *) & mmax);
	if (nrf && pert(rf, nome)) {
	    *num = i;
	    break;
	}
	if (nbp && pert(bp, nome)) {
	    *num = i;
	    break;
	}
    }
    if (i == nvar)
	*num = -1;
}

/*.........................................................................*/

void hashcalc(v)
int *v;
{
    extern char *gcvt();
    int tam;
    char nome[50];
    char a[50];
    double valor;
    int (*Fcompn) ();

    tam = apnome;
    Fcompn = compn;

    qsort((void *) nomes[0], (size_t) tam, (size_t) sizeof(char *),
	  Fcompn);

    if (ehashing) {
	if (pert(est, "moda")) {
	    moda[*v] = calc_moda(tam);
	}
	if (pert(est, "mediana")) {
	    mediana[*v] = calc_percentil(tam, 0.50);
	}
	if (cpercentis) {
	    percentis[*v] = L_Create();
	    if (pert(est, "p05")) {
		valor = calc_percentil(tam, 0.05);
		strcpy(nome, "p05");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p10")) {
		valor = calc_percentil(tam, 0.10);
		strcpy(nome, "p10");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p15")) {
		valor = calc_percentil(tam, 0.15);
		strcpy(nome, "p15");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p20")) {
		valor = calc_percentil(tam, 0.20);
		strcpy(nome, "p20");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p25")) {
		valor = calc_percentil(tam, 0.25);
		strcpy(nome, "p25");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p30")) {
		valor = calc_percentil(tam, 0.30);
		strcpy(nome, "p30");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p35")) {
		valor = calc_percentil(tam, 0.35);
		strcpy(nome, "p35");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p40")) {
		valor = calc_percentil(tam, 0.40);
		strcpy(nome, "p40");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p45")) {
		valor = calc_percentil(tam, 0.45);
		strcpy(nome, "p45");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p50")) {
		valor = calc_percentil(tam, 0.50);
		strcpy(nome, "p50");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p55")) {
		valor = calc_percentil(tam, 0.55);
		strcpy(nome, "p55");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p60")) {
		valor = calc_percentil(tam, 0.60);
		strcpy(nome, "p60");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p65")) {
		valor = calc_percentil(tam, 0.65);
		strcpy(nome, "p65");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p70")) {
		valor = calc_percentil(tam, 0.70);
		strcpy(nome, "p70");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p75")) {
		valor = calc_percentil(tam, 0.75);
		strcpy(nome, "p75");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p80")) {
		valor = calc_percentil(tam, 0.80);
		strcpy(nome, "p80");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p85")) {
		valor = calc_percentil(tam, 0.85);
		strcpy(nome, "p85");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p90")) {
		valor = calc_percentil(tam, 0.90);
		strcpy(nome, "p90");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	    if (pert(est, "p95")) {
		valor = calc_percentil(tam, 0.95);
		strcpy(nome, "p95");
		strcat(nome, gcvt(valor, 15, a));
		ins(percentis[*v], nome);
	    }
	}
	if (pert(est, "D_Kolmog.")) {
	    kolmogorov[*v] = calc_kolmogorov(tam);
	}
    }
}

int compn(xx, yy)
void *xx;
void *yy;
{
    char **x;
    char **y;

    x = (char **) xx;
    y = (char **) yy;

    if (atof(*x) > atof(*y))
	return (1);
    else
	return (-1);
}
