/***********************************************************************************
 ESTAT - Clculo de Estatsticas Descritivas
 
 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
 termos da Licena Pblica Geral GNU conforme publicada pela Free Software
 Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
 posterior.

 Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
 programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
 Temple Street, Suite 330, Boston, MA 02111-1307 USA.

 **********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

    Mdulo integrante do software ESTAT, que  calcula estatsticas descritivas 
    de variveis  numricas. As  estatsticas  que  podem  ser requeridas so: nmero 
    de observaes, valor mnimo, valor mximo, total, mdia, varincia, desvio padro, 
    erro padro, coeficiente de variao, assimetria, curtose, mediana, estatstica D 
    de Kolmogorov, percentis e modas.
    Este  o mdulo que se encarrega de exibir os clculos das estatsticas. 
    
 Desenvolvido por:

 	  Dalton Francisco Andrade
 	  Maria Fernanda Moura
	  Maurcio Noronha Festa
	  Slvio Roberto Medeiros Evangelista
	  
 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/

#include "estat.h"
#ifdef MEM_DEBUG
#include "mshell.h"
#endif

/* ----------- ordem de impressao para as estatisticas ---------------*/

static char *ordem[] = { "n", "minimo", "maximo", "total", "media",
    "variancia", "desv.padrao", "err.padrao", "coef.var.",
    "assimetria", "curtose", "mediana", "D_Kolmog.", "p05",
    "p10", "p15", "p20", "p25", "p30", "p35", "p40", "p45",
    "p50",
    "p55", "p60", "p65", "p70", "p75", "p80", "p85", "p90",
    "p95", "moda1", "moda2", "moda3", "moda4", "moda5",
    0
};

/* ----------- tamanho total do formato de cada estatistica ----------*/

static int tot[] = { 5, 11, 11, 12, 11,
    15, 13, 11, 13,
    11, 11, 11, 11, 11,
    11, 11, 11, 11, 11, 11, 11, 11,
    11,
    11, 11, 11, 11, 11, 11, 11, 11,
    11, 11, 11, 11, 11, 11,
    0
};

/* ----------- tamanho das decimais de cada formato respectivo --------*/

static int dec[] = { 0, 4, 4, 4, 4,
    4, 4, 4, 4,
    4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4, 4, 4,
    4,
    4, 4, 4, 4, 4, 4, 4, 4,
    4, 4, 4, 4, 4, 4,
    0
};

/* -----------  tamanho do vetor ordem ---------------------------------*/

int t_ord = 37;

/* ----- quais estatisticas serao impressas na pagina corrente ---------*/

int quais[20];

/* ---------------- tamanho em colunas da pagina corrente --------------*/

int tam;

/* ----------------- flags do programa principal ---------------------- */

extern int flagP;		/* mudanca de pagina            */
extern int linhas;		/* numero de linhas por pagina  */
extern int ncolunas;		/* numero de colunas por pagina */

/* -------------------------------------------------------------------- */

void imprime()
{
    int i, k;
    int tem = 0;

    tem = ins_modas();
    tam = 8;
    i = 0;
    for (k = 0; k < t_ord; k++) {
	if (pert(est, ordem[k])) {
	    if ((tam + 1 + tot[k]) <= ncolunas) {
		tam += (tot[k] + 1);
		quais[i++] = k;
	    } else {
		relatorio(i);
		i = 0;
		quais[i++] = k;
		tam = 8 + tot[k] + 1;
	    }
	}
    }
    relatorio(i);
    if (tem)
	observacoes(tem);
}

/* -------------------------------------------------------------------- */

void relatorio(lim)
int lim;
{
    int i, k;
    int brancos;
    int contlinha = 0;
    double valor;

    brancos = (ncolunas - tam) / 2;

    for (i = 0; i < nvar; i++) {
	if (contlinha == 0)
	    contlinha = imptl(brancos, lim);
	if (brancos > 0)
	    printf("%*s", brancos, " ");
	printf("%-8s ", center(L_Retrieve(var, (i + 1)), 8));
	for (k = 0; k < lim; k++) {
	    valor = calculo(quais[k], i);
	    if (valor == MISSING) {
		printf("%*s ", tot[quais[k]], center(".", tot[quais[k]]));
	    } else {
		if (fabs(valor) > pow(10.0, (double)
				      (tot[quais[k]] - dec[quais[k]] - 1))) {
		    /* Antes :
		       printf("%*.*g ", tot[quais[k]],dec[quais[k]],
		       valor);
		     */
		    char buf[20];
		    printf("%*s ", tot[quais[k]],
			   gcvt(valor, (tot[quais[k]] - 5), buf));
		} else
		    printf("%*.*f ", tot[quais[k]], dec[quais[k]], valor);
	    }
	}
	printf("\n");
	if ((++contlinha) >= linhas)
	    contlinha = 0;
    }
}

/* -------------------------------------------------------------------- */

int imptl(brancos, lim)
int brancos;
int lim;
{
    int cont = 0;
    int i, t, aux, de;

    if (flagP)
	printf("\f");
    else
	printf("\n");
    for (i = 0; i < ntit; i++)
	printf("%-s\n", center(titulo[i], ncolunas));
    printf("\n");
    cont = ntit + 1;
    if (nby) {
	t = de = 0;
	for (i = 0; i < nby; i++) {
	    if (tipby[i] == 'd')
		aux = strlen(L_Retrieve(class, (i + 1))) + 14;
	    else
		aux = strlen(L_Retrieve(class, (i + 1))) +
		    strlen(ant[i]) + 4;
	    if ((t + aux) < tam)
		t += aux;
	    else {
		impby(tam, t, brancos, de, i);
		t = aux;
		de = i;
		++cont;
	    }
	}
	impby(tam, t, brancos, de, i);
	++cont;
    } else {
	if (brancos > 0)
	    printf("%*s", brancos, " ");
	for (i = 0; i < tam; i++)
	    printf("-");
	printf("\n");
	++cont;
    }
    if (brancos > 0)
	printf("%*s", brancos, " ");
    printf("%-8s ", center("variavel", 8));
    for (i = 0; i < lim; i++)
	printf("%*s ", tot[quais[i]], center(ordem[quais[i]],
					     tot[quais[i]]));
    printf("\n");
    ++cont;
    if (brancos > 0)
	printf("%*s", brancos, " ");
    for (i = 0; i < tam; i++)
	printf("-");
    printf("\n");
    ++cont;
    return (cont);
}

/* -------------------------------------------------------------------- */

void impby(tam, tamby, brancos, de, ate)
int tam;
int tamby;
int brancos;
int de;
int ate;
{
    int k, impar, i;

    k = (tam - tamby) / 2;
    impar = (tam - tamby) % 2;
    if (brancos > 0)
	printf("%*s", brancos, " ");
    for (i = 1; i <= k; i++)
	printf("-");
    for (i = de; i < ate; i++)
	if (tipby[i] == 'd') {
	    char pdata[15];

	    dtoa(atof(ant[i]), pdata);
	    printf(" %s= %s ", L_Retrieve(class, (i + 1)), pdata);
	} else
	    printf(" %s= %s ", L_Retrieve(class, (i + 1)), ant[i]);
    k += impar;
    for (i = 1; i <= k; i++)
	printf("-");
    printf("\n");
}

/* -------------------------------------------------------------------- */

char *center(cadeia, tam)
char *cadeia;
int tam;
{
    static char buf[256];
    char *char_str();

    strcpy(buf, char_str((tam - strlen(cadeia)) / 2, '\040'));

    strcat(buf, cadeia);

    return (buf);
}

/* -------------------------------------------------------------------- */

char *char_str(comp, c)

int comp;
char c;

{

    static char loc[256];

    char *ptr = loc;

    while (comp--)
	*ptr++ = c;

    *ptr = '\0';

    return (loc);

}

/* -------------------------------------------------------------------- */

int ins_modas()
{
    int i, c;
    int maior = 0;
    int sai = 0;
    char nome[10], buffer[10];

    if (pert(ests, "moda"))
	sai = 1;
    if (pert(est, "moda")) {
	maior = 1;
	for (i = 0; i < nvar; i++)
	    if ((c = card(moda[i])) > maior)
		maior = c;
    }
    if (maior) {
	for (i = 1; (i <= MAXMODA) && (i <= maior); i++) {
	    strcpy(nome, "moda");
	    strcat(nome, itoa(i, buffer, 10));
	    ins(est, nome);
	    if (sai)
		ins(ests, nome);
	}
    }
    return (maior);
}

/* -------------------------------------------------------------------- */

void observacoes(tem)
int tem;
{
    int i, c;
    int prim = 0;

    for (i = 0; i < nvar; i++) {
	if (((c = card(moda[i])) > MAXMODA) || (c == 0)) {
	    if (prim == 0) {
		if (flagP)
		    printf("\f");
		else
		    printf("\n\n");
		printf("Variavel : Observacao sobre a Moda\n");
		printf("==================================\n\n");
		prim = 1;
	    }
	    if (c == 0)
		mensagem(0, i);
	    if (c > MAXMODA)
		mensagem(1, i);
	}
    }
}
static char *msg[] = {
    "Todas as observacoes tem frequencia um",
    "Existem  mais de cinco modas",	/* MAXMODA */
    0
};
void mensagem(n, nv)
int n;
int nv;
{
    printf("%-8s : ", center(L_Retrieve(var, (nv + 1)), 8));
    printf("%s\n\n", msg[n]);
}

/* -------------------------------------------------------------------- */

double calculo(k, i)
int k;				/* estatistica */
int i;				/* variavel    */
{
    double aux;
    double dp;
    int q;
    Dado P;

    if (k == 0)			/* n , numero de observacoes */
	return (wfreq[i]);

    if (k == 1)			/* minimo */
	return (xmin[i]);

    if (k == 2)			/* maximo */
	return (xmax[i]);

    if (k == 3)			/* total , sem ponderacao */
	return (soma[i]);

    if (k == 4)			/* media , com ponderacao */
	return (xmean[i]);

    if (k == 5) {		/* variancia , com ponderacao */
	if (wsum[i] <= 1.0) {
	    if (wsum[i] == 1.0)
		aux = 0.0;
	    else
		aux = MISSING;
	}

	else {
	    aux = s2[i] / (wfreq[i] - 1);
	}
    }

    if (k == 6) {		/* desvio padrao , com ponderacao */
	if (wsum[i] <= 1.0) {
	    if (wsum[i] == 1.0)
		aux = 0.0;
	    else
		aux = MISSING;
	}

	else {
	    aux = sqrt(s2[i] / (wfreq[i] - 1));
	}
    }

    if (k == 7) {		/* erro padrao , com ponderacao */
	if (wsum[i] <= 1.0) {
	    if (wsum[i] == 1.0)
		aux = 0.0;
	    else
		aux = MISSING;
	}

	else {
	    aux = sqrt(s2[i] / (wfreq[i] - 1)) / sqrt(wfreq[i]);
	}
    }

    if (k == 8) {		/* coef.variacao , com ponderacao */
	if (xmean[i] == 0.0)
	    return (MISSING);

	if (wfreq[i] <= 1.0) {
	    if (wsum[i] == 1.0) {
		aux = 0.0;
	    }

	    else {
		aux = MISSING;
	    }
	}

	else {
	    aux = 100 * sqrt(s2[i] / (wfreq[i] - 1)) / xmean[i];
	}
    }

    if (k == 9) {		/* assimetria , sem ponderacao */
	if (wfreq[i] <= 2.0)
	    aux = MISSING;

	else {
	    dp = sqrt(s22[i] / (wfreq[i] - 1));

	    if (dp <= 0.0) {
		aux = 0.0;
	    }

	    else {
		aux = (wfreq[i] * s3[i]) /
		    ((wfreq[i] - 1) * (wfreq[i] - 2) * dp * dp * dp);
	    }
	}
    }

    if (k == 10) {		/* curtose , sem ponderacao */
	if (wfreq[i] <= 3.0)
	    aux = MISSING;

	else {
	    dp = sqrt(s22[i] / (wfreq[i] - 1));

	    if (dp <= 0.0)
		aux = MISSING;

	    else {
		aux = (wfreq[i] * (wfreq[i] + 1) *
		       s4[i] - 3.0 * s22[i] * s22[i] * (wfreq[i] - 1)) /
		    ((wfreq[i] - 1) * (wfreq[i] - 2) * (wfreq[i] - 3) * dp
		     * dp * dp * dp);
	    }
	}
    }

    if (k == 11) {
	return (mediana[i]);
    }

    if (k == 12) {
	return (kolmogorov[i]);
    }

    if (k == 13) {		/* p05 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p05", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 14) {		/* p10 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p10", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 15) {		/* p15 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p15", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 16) {		/* p20 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p20", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 17) {		/* p25 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p25", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 18) {		/* p30 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p30", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 19) {		/* p35 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p35", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 20) {		/* p40 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p40", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 21) {		/* p45 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p45", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 22) {		/* p50 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p50", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 23) {		/* p55 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p55", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 24) {		/* p60 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p60", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 25) {		/* p65 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p65", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 26) {		/* p70 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p70", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 27) {		/* p75 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p75", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 28) {		/* p80 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p80", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 29) {		/* p85 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p85", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 30) {		/* p90 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p90", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 31) {		/* p95 */
	if (wsum[i] == 0.0)
	    return (MISSING);

	for (q = 1; (P = L_Retrieve(percentis[i], q)) != DNULL; q++) {
	    if (strncmp(P, "p95", 3) == 0) {
		aux = atof(&P[3]);

		break;
	    }
	}
    }

    if (k == 32) {		/* moda 1 */
	if (card(moda[i]) >= 1) {
	    P = L_Retrieve(moda[i], 1);
	    aux = atof(P);
	}

	else
	    aux = MISSING;
    }

    if (k == 33) {		/* moda 2 */
	if (card(moda[i]) >= 2) {
	    P = L_Retrieve(moda[i], 2);
	    aux = atof(P);
	}

	else
	    aux = MISSING;
    }

    if (k == 34) {		/* moda 3 */
	if (card(moda[i]) >= 3) {
	    P = L_Retrieve(moda[i], 3);
	    aux = atof(P);
	}

	else
	    aux = MISSING;
    }

    if (k == 35) {		/* moda 4 */
	if (card(moda[i]) >= 4) {
	    P = L_Retrieve(moda[i], 4);
	    aux = atof(P);
	}

	else
	    aux = MISSING;
    }

    if (k == 36) {		/* moda 5 */
	if (card(moda[i]) >= 5) {
	    P = L_Retrieve(moda[i], 5);
	    aux = atof(P);
	}

	else
	    aux = MISSING;
    }

    return (aux);
}

/* -------------------------------------------------------------------- */

void outhdr()
{
    static char *sufixo[] =
	{ "n", "mn", "mx", "tt", "m", "v", "dp", "ep", "cv",
	"a", "c", "me", "D", "p05", "p10", "p15", "p20", "p25", "p30",
	    "p35", "p40",
	"p45", "p50", "p55", "p60", "p65", "p70", "p75", "p80", "p85",
	    "p90", "p95",
	"mo1", "mo2", "mo3", "mo4", "mo5", 0
    };
    int i, k;
    char tipo;
    char nome[16];
    double mmin, mmax;
/*	int ret;  Luciana */
    char tmpbuf[128];		/* Luciana */
    long ret;

    for (i = 0; i < nby; ++i) {
	if ((ret =
	     (long) pega_var(fd, posclass[i], nome, &tipo,
			     (MINVAL *) & mmin, (MAXVAL *) & mmax)) < 0L) {
	    fprintf(stderr, PegaMsg(1062));
#ifdef MEM_DEBUG
	    if (Mem_Used() != 0) {
		printf("***Lista de memoria nao esta vazia ***\n");
		Mem_Store("estat");
		exit(0);
	    }
#endif
	    setOSstdin();
	    exit(2);
	}
	if (ret = def_var(fout, nome, tipo) == -1) {
	    sprintf(tmpbuf, PegaMsg(1013), nome, ret);
	    fprintf(stderr, tmpbuf);
	    setOSstdin();
#ifdef MEM_DEBUG
	    if (Mem_Used() != 0) {
		printf("***Lista de memoria nao esta vazia ***\n");
		Mem_Store("estat");
		exit(100);
	    }
#endif

	    exit(100);
	}
    }
    for (i = 0; i < nvars; ++i) {
	tipo = 'n';
	for (k = 0; k < t_ord; k++) {
	    if (pert(ests, ordem[k])) {
		strcpy(nome, L_Retrieve(vars, (i + 1)));
		if (strlen(nome) > 5)
		    nome[5] = '\0';
		strcat(nome, sufixo[k]);
		if (ret = def_var(fout, nome, tipo) == -1) {
		    sprintf(tmpbuf, PegaMsg(1013), nome, ret);
		    fprintf(stderr, tmpbuf);
		    setOSstdin();
#ifdef MEM_DEBUG
		    if (Mem_Used() != 0) {
			printf("***Lista de memoria nao esta vazia ***\n");
			Mem_Store("estat");
			exit(100);
		    }
#endif
		    exit(100);
		}
	    }
	}
    }
}

/* -------------------------------------------------------------------- */

char *obs2[MAXVAR];
char obsxxx[MAXTOBS];
char *p2;
void grava_ntia()
{
    extern char *gcvt();
    int i, k, prim;
    int v;			/* posicao da variavel no conjunto "var" */
    double valor;
    long ret;

    p2 = obsxxx;
    for (i = 0; i < nby; i++) {
	obs2[i] = p2;
	strcpy(p2, ant[i]);
	p2 += (strlen(p2) + 1);
    }
    prim = nby;
    for (i = 0; i < nvars; i++) {
	v = elem(var, L_Retrieve(vars, (i + 1)));
	for (k = 0; k < t_ord; k++) {
	    if (pert(ests, ordem[k])) {
		obs2[prim++] = p2;
		valor = calculo(k, (v - 1));
		if (valor == MISSING)
		    strcpy(p2, ".");
		else
		    gcvt(calculo(k, (v - 1)), 15, p2);
		p2 += (strlen(p2) + 1);
	    }
	}
    }
    if ((ret = grav_obs(fout, obs2)) < 0L) {
	fprintf(stderr, PegaMsg(5018), "");
#ifdef MEM_DEBUG
	if (Mem_Used() != 0) {
	    printf("***Lista de memoria nao esta vazia ***\n");
	    Mem_Store("estat");
	    exit(0);
	}
#endif
	setOSstdin();
	exit(2);
    }
}
