/***********************************************************************************
 ESTAT - Clculo de Estatsticas Descritivas
 
 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
 termos da Licena Pblica Geral GNU conforme publicada pela Free Software
 Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
 posterior.

 Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
 programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
 Temple Street, Suite 330, Boston, MA 02111-1307 USA.

 **********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

    Mdulo integrante do software ESTAT, que  calcula estatsticas descritivas 
    de variveis  numricas. As  estatsticas  que  podem  ser requeridas so: nmero 
    de observaes, valor mnimo, valor mximo, total, mdia, varincia, desvio padro, 
    erro padro, coeficiente de variao, assimetria, curtose, mediana, estatstica D 
    de Kolmogorov, percentis e modas.
    Este  o mdulo que implementa os operadores usados pelo tipo Lista Encadeada
    O tipo lista implementado aqui foi baseado na lista definida em "Aho, A.V. & Hopcroft, 
    J.E & Ullman, J.D.Data Structures and Algorithms, Addison-Wesley, 1983"
    
 Desenvolvido por:

 	  Antonio Luis Peloso
	  Carla Geovana do Nascimento Macrio
	  Fani Barbosa Camargo
	  Maria Fernanda Moura
	  
 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/

#include	<stdio.h>
#include	<string.h>
#include	"atomic.h"
#include	"memoria.h"
#include	"lista.h"
#include	"pcsgbd.h"	/* Luciana */

int ldebug = 0;

Lista L_Create()
{

    if (ldebug) {
	printf("Create \n");
	printf("----------------\n");
    }

    return ((Lista) M_Aloca(1, L_Size(), __LISTA));
}

void L_Destroy(L)
Lista L;
{
    No *Tmp, *Lnk;

    Lnk = Tmp = L->Prim;
    while (Tmp) {
	Tmp = Tmp->Prox;
	M_Libera((Memoria) Lnk, __LISTA);
	Lnk = Tmp;
    }
    M_Libera((Memoria) L, __LISTA);
}

void L_MakeNull(L)
Lista L;
{
    No *Tmp, *Lnk;

    Lnk = Tmp = L->Prim;
    while (Tmp) {
	Tmp = Tmp->Prox;
	M_Libera((Memoria) Lnk, __LISTA);
	Lnk = Tmp;
    }

    L->Tam = 0;
    L->Prim = (No *) NULL;
    L->Ulti = (No *) NULL;
    L = (Lista) NULL;
}

void L_Insert(L, Elem, Pos)
Lista L;
Dado Elem;
Node Pos;
{
    No *Novo, *Aux;

/****	Se Lista ainda nao foi criada retorne ERRO	****/

    if (Pos < 0 || L == (Lista) NULL) {
	printf(PegaMsg(5029));
	return;
    }

/****	Primeiro no a ser inserido	****/

    if (L->Prim == (No *) NULL) {
	Novo = L->Prim = L->Ulti =
	    (No *) M_Aloca(1, sizeof(struct No), __LISTA);
	Novo->Posicao = Pos;
	Novo->Elemento = Elem;
	L->Tam = 1;
    }

    else {
	Aux = L->Prim;

/****	Posiciona local de insercao	****/

	for (; ((Aux->Prox != NULL) && Pos >= Aux->Prox->Posicao);
	     Aux = Aux->Prox);


	if (Aux->Posicao == Pos)      /****	Insere na mesma posicao	****/
	    Aux->Elemento = Elem;

	else {
	    Novo = (No *) M_Aloca(1, sizeof(struct No), __LISTA);
	    Novo->Posicao = Pos;
	    Novo->Elemento = Elem;

	    if (Aux->Ant == NULL && Aux->Posicao > Pos) {
/****	Insere no inicio ****/
		Novo->Prox = Aux;
		Novo->Ant = NULL;
		Aux->Ant = L->Prim = Novo;
		++(L->Tam);
	    }

	    else if (Aux->Prox == NULL) {
/** Insere no fim **/
		Novo->Prox = NULL;
		Novo->Ant = Aux;
		Aux->Prox = L->Ulti = Novo;
		++(L->Tam);
	    }

	    else {
/****	Insere no meio	****/

		Novo->Prox = Aux->Prox;
		Novo->Ant = Aux;
		Aux->Prox = Aux->Prox->Ant = Novo;
		++(L->Tam);
	    }
	} /** fim do else **/
    }	  /** fim do else **/
    if (ldebug) {
	for (Aux = L->Prim; Aux != NULL; Aux = Aux->Prox) {
	    printf("Insert!!!\tTam: %d\n", L->Tam);
	    printf("L->Prim: %s\n", L->Prim->Elemento);
	    printf("Elemento: %s\tPosicao: %d\n", Aux->Elemento,
		   Aux->Posicao);
	}
	printf("------------------------\n");
    }
}

void L_Delete(L, N)
Lista L;
Node N;
{

    No *Pont;

/****	Se Lista ainda nao foi criada retorne ERRO	****/

    if (L == (Lista) NULL || L->Tam == 0) {
	printf(PegaMsg(5030));
	return;
    }

/****	Posiciona local de insercao	****/

    for (Pont = L->Prim; Pont->Posicao != N && Pont != NULL;
	 Pont = Pont->Prox);

    if (Pont->Posicao != N)	    /** Se posicao invalida, retorne ERRO **/
	printf(PegaMsg(5031));

    if (Pont == L->Prim) {
/** deleta primeiro no **/
	L->Prim = Pont->Prox;
	if (L->Tam != 1)
	    Pont->Prox->Ant = NULL;
	M_Libera((Memoria) Pont, __LISTA);
    }

    else if (Pont == L->Ulti) {
/** deleta ultimo no **/
	L->Ulti = Pont->Ant;
	Pont->Ant->Prox = NULL;
	M_Libera((Memoria) Pont, __LISTA);
    }

    else {
/** deleta no' no meio da lista **/

	Pont->Ant->Prox = Pont->Prox;
	Pont->Prox->Ant = Pont->Ant;
	M_Libera((Memoria) Pont, __LISTA);
    }

    --(L->Tam);
    if (ldebug) {
	No *Aux;
	for (Aux = L->Prim; Aux != NULL; Aux = Aux->Prox) {
	    printf("Delete!!\tTam: %d\n", L->Tam);
	    printf("L->Prim: %s\n", L->Prim->Elemento);
	    printf("Elemento:\t%s\n", Aux->Elemento);
	}
	printf("---------------------\n");
    }
}

void L_Split(L, L1, L2, pos)
Lista L, L1, L2;
Node pos;
{
    No *Pont;

    if (L == (Lista) NULL || L->Tam == 0 || L1 == (Lista) NULL
	|| L2 == (Lista) NULL) {
	printf(PegaMsg(5030));
	return;
    }

    for (Pont = L->Prim; Pont->Posicao != pos && Pont != NULL;
	 Pont = Pont->Prox) {
	L_Insert(L1, Pont->Elemento, Pont->Posicao);
    }
    if (ldebug) {
	No *Aux;
	printf("----------------------- SPLIT --------------------\n");
	for (Aux = L1->Prim; Aux != NULL; Aux = Aux->Prox)
	    printf("L1->Pos: %d\tL1->Elem%s\n", Aux->Posicao,
		   Aux->Elemento);
    }

    if (Pont->Posicao != pos)
	printf(PegaMsg(5031));

    for (; Pont != NULL; Pont = Pont->Prox) {
	L_Insert(L2, Pont->Elemento, Pont->Posicao);
    }
    if (ldebug) {
	No *Aux;
	printf("----------------------- SPLIT2 --------------------\n");
	for (Aux = L2->Prim; Aux != NULL; Aux = Aux->Prox)
	    printf("L2->Pos: %d\tL2->Elem%s\n", Aux->Posicao,
		   Aux->Elemento);
    }
}

void L_Join(L, L1, L2)
Lista L, L1, L2;
{
    No *Pont;
    int pos;

    if (L == (Lista) NULL || L->Tam == 0 || L1 == (Lista) NULL
	|| L2 == (Lista) NULL) {
	printf(PegaMsg(5030));
	return;
    }

    for (Pont = L1->Prim, pos = Pont->Posicao; Pont != NULL;
	 Pont = Pont->Prox, ++pos) {
	L_Insert(L, Pont->Elemento, pos);
    }

    for (Pont = L2->Prim; Pont != NULL; Pont = Pont->Prox, ++pos) {
	L_Insert(L, Pont->Elemento, pos);
    }

    if (ldebug) {
	No *Aux;
	printf("----------------------- JOIN --------------------\n");
	for (Aux = L->Prim; Aux != NULL; Aux = Aux->Prox)
	    printf("L->Pos: %d\tL->Elem%s\n", Aux->Posicao, Aux->Elemento);
    }
}

void L_Swap(L, p1, p2)
Lista L;
Node p1, p2;
{
    No *Pont1, *Pont2;

    if (L == (Lista) NULL || L->Tam == 0) {
	printf(PegaMsg(5030));
	return;
    }

    for (Pont1 = L->Prim; Pont1->Posicao != p1 && Pont1->Prox != NULL;
	 Pont1 = Pont1->Prox);
    if (Pont1->Posicao != p1)
	printf(PegaMsg(5031));

    for (Pont2 = L->Prim; Pont2->Posicao != p2 && Pont2 != NULL;
	 Pont2 = Pont2->Prox);
    if (Pont2->Posicao != p2)
	printf(PegaMsg(5031));

    TrocaDado(Pont1->Elemento, Pont2->Elemento);
}

Dado TrocaDado(d1, d2)
Dado d1, d2;
{
    Dado aux;

    aux = M_Aloca(1, sizeof(Dado), __LISTA);

    aux = strcpy(aux, d1);
    d1 = strcpy(d1, d2);
    d2 = strcpy(d2, aux);
    return (aux);
}


Dado L_Retrieve(L, P)
Lista L;
int P;
{
    No *Pont;

/****	Se Lista ainda nao foi criada retorne ERRO	****/

    if (P < 0 || L == (Lista) NULL)
	return (NULL);

    for (Pont = L->Prim; Pont != NULL && Pont->Posicao != P;
	 Pont = Pont->Prox);

    if (Pont == NULL)
	return (NULL);

    return (Pont->Elemento);
}

int L_Locate(L, E)
Lista L;
Dado E;
{
    No *Pont;
    int i;
    i = -1;

    if (ldebug) {
	printf("1 LOCATE!!\n");
	printf("  Elemento: %s\n", E);
    }

/****	Se Lista ainda nao foi criada retorne ERRO	****/

    if (L == (Lista) NULL)
	return (-1);

    for (Pont = L->Prim;
	 Pont != NULL && (i = (strcmpi(Pont->Elemento, E)) != 0);
	 Pont = Pont->Prox) {

    }

    if (i != 0)
	return (-1);
    if (ldebug) {
	printf("2 LOCATE!!\n");
	printf("  Elemento1: %s\tPont->Elem: %s\n", E, Pont->Elemento);
    }

    return (Pont->Posicao);
}

int L_Next(L, P)
Lista L;
int P;
{
    No *Pont;

/****	Se Lista ainda nao foi criada retorne ERRO	****/

    if (P < 0 || L == (Lista) NULL)
	return (-1);

    for (Pont = L->Prim; Pont != NULL && Pont->Posicao != P;
	 Pont = Pont->Prox);

    if (Pont->Posicao != P || Pont->Prox == NULL)
	return (-1);

    return (Pont->Prox->Posicao);
}

int L_Previous(L, P)
Lista L;
int P;
{
    No *Pont;

/****	Se Lista ainda nao foi criada retorne ERRO	****/

    if (P < 0 || L == (Lista) NULL)
	return (-1);

    for (Pont = L->Prim; Pont != NULL && Pont->Posicao != P;
	 Pont = Pont->Prox);

    if (Pont->Posicao != P)
	return (-1);

    return (Pont->Ant->Posicao);
}

int L_End(L)
Lista L;
{
/****	Se Lista ainda nao foi criada retorne ERRO	****/

    if (L == (Lista) NULL)
	return (-1);

    return ((int) (L->Tam == 0 ? (Node) NULL : L->Ulti->Posicao));
}

Dado L_First(L)
Lista L;
{
/****	Se Lista ainda nao foi criada retorne ERRO	****/

    if (L == (Lista) NULL)
	return (NULL);

    return (L->Tam == 0 ? (Dado) NULL : L->Prim->Elemento);
}

/*
//	Final de lista.c
*/
