/***********************************************************************************
 Genese - Linguagem de programao para a gerao e manipulao de arquivo .dat

 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
 termos da Licena Pblica Geral GNU conforme publicada pela Free Software
 Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
 posterior.

 Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
 programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
 Temple Street, Suite 330, Boston, MA 02111-1307 USA.

 ***********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

	Arquivo que contem as funcoes utilizadas para a concatenacao de arquivos

 Desenvolvido por:

      Carlos Eduardo Assis Paniago
	Oscar Ivan Palma Pacheco
	Maria Fernanda Moura
	Adriana Delfino dos Santos
	

 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#ifndef VAXVMS
#include <malloc.h>
#endif

#include "socasm.h"
#include "pcsgbd.h"
#include "genese.h"

#ifdef MEM_DEBUG
#include "mshell.h"
#endif


/* ----------------------------------------------------------------------------
**
**	
**
** Arquivo que contem as funcoes para realizar a concatenacao de arquivos
**
** Funcoes:
**
**		. concv(prim, ult, chaves, nch, area): escolhe o arquivo mestre
**		para a concatenacao vertical. Caso nao exista arq. mestre, nao 
**		realiza a concatenacao por nao existirem arqs.. Caso contrario,
**		le uma obs. neste arq. e verifica se e' EOF, escolhendo neste 
**		caso outro arq. mestre. Ao final, se for possivel a concatenacao
**		coloca o no. da obs. corrente no mestre na pilha de dados.
**
**
**		. CvEscolheMestre(mestre, chave, nch, chaves, prim, ult, 
**		tip_ord): escolhe o arq. mestre , dentre os arqs. que serao 
**		usados para concatenacao, de acordo com o tipo de ordenacao, 
**		se existir. A funcao coloca em mestre o indice do arq. mestre.
**
**		. ChErro(): imprime mensagem de advertencia quando os arquivos
**		nao foram previamente ordenados para a concatenacao.
**
**		. CvLeCopiaChave(qual, ChAtual, chaves, nch): le uma obs. no
**		arq. de indice qual, copiando seu conteudo para ChAtual, que
**		contem as vars. que compoem a chave de ordenacao (chaves[]). 
**		Retorna o apontador do arq. para a sua posicao antes da leitura
** 		da observacao. Coloca valor missing nas vars. do arquivo e 
**		retorna o no. da obs. lida.
**
**		. CvCompara(mestre, lida, nch, chaves): compara o conteudo das
** 		vars. contidas em mestre com as vars. de lida, retornando 1 se
**		forem diferentes.
**
**		. conch(prim, ult, chaves, nch, area): escolhe o arquivo 
**		para a concatenacao horizontal. Caso nao defina o arq. mestre
**		nao realiza a concatenacao. Caso contrario mosta a observacao,
**		colocando na pilha o no. da obs. corrente.
**
**		. ConcHor(prim, ult, area): realiza a concatenacao horizontal
**		caso nao exista chave de ordenacao.
**
**		. ChSLe(i): le uma obs. no arquivo de indice i, atribuindo o 
**		seu conteudo `as vars. do arquivo. Retorna o no. da obs lida.
**
**		. ChaveEscolheMestre(prim, ult, chaves, nch, area): escolhe
**		o arq. mestre da concatenacao.
**	
**		. ChChave(prim, utl, chaves, nch, area): realiza a concatenacao
**		horizontal com chave de ordenacao.
**
**		. MontaObs(i, prim, ult, chaves, nch, area, nret): monta a 
**		observacao resultante na concatenacao horizontal.
**
**		. PoemMissing(prim, ult): coloca valor missing em todas as vars.
**		dos arquivos com indices entre prim e ult inclusive. Estes 
**		valores sao atualizados na tabela de variaveis
**
**
** Documentado por:
**
**		Carla Geovana do N. Macario
**
**
** Data da ultima alteracao:
**
**		97/01/28 10:56:09
**
** ----------------------------------------------------------------------------
*/

/* ---------------------------------------------
        Para usar FTELL e FSEEK :
	
	#include <stdio.h> 
	int fseek(stream,offset,origin);
	FILE *stream;
	long offset;
	int  origin;   

	long ftell(stream);
	FILE *stream;

   ----------------------------------------------- */


extern struct variavel variavel[];
extern struct arquivos arquivos[];
extern double pilha[];
extern int idpilha;
extern int fim;
extern int pareeof;
extern char *obs[OBSSIZE];

extern int warnings;


/* ---------------- GLOBAIS --------------------------------------- */

static char ChAtu[MAXCHAVES][TAMCAMPO];
static char ChAnt[MAXCHAVES][TAMCAMPO];
static char ChMestre[MAXCHAVES][TAMCAMPO];

/* --------------------- Rotina :   CONCV -------------------------- */

void concv(prim,ult,chaves,nch,area)
int prim,ult;		/* arquivos inicial e final */
int *chaves;		/* indices das variaveis que formam a chave */
int nch;		/* numero de variaveis que formam a chave */
struct WorkArea *area;  /* apontador para a area de trabalho */
{
	long nobs;
	int  i;
	
	/*
	** Escolhe um arquivo mestre para concatencao. Caso seja um valor
	** nulo nao realiza a concatenacao e termina a execucao do programa
	*/

	if (area->mestre == 0 ) {
		CvEscolheMestre(&(area->mestre), area->chave, nch, chaves,
			prim, ult, area->tip_ord);
		area->obscorr = 0L;

		/* Nao encontrou arquivo mestre  */

		if ( area->mestre == -1 ) {
			if (warnings) {
				fprintf(stderr, "\n\n\t\tAdvertencia : \n");
				fprintf(stderr, "\t\tArquivos vazios - CONCVER \n\n");
				Pausa_de_Advertencia ();
			}
			PoemMissing(prim, ult);
			goto final;
		}
	}

	/* 
	** Atribui valor missing para as variaveis dos arquivos da concatencao
	*/

	PoemMissing(prim, ult);

	/*
	** Existe um arquivo mestre
	*/

	if (area->mestre != -1) {

		/*
		** Obtem a posicao atual do apontador do arq. mestre e le a
		** observacao corrente, atribuindo o seu conteudo para
		** as vars. da chave mestre. Reposiciona o apontador do arq.
		** na posicao anterior `a leitura da obs. Rele a obs., atribui-
		** seu conteudo `as vars. do arq.
		*/

		area->apmestre = Curr_OffSet(arquivos[area->mestre].arq.ptraa);
		nobs = CvLeCopiaChave(area->mestre, ChMestre, chaves, nch);
		nobs = ChSLe(area->mestre);

		/*
		** Se foi detectado EOF no mestre ou o conteudo de alguma var. 
		** de ChMestre e' diferente das vars. que compoem a chave na 
		** area de trabalho
		*/

		if ((nobs == -1L) ||
			CvCompara(ChMestre, area->chave, nch, chaves)) {

			/* 
			** Reposiciona o apontador do arq. mestre e verifica
			** se o mestre deve ser trocado
			*/

			Set_OffSet(arquivos[area->mestre].arq.ptraa,
				area->apmestre);
			CvEscolheMestre(&(area->mestre), area->chave, nch,
				chaves, prim, ult, area->tip_ord);

			if ( area->mestre != -1 ) {

				/*
				** Atualiza o apontador para o arq. mestre, que
				** agora pode ser outro. Le uma obs. neste arq. 
				** e atribuindo o seu conteudo para as vars. de
				** ChMestre. Reposiciona o apontador do arq.
				** Rele a obs., atribuindo seu conteudo `as  
				** vars. do arq.
				*/

				area->apmestre =
				Curr_OffSet(arquivos[area->mestre].arq.ptraa);
				nobs = CvLeCopiaChave(area->mestre, ChMestre,
					chaves, nch);
				nobs = ChSLe(area->mestre);
			}
		}
	}
	
	/*
	** Verifica se nao existe arquivo mestre para a concatenacao, atribuindo
	** neste caso, EOF para os arquivos e verificando se todos os EOF 
	** foram atingidos para terminar a execucao do programa.
	** Caso contrario coloca na pilha o no. da observacao corrente da 
	** area de trabalho
	*/

final: 	
	if (area->mestre == -1 ) {
		for ( i = prim ; i <=ult ; i++ )
			arquivos[i].obs = (long) EOF;
		pilha[idpilha++] = (double) -1L;
		if ( pareeof ) 	if (veriarq()) 	fim = 0;
	}
	else			 pilha[idpilha++] = (double)(++area->obscorr);

} 

/* --------- CvEscolheMestre ---------------------------------------------- */

void   	CvEscolheMestre(mestre, chave, nch, chaves, prim, ult, tip_ord)
int	*mestre;
char	chave[MAXCHAVES][TAMCAMPO];
int    	nch;
int    	*chaves;
int    	prim, ult;
char	tip_ord;
{
	int    i,j,m,n;
	int    atumestre;
	long   nobs,nchave;
	double a1,a2;
	int    qual=0;
	int    igual=1;

	nchave = -1L;

	/*
	** Le uma obs. no arq. e atualiza os valores das vars. de ChAnt 
	** com os valores da nova observacao lida. Reposiciona o apontador
	** do arq. para sua posicao antes da leitura da obs. e retorna o no. 
	** da obs. corrente. Faz isso para todos os arqs. de indices entre 
	** prim e ult., enquanto nao for encontrado um EOF
	*/

	for (i=prim ; (i<=ult) && (nchave==-1L) ; i++) {
		atumestre = i;
		nchave = CvLeCopiaChave(i, ChAnt, chaves, nch);
	}

	/* 
	** Se existe chave  e nao foi detectado EOF. Verifica se um arq. dos 
	** que restaram sera' o arq. mestre segundo o tipo de ordenacao
	*/

	if ( nch  && (nchave != -1L) ) {

		for (i= (atumestre+1) ; i<=ult ; i++) {

			/*
			** Le uma nova observacao no arquivo e atualiza os 
			** valores das vars. de ChAtu com os valores da nova 
			** observacao lida. Reposiciona o ponteiro do arq.
			** Retorna o no. da obs. corrente
			*/

			nobs = CvLeCopiaChave(i, ChAtu, chaves, nch);

			/* 
			** Nao e' EOF. Verifica qual vai ser o arq. mestre 
			** segundo o tipo de ordenacao
			*/

			if ( nobs != -1L ) {
				j = 0;

				/* 
				** Compara o conteudo das vars. de ChAnt
				** com as vars. de ChAtu, verificando sua
				** precedencia segundo o tipo de ordenacao.
				** Faz isso enquanto seus valores forem iguais 
				*/

				while ( igual && (j<nch) ) {

					m = chaves[j];

					/* variavel string */

					if ( variavel[m].tipo == 'c' ) {
						n = strcmpi(ChAnt[j], ChAtu[j]);
						if ( n != 0 )  igual = 0;
						if (tip_ord == 'a' && n > 0)
							qual = 2;
						else if (tip_ord=='d' && n<0)
							qual =2;
					}

					/* variavel numerica */

					else {
						if (strcmpi(ChAnt[j], ".")==0) 
							a1 = MISSING;
						else 	a1 = atof(ChAnt[j]);
						if (strcmpi(ChAtu[j], ".")==0) 
							a2 = MISSING;
						else 	a2 = atof(ChAtu[j]);
						if (a1 != a2)
							igual = 0;
						if (tip_ord == 'a' && a1 > a2)
							qual = 2;
						else if (tip_ord=='d' && a1<a2)
							qual =2;
					}
					++j;
				}
				igual = 1;

				/* 
				** Caso os valores de ChAnt e ChAtu estejam 
				** trocados na ordenacao, copia os valores de
				** ChAtu em ChAnt e o mestr passa a ser o arq.
				** corrente
				*/

				if ( qual == 2 ) {
					for (j=0;j<nch;j++) 
						strcpy(ChAnt[j], ChAtu[j]);
					atumestre = i;
					nchave = nobs;
					qual = 0;
				}
			}
		}
	}

	/*
	** Se encontrou EOF no arquivo mestre atribui valor missing as vars. 
	** que compoem a chave
	*/

	if ( nchave == -1L ){
	 	*mestre = -1;
		for ( j=0 ; j < nch ; j++ ) {
			if ( variavel[chaves[j]].tipo == 'c' )
				strcpy(chave[j], ""); 
			else
				strcpy(chave[j], "."); 
		}
	}

	/*
	** Nao e' EOF. Compara o conteudo das vars. da chave do arquivo mestre
	** com o conteudo das vars. em chave. Verifica se sao iguais ou se 
	** estao na ordem correta de ordenacao
	*/

	else {
		if (*mestre != 0) {
			igual = 1;
			for (j = 0; j < nch && igual; j++) {
				
				m = chaves[j];
				if ( variavel[m].tipo == 'c' ) {
					if ((n=strcmpi(chave[j], ChAnt[j]))!=0) 
						igual =0;
						break;
				}
				else {
					if (strcmpi(chave[j], ".")==0) 
						a1 = MISSING;
					else 	
						a1 = atof(chave[j]);
					if (strcmpi(ChAnt[j], ".")==0) 
						a2 = MISSING;
					else 	
						a2 = atof(ChAnt[j]);
					if (a1 != a2)  {
						if (a1 > a2)
							n = 1;
						else
							n = -1;
						igual = 0;
						break;
					}
				}
			}

			/* 
			** O conteudo das vars. nao e' igual. Verifica se 
			** existe erro na ordenacao
			*/

			if ( igual == 0 ) {
				if ( (tip_ord=='a') && (n>0) ) 
					ChErro();
				else
					if (tip_ord == 'd' && n < 0)
						ChErro();
			}
		}
		*mestre = atumestre;
		for ( j=0 ; j < nch ; j++ ) 
			strcpy(chave[j], ChAnt[j]);
	}
}

void	ChErro()
{
	printf("\n\n\t\tAdvertencia \n");
	printf("\t\t\t O(s) arquivo(s) nao esta(o) ordenado(s) pela chave\n");
	printf("\t\t\t Use o modulo ORDENA para obter ordenacao apropriada\n\n");
	Pausa_de_Advertencia ();
	termina(2);
}

/* ------------------------- CvLeCopiaChave -------------------- */

long CvLeCopiaChave(qual, ChAtual, chaves, nch)
int  qual;
char ChAtual[MAXCHAVES][TAMCAMPO];
int  *chaves;
int  nch;
{
	long guarda,nobs;
	int  i, j, k, m, nv;

	/*
	** Obtem a posicao atual do apontador do arquivo.
	** Le a observacao corrente, retornando seu numero.
	** Reposiciona o apontador do arquivo para sua posicao antes da
	** leitura da observacao. Obtem o no. de vars. do arq.
	*/

	guarda = Curr_OffSet(arquivos[qual].arq.ptraa);
	nobs   = ler_obs(arquivos[qual].arq.ptraa,obs);
	k      = Set_OffSet(arquivos[qual].arq.ptraa, guarda);
	nv     = arquivos[qual].nvar;

	k      = 0;

	/* 
	** Faz para o loop nch vezes, que corresponde ao no. de vars. que
	** compoem a chave
	*/

	for ( j=0 ; j<nch ; j++ ) {

		for ( i=0 ; i<nv ; i++ ) {
			
			/*
			** Obtem o indice da var. na tabela de vars. e verifica
			** se ela faz parte da chave
			*/

			m = arquivos[qual].pos[i];
			if (chaves[j] == m) {

				/*
				** Se for EOF coloca valor missing
				** na chave atual. Caso contrario, copia o 
				** conteudo da obs. para ChAtual
				*/

				if ( nobs == (long) EOF ) {
					if (variavel[m].tipo == 'c')
						strcpy(ChAtual[k++], "");
					else
						strcpy(ChAtual[k++], ".");
				}
				else	
					strcpy(ChAtual[k++], obs[i]);
				break;
			}
		}
	}

	/*
	** Atribui valor missing as vars. do arquivo e retorna o no. da obs.
	*/

	PoemMissing(qual, qual);
	return(nobs);
}

/* ----------------------- CvCompara -------------------------- */

int 	CvCompara(mestre, lida, nch, chaves)
char 	mestre[MAXCHAVES][TAMCAMPO],
	lida[MAXCHAVES][TAMCAMPO];
int	nch;
int	*chaves;
{
	int 	i,m;
	double	a1,a2;

	/*
	** Faz um loop para todas as vars. que compoem a chave
	*/

	for ( i=0 ; i < nch ; i++ ) {
		
		/*
		** Obtem o indice da var. na tabela de vars.
		*/

		m = chaves[i];

		/*
		** variavel string: compara o conteudo da variavel do 
		** arquivo mestre com a var. do arquivo lida, retornando
		** 1 caso sejam diferentes
		*/

		if ( variavel[m].tipo == 'c' ) {
			if ( strcmpi(mestre[i], lida[i]) != 0 )
					return(1);
		}

		/* 
		** variavel numerica: verifica o conteudo das vars. dos
		** arqs. mestre e lida e retorna 1 caso seus valores sejam
		** diferentes
		*/

		else {
			if ( strcmpi(mestre[i], ".") == 0) {
				a1 = MISSING;
			}
			else {
				a1 = atof(mestre[i]);
			}
			if ( strcmpi(lida[i], ".") == 0) {
				a2 = MISSING;
			}
			else {
				a2 = atof(lida[i]);
			}
			if ( a1 != a2 )
				return(1);
		}
	}
	return(0);
}

/* ----------- ROTINA conch -------------------------------------- */

void 	conch(prim, ult, chaves, nch, area)
int	prim, ult;
int	*chaves;
int	nch;
struct 	Ch_Area	*area;
{
	int i;

	/*
	** Verifica se existe chave de ordenacao
	*/

	if ( nch == 0 )
		ConcHor(prim, ult, area);
	else
		ChChave(prim, ult, chaves, nch, area);

	/*
	** Escolhe um arquivo mestre para concatencao. Caso seja um valor
	** nulo nao realiza a concatenacao e termina a execucao do programa.
	** Caso contrario coloca na pilha o no. da observacao corrente
	*/

	if ( area->mestre == -1 ) {
		for ( i = prim ; i <=ult ; i++ )
			arquivos[i].obs = (long) EOF;
		pilha[idpilha++] = (double) -1L;
		if (pareeof && veriarq())
			fim = 0;
	}
	else
		pilha[idpilha++] = (double) (++area->obscorr);

}

/* ----------- ROTINA ConcHor -------------------------------------- */

void	ConcHor(prim, ult, area)
int	prim, ult;
struct  Ch_Area	*area;
{
	long nobs,guarda;
	unsigned ubytes;
	int  i, k, j, m, l, l1, l2, l3, pert;
	char *qp;
#ifdef EDIX
	double *dp;
#endif


	/*
	** Verifica se existe arq. mestre definido
	*/

	if (area->mestre== 0) {
		
		for (i = prim ; i <= ult ; i++) {

			/*
			** Obtem a posicao atual do apontador do arquivo.
			** Le a observacao corrente do arquivo.
			** Reposiciona o apontador na observacao corrente 
			*/

			guarda = Curr_OffSet(arquivos[i].arq.ptraa);
			nobs   = ler_obs(arquivos[i].arq.ptraa, obs);
			k      = Set_OffSet(arquivos[i].arq.ptraa, guarda);

			/* 
			** verifica se e' EOF. Caso nao seja, o
			** arquivo corrente passa a ser o mestre
			*/

			if ( nobs != -1L ) {
				area->mestre = i;
				break;
			}
		}

		/*
		** Verifica se foi definido o arq. mestre, lendo as observacoes
		** dos outros arquivos
		*/

		if ( area->mestre != 0 ) {
			area->numLista = 0;
			area->obscorr  = 0L;
			for ( i = area->mestre; i <= ult ; i++) {

				/*
				** Obtem a posicao atual do apontador do arq.
				** Le a observacao corrente do arquivo.
				** Reposiciona o apontador na obs. corrente 
				** Verifica se o arquivo nao atingiu EOF. Neste
				** caso, coloca seu indice na lista de arquivos
				** que ainda restam observacoes para concatenar
				*/

				guarda = Curr_OffSet(arquivos[i].arq.ptraa);
				nobs   = ler_obs(arquivos[i].arq.ptraa, obs);
				k      = Set_OffSet(arquivos[i].arq.ptraa,
						guarda);

				if ( nobs != -1L ) {
					area->Lista[area->numLista++] = i;
				}
			}
		}

		/* 
		** Nao conseguiu definir o arq. mestre. Os arquivos para 
		** concatenacao estao vazios. Nao realiza a concatenacao
		*/

		else {
			if (warnings) {
				fprintf(stderr, "\n\n\t\tAdvertencia : \n\n");
				fprintf(stderr, "\t\t\tArquivos vazios - conchor \n\n");
				Pausa_de_Advertencia ();
			}

			area->mestre = -1;
			PoemMissing(prim, ult);
			goto final;
		}
	}
	pert = 0;

	for ( i= prim ; i <= ult ; i++ ) {

		/*
		** Verifica se o arquivo esta na lista de arquivos 
		** existentes para a concat. Caso nao esteja, atribui valor
		** missing `as suas vars.
		*/

		for ( j=0 ; j < area->numLista ; j++ ) {
			if ( i == area->Lista[j] ) {
				pert = 1;
				break;
			}
		}
		if ( pert == 0 )
			PoemMissing(i , i);
		else
			pert = 0;
	}
	
	/* 
	** Verifica se foi definido arq. mestre, lendo as obs. dos arquivos
	** que restam
	*/

	if (area->mestre != -1) {
		for ( i = 0 ; i < area->numLista ; i ++ ) {
			nobs = ler_obs(arquivos[area->Lista[i]].arq.ptraa, obs);

			/*
			** Seta  o no. da observacao corrente e obtem o no. de
			** vars. do arquivo
			*/

			arquivos[area->Lista[i]].obs = nobs;
			k = arquivos[area->Lista[i]].nvar;


			for ( j=0; j<k ; j++ ) {
				pert = 0;

				/*
				** Obtem o indice da var.
				** Verifica se foi atingido EOF do arq. Neste
				** caso, procura nos arqs. anteriores a 
				** mesma variavel
				*/

				m = arquivos[area->Lista[i]].pos[j];

				if ( nobs == (long) EOF ) {
					for (l=0 ; l<i ; l++ ) {
						l1 = arquivos[area->Lista[l]].nvar;
						for ( l2=0 ; l2<l1 ; l2++) {
							l3 = arquivos[area->Lista[l]].pos[l2];
							if ( l3 == m) {
								pert = 1;
								break;
							}
						}
						if ( pert ) break;
					}
				}

				/*
				** Nao encontrou a variavel em outros arqs. 
				** ou nao foi EOF, copia o conteudo da obs.
				** para a variavel com indice m
				*/

				if ( !pert ) {

					/* variavel string  */

					if (variavel[m].tipo == 'c') {
						free(variavel[m].ptrval.ptrs);
						variavel[m].ptrval.ptrs = (char *) NULL;
						
						ubytes =
						(unsigned) (strlen(obs[j]) + 1);
						qp = (char *) malloc(ubytes);
						if (qp == (char *) NULL) {

/* *** Padroniz. msg erro - Adriana 09/01/95
							execerror(
							"Falta memoria", "");
*** */

							execerror(
							PegaMsg (1039), "");
							termina(2);
						}
						if (nobs == (long) EOF)
							*qp = '\0';
						else
							strcpy(qp, obs[j]);
						variavel[m].ptrval.ptrs = qp;
					}

					/* variavel missing  */

					else {
						if (nobs == (long) EOF ||
							strcmpi(obs[j], ".") == 0) {
#ifdef EDIX
							dp = &(variavel[m].ptrval.ptrf);
							*dp = MISSING;
#else
							variavel[m].ptrval.ptrf = MISSING;
#endif
						}

						/* variavel numerica  */

						else {
#ifdef EDIX
							dp = &(variavel[m].ptrval.ptrf);
							*dp = atof(obs[j]);
#else
							variavel[m].ptrval.ptrf = atof(obs[j]);
#endif
						}
					}
				}
			}

			/*
			** Apos percorrer todas as vars. do arq., verifica 
			** se e' EOF. Retira este arquivo da lista de
			** arquivos restantes para concat.
			*/

			if ( nobs == (long) EOF ) {
				l1 = 0;
				for ( l=0 ; l < area->numLista ; l++) 
					if ( l != i )
						area->Lista[l1++] = area->Lista[l];
				area->numLista = l1;
				i--;
			}
		}

		/*
		** Verifica se ainda existem arquivos para concatencao
		*/

		if ( area->numLista == 0 ) area->mestre = -1;
	}
final: 	;
}

/* ------------ ROTINA ChSLe ----------------------------------------- */

long 	ChSLe(i)
int	i;
{
	long nobs;
	unsigned ubytes;
	int  k,j,m;
	char *qp;
#ifdef EDIX
	double *dp;
#endif

	/*
	** Le a observacao corrente do arquivo, retornando seu no.
	*/

	nobs = ler_obs(arquivos[i].arq.ptraa,obs);
	arquivos[i].obs = nobs;

	/*
	** Obtem o no. de vars. do arquivo, fazendo um loop para todas elas
	*/

	k = arquivos[i].nvar;
	for (j=0; j<k; ++j) {
		
		/*
		** Obtem o indice da cada var. na tabela de vars., atribuindo
		** para elas o seu valor correspondente na observacao. Se for
		** final de arquivo o valor atribuido e' missing
		*/

		m = arquivos[i].pos[j];

		/* Variavel string: desaloca a area de memoria ocupada e aloca
		   uma nova area */

		if (variavel[m].tipo == 'c') {
			free(variavel[m].ptrval.ptrs);
			variavel[m].ptrval.ptrs = (char *) NULL;
			if (nobs == (long) EOF) {
				if ((qp = (char *) malloc(1)) == (char *) NULL) {
					execerror( PegaMsg (1039), "");
					termina(2);
				}
				*qp = '\0';
			}
			else {
				ubytes = (unsigned) (strlen(obs[j])+1);
				if ((qp = (char *) malloc(ubytes)) == (char *) NULL) {

/* *** Padroniz. msg erro - Adriana 09/01/95
					execerror("Falta memoria", "");
*** */

					execerror( PegaMsg (1039), "");
					termina(2);
				}
				strcpy(qp, obs[j]);
			}
			variavel[m].ptrval.ptrs = qp;
		}

		/* variavel numerica */

		else {
			if (nobs == (long) EOF || strcmpi(obs[j], ".") == 0) {
#ifdef EDIX
				dp = &(variavel[m].ptrval.ptrf);
				*dp = MISSING;
#else
				variavel[m].ptrval.ptrf = MISSING;
#endif
			}
			else {
#ifdef EDIX
				dp = &(variavel[m].ptrval.ptrf);
				*dp = atof(obs[j]);
#else
				variavel[m].ptrval.ptrf = atof(obs[j]);
#endif
			}
		}
	}
	return(nobs);
}


/* --------- ChaveEscolheMestre --------------------------------------- */

void   ChaveEscolheMestre(prim,ult,chaves,nch,area)
int    prim,ult;
int    *chaves;
int    nch;
struct Ch_Area *area;
{
	int    i,j;

	/*
	** Escolhe o arquivo mestre da concatenacao
	*/

	CvEscolheMestre(&(area->mestre),area->chave,nch,chaves,prim,ult,area->tip_ord);

	/*
	** Verifica se conseguiu definir o mestre
	*/

	if (area->mestre != -1) {
		i = 0;

		/*
		** Coloca o indice do arq. mestre na lista de indices de arqs. 
		** a serem usados na concatencao. Obtem o endereco do ponteiro
		** deste arquivo e coloca estas informacoes na area de trabalho
		*/

		area->Lista[i++] = area->mestre;
		area->apmestre = Curr_OffSet(arquivos[area->mestre].arq.ptraa);
		area->ARQS[area->mestre].apnivel = Curr_OffSet(arquivos[area->mestre].arq.ptraa);

		/*
		** Faz um loop para cada um dos arqs. restantes
		*/

		for (j=((area->mestre)+1); j<=ult ; j++) {

			/* 
			** Le a observacao do arq. colocando o conteudo das
			** vars. da obs. lida nas vars. de ChAnt. Reposiciona 
			** o apontador do arquivo.
			*/

			CvLeCopiaChave(j,ChAnt,chaves,nch);

			/*
			** Compara o conteudo de cada var. de ChAnt e da
			** chave da area de trabalho, retornado 1 se alguma 
			** delas for diferentes. Caso sejam todas iguais, 
			** coloca o indice do arq. na lista de arqs. para 
			** concat. e atualiza apnivel com o apontador 
			** deste arquivo
			*/

			if (CvCompara(ChAnt,area->chave,nch,chaves)== 0) {
				area->Lista[i++] = j; 
				area->ARQS[j].apnivel = Curr_OffSet(arquivos[j].arq.ptraa);
			}
		}

		/*
		** Seta o no. de arqs. que compoem a concatencao
		*/

		area->numLista = i;
	}

	/*
	** Nao encontrou um arq. mestre. logo nao tem arqs. a serem concate-
	** nados
	*/

	else	area->numLista = 0;
}

/* --------------------- Rotina ChChave --------------------------------- */

void 	ChChave(prim,ult,chaves,nch,area)
int	prim,ult;
int	*chaves;
int	nch;
struct	Ch_Area *area;
{
	long	guarda;
	int	i,j;
	int 	nret,pert; 

	/*
	** Verifica se ja' foi escolhido o arq. mestre . Caso isto ainda
	** nao tenha acontecido, tenta escolher um, verificando se o arq.
	** foi definido
	*/

	if ( area->mestre == 0 ) {
		area->obscorr = 0L;
		ChaveEscolheMestre(prim,ult,chaves,nch,area);

		/*
		** Nao conseguiu escolher o arq. mestre. Logo nao existem
		** arqs. a serem concatenados
		*/

		if (area->mestre == -1) {
			if (warnings) {
				fprintf(stderr,"\n\n\t\tAdvertencia : \n\n");
				fprintf(stderr,"\t\t\tArquivos vazios - CONCHOR \n\n");
				Pausa_de_Advertencia ();
			}

			PoemMissing(prim,ult);
			goto final;
		}
	}
	pert = 0;

	/*
	** Faz um loop para todos os arqs. da concatenacao para verificar
	** os que fazem parte da lista de arqs. que ainda tem obs. para
	** concatenar. Os que nao tiverem mais obs. terao valor missing em 
	** suas vars.
	*/

	for ( i=prim ; i<=ult ; i++ ) {
		for ( j=0 ; j < (area->numLista - 1) ; j++) {
			if ( i == area->Lista[j] ) {
				pert = 1;
				break;
			}
		}
		if ( !pert ) 	PoemMissing(i,i);
		else		pert = 0;
	}

	/*
	** Verifica se  arq. mestre foi definido. Neste caso, le as 
	** observacoes dos proximos arqs. e monta a obs. do arq. resultante
	*/

	if (area->mestre != -1) {
		for ( i=0 ; i < (area->numLista -1) ; i++ ) {
			guarda = Curr_OffSet(arquivos[area->Lista[i]].arq.ptraa);
			/*
			** Le um obs. e atribui o conteudo das vars. da obs. 
			** lida `as vars. do arq.. Feito isso, reposiciona o 
			** ponteiro do arq.
			*/

			ChSLe(area->Lista[i]);
			Set_OffSet(arquivos[area->Lista[i]].arq.ptraa, guarda);
		}

		/* no. de arqs. restante para a concatenacao  */

		nret = area->numLista - 1;

		MontaObs(area->numLista-1, &prim, &ult, chaves, &nch, area, &nret);
	}
final: 	;
}

/* --------------- Rotina : MontaObs --------------------------------- */

void	MontaObs(i,prim,ult,chaves,nch,area,nret)
int	i;
int	*prim;
int	*ult;
int	*chaves;
int	*nch;
struct 	Ch_Area *area;
int	*nret;
{
	long	guarda;
	long	nobs;
	int	j;

	/*
	** Le uma observacao do arquivo a ser concatenado, colocando 
	** o conteudo das vars. da obs lida na vars. de ChAtu. Reposiciona o
	** apontador do arquivo
	*/

	nobs   = CvLeCopiaChave(area->Lista[i],ChAtu,chaves,*nch);

	/*
	** obtem a posicao atual do apontador do arquivo. Rele a obs, 
	** atribuindo seu conteudo `as vars. do arq.
	*/

	guarda = Curr_OffSet(arquivos[area->Lista[i]].arq.ptraa);
	nobs   = ChSLe(area->Lista[i]);

	/* 
	** Se for EOF ou o conteudo de alguma var. de ChAtu for diferente do
	** conteudo da  var da area de trabalho
	*/

	if ((nobs == -1L)||(CvCompara(ChAtu,area->chave,*nch,chaves))){

		/*
		** Ainda tem arq. para a concatenacao
		*/

		if (i != 0) {
			
			/*
			** Faz o apontador da area de trabalho que contem o end.
			** do prox. arq. ser o end. deste arq., reposicionando o
			** seu ponteiro na posicao antes de ler a obs. Le uma 
			** obs. no arq. anterior a este e monta a obs. do arq.
			** final
			*/

			area->ARQS[area->Lista[i]].approx = guarda;
			Set_OffSet(arquivos[area->Lista[i]].arq.ptraa,
				area->ARQS[area->Lista[i]].apnivel);
			nobs = ler_obs(arquivos[area->Lista[i-1]].arq.ptraa,obs);
			MontaObs(i-1,prim,ult,chaves,nch,area,nret);

			/*
			** Verifica se tem arq. mestre e ainda existem arqs. 
			** para a concat. Caso existam arqs. 
			** manda ler uma obs. do arq. e seta as vars.
			** do arq. com os valores da obs.
			*/

			if ((area->mestre != -1) && (*nret > 0)){
				
				/* tem mais de um arq. para a concatenacao  */

				if ( *nret > 1 ) {
					guarda = Curr_OffSet(arquivos[area->Lista[i]].arq.ptraa);
					nobs = ChSLe(area->Lista[i]);
					Set_OffSet(arquivos[area->Lista[i]].arq.ptraa,guarda);
				}
				else 	nobs = ChSLe(area->Lista[i]);

				/* decrementa o no. de arqs. restantes  */

				-- *nret;
			}
		}

		/*  arquivo mestre  */

		else {	

			/*
			** Faz um encadeamento dos arqs. da concatenacao
			*/

			area->ARQS[area->Lista[0]].approx = guarda;
			for (j=0;j<area->numLista;j++)
				Set_OffSet(arquivos[area->Lista[j]].arq.ptraa,
					area->ARQS[area->Lista[j]].approx);

			/*
			** Manda escolher o mestre e verifica se existe o mestre
			** Manda ler a obs.
			*/

			ChaveEscolheMestre(*prim,*ult,chaves,*nch,area);
			if (area->mestre != -1) {
				if ( area->numLista > 1) {
					guarda = Curr_OffSet(arquivos[area->Lista[0]].arq.ptraa);
					nobs = ChSLe(area->Lista[0]);
					Set_OffSet(arquivos[area->Lista[0]].arq.ptraa,guarda);
				}
				else 
					nobs = ChSLe(area->Lista[0]);

				if ( *nret > (area->numLista -1)) 
					*nret = area->numLista - 1;
				else {
					if ( *nret < (area->numLista-1) ) {
						for ( i=1 ; i < (area->numLista-1) ; i++) {
							guarda = Curr_OffSet(arquivos[area->Lista[i]].arq.ptraa);
							nobs = ChSLe(area->Lista[i]);
							Set_OffSet(arquivos[area->Lista[i]].arq.ptraa,guarda);
						}
						nobs = ChSLe(area->Lista[area->numLista-1]);

						*nret = 0;
					}
				}
			}
		}
	}

	/* 
	** e' EOF e o conteudo de alguma var. de ChAtu e' diferente do
	** conteudo de alguma var da area de trabalho
	*/

	else {
		if ( i != (area->numLista - 1) )
			Set_OffSet(arquivos[area->Lista[i]].arq.ptraa,guarda);
	}
}

/* ---------------------- PoemMissing --------------------------------- */

void	PoemMissing(prim,ult)
int	prim;
int	ult;
{
	int i,j,nv,m;
	char *qp;
#ifdef EDIX
	double *dp;
#endif

	/* 
	** Pega todos os arquivos desde prim ate' ult
	*/

	for ( i = prim ; i <= ult ; i++ ) {

		/* 
		** Pega o numero de vars. do arquivo, para atribuir valor 
		** missing para todas elas 
		*/

		nv = arquivos[i].nvar;
		for (j=0 ; j < nv ; j++) {

			/*
			** Pega o indice da variavel do arquivo na tabela de 
			** vars.
			*/

			m = arquivos[i].pos[j];

			/*
			** variavel string. Libera a area de memoria ocupada
			** por ela
			*/

			if (variavel[m].tipo == 'c') {
				free(variavel[m].ptrval.ptrs);
				variavel[m].ptrval.ptrs = (char *) NULL;
				if ((qp = (char *) malloc(2)) == (char *)NULL) {

/* *** Padroniz. msg erro - Adriana 09/01/95
					execerror("Falta memoria", "");
*** */

					execerror( PegaMsg (1039), "");
					termina(2);
				}
				*(qp) = '\0';
				variavel[m].ptrval.ptrs = qp;
			}

			/*
			** variavel numerica. 
			*/

			else {
#ifdef EDIX
				dp = &(variavel[m].ptrval.ptrf);
				*dp = MISSING;
#else
				variavel[m].ptrval.ptrf = MISSING;
#endif
			}
		}
	}
}

