/***********************************************************************************
 Genese - Linguagem de programao para a gerao e manipulao de arquivo .dat

 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
 termos da Licena Pblica Geral GNU conforme publicada pela Free Software
 Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
 posterior.

 Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
 programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
 Temple Street, Suite 330, Boston, MA 02111-1307 USA.

 ***********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

	Funcao func que trata funcoes matematicas (seno, coseno, tangente, log, exp, etc.) e 
	algumas funcoes que manipulam datas

 Desenvolvido por:

      Oscar Ivan Palma Pacheco
      Carla Geovana do N. Macario
	Maria Fernanda Moura
	Adriana Delfino dos Santos
	

 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/

/* ----------------------------------------------------------------------------
**
**	
**
** Funcao que trata funcoes matematicas e algumas funcoes que manipulam datas
**
**
** Documentado por:
**
**		Carla Geovana do N. Macario
**		Oscar Ivan Palma Pacheco
**
**
** Data da ultima alteracao:
**
**		97/09/24 16:35:35
**
** ----------------------------------------------------------------------------
*/


/* ................................................................................*/
/* Nova versao para a funcao MONET.*/
/* ................................................................................*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>

#ifdef MSDOS
#include <io.h>
#endif

#ifndef VAXVMS
#include <malloc.h>
#endif

#ifdef SUN
extern time_t time();
#endif

#include "socasm.h"
#include "pcsgbd.h"
#include "genese.h"

#if defined(ULTRIX) | (defined(SUN) & !defined(RS6000)) | defined(VAXVMS)

#define	MAXRAND		2147483647.

#else

#if defined(RS6000) | defined(MSDOS) | defined(EDIX)

#define	MAXRAND		37267.

#else

#if defined(SGI)

#define	MAXRAND		RAND_MAX

#endif
#endif
#endif

#ifdef MEM_DEBUG
#include "mshell.h"
#endif

/* ................................................................................*/
/* Esta era a versao 3.12, que estava funcionando ok........................*/
/* ................................................................................*/

#ifdef DEBUG
extern int	f_1;
#endif

extern	unsigned int	ins[];	/* vetor de instrucoes a serem executadas  */
extern	double	pilha[];	/* pilha de dados  */
extern	int	idpilha;	/* indice da pilha de dados  */
extern	char	*spilha[];	/* pilha de dados strings */
extern	int	idspilha;	/* indice da pilha de strings  */
extern unsigned int inst;	/* instrucao do vetor de instrucoes a ser 
				   executada  */
extern	double	randseed;

extern	double	op1, op2, op3;
extern	char	*ops1, *ops2, *ops3;
extern	char	buffer[];

/*
** Tabela com o total de dias no ano apos cada mes.
*/

static double N_DIA_MES[] = {0.0, 31.0, 59.0, 90.0, 120.0, 151.0, 181.0, 212.0,
			     243.0, 273.0, 304.0, 334.0};

#if defined(MSDOS) | defined(SCO)
# define RND_T		unsigned
#else
# define RND_T		int
#endif

void func()
{
	int i;
	RND_T irand;
	time_t hora;
	struct tm *dthr;
	double nday;
	double temp;
	double temp1;

	/*
	** Pega o parametro da funcao a ser executada.
	*/

	i = OPER1;

	if (i != ODATE && i != OTIME && i != OSTRLEN && i != OSCAN && 
		i != OEXTARQ)
		op2 = pilha[--idpilha];

	switch (i) {

	/*
	** Data no formato aaaammdd.
	*/

	case ODATE:

#ifdef DEBUG
		D1print("DATE\n");
#endif

		time(&hora);
		dthr = localtime(&hora);

		/*
		** Coloca a data na pilha de dados.
		*/

		pilha[idpilha++] = (double) ((1900 + dthr -> tm_year) * 10000L +
			(dthr -> tm_mon + 1) * 100L + dthr -> tm_mday);

		break;

	/*
	** Hora no formato hhmmss.
	*/

	case OTIME:

#ifdef DEBUG
		D1print("TIME\n");
#endif

		time(&hora);
		dthr = localtime(&hora);

		/*
		** Coloca a data na pilha de dados.
		*/

		pilha[idpilha++] = (double) (dthr -> tm_hour * 10000L +
			dthr -> tm_min * 100L + dthr -> tm_sec);

		break;

	/* 
	** Obtem o seno de op2, caso este nao seja um valor
	** missing, colocando o resultado na pilha de dados.
	*/

	case OSIN:

#ifdef DEBUG
		D1print("SIN\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = sin(op2);

		break;

	/* 
	** Obtem o cosseno de op2, caso este nao seja um valor
	** missing, colocando o resultado na pilha de dados.
	*/

	case OCOS:

#ifdef DEBUG
		D1print("COS\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = cos(op2);

		break;

	/* 
	** Obtem a tangente de op2, caso este nao seja um valor
	** missing, colocando o resultado na pilha de dados.
	*/

	case OTAN:

#ifdef DEBUG
		D1print("TAN\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = tan(op2);

		break;

	/* 
	** Obtem o arcoseno de op2, caso este nao seja um valor
	** missing, colocando o resultado na pilha de dados.
	*/

	case OASIN:

#ifdef DEBUG
		D1print("ASIN\n");
#endif

		if ( (lfuzz (op2, (double) -1.0) == -1) ||
		     (lfuzz (op2, (double) 1.0) == 1) )
		{
		  printf (PegaMsg (1161), "asin" );
		  termina (1);
		}

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = asin(op2);

		break;

	/* 
	** Obtem o arcocosseno de op2, caso este nao seja um valor
	** missing, colocando o resultado na pilha de dados.
	*/

	case OACOS:

#ifdef DEBUG
		D1print("ACOS\n");
#endif
		if ( (lfuzz (op2, (double) -1.0) == -1) ||
		     (lfuzz (op2, (double) 1.0) == 1) )
		{
		  printf (PegaMsg (1161), "acos" );
		  termina (1);
		}

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = acos(op2);

		break;

	/* 
	** Obtem o arcotangente de op2, caso este nao seja um valor
	** missing, colocando o resultado na pilha de dados.
	*/

	case OATAN:

#ifdef DEBUG
		D1print("ATAN\n");
#endif

/* Arco tangente vai de -infinito a +infinito

		if ( (lfuzz (op2, (double) -1.0) == -1) ||
		     (lfuzz (op2, (double) 1.0) == 1) )
		{
		  printf (PegaMsg (1161), "atan" );
		  termina (1);
		}
*/


		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = atan(op2);

		break;

	/* 
	** Obtem o exponencial de op2, caso este nao seja um valor
	** missing, colocando o resultado na pilha de dados.
	*/

	case OEXP:

#ifdef DEBUG
		D1print("EXP\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = exp(op2);

		break;

	/* 
	** Obtem o logaritmo natural de op2, caso este nao seja um 
	** valor missing, colocando o resultado na pilha de dados.
	*/

	case OLOG:

#ifdef DEBUG
		D1print("LOG\n");
#endif

		if (lfuzz (op2, (double) 0.0) <= 0)
		{
		  printf (PegaMsg (1162), "log" );
		  termina (1);
		}

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = log(op2);

		break;

	/* 
	** Obtem o logaritmo na base 10 de op2, caso este nao seja um 
	** valor missing, colocando o resultado na pilha de dados.
	*/

	case OLOG10:

#ifdef DEBUG
		D1print("LOG10\n");
#endif

		if (lfuzz (op2, (double) 0.0) <= 0)
		{
		  printf (PegaMsg (1162), "log10" );
		  termina (1);
		}

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = log10(op2);

		break;

	/* 
	** Obtem o valor absoluto de op2, caso este nao seja um 
	** valor missing, colocando o resultado na pilha de dados.
	*/

	case OABS:

#ifdef DEBUG
		D1print("ABS\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = fabs(op2);

		break;

	/* 
	** Obtem a raiz quadrada de op2, caso este nao seja um 
	** valor missing, colocando o resultado na pilha de dados.
	*/

	case OSQRT:

#ifdef DEBUG
		D1print("SQRT\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = sqrt(op2);

		break;

	/* 
	** Elimina os numeros decimais de  op2, caso este nao seja um 
	** valor missing, colocando o resultado na pilha de dados.
	*/

	case OINT:

#ifdef DEBUG
		D1print("INT\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = (double) ((long) op2);

		break;

	/* 
	** Indica o menor no. inteiro maior ou igual a op2, caso este 
	** nao seja um valor missing, colocando o resultado na 
	** pilha de dados.
	*/

	case OCEIL:

#ifdef DEBUG
		D1print("CEIL\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = ceil(op2);

		break;

	/* 
	** Indica o maior no. inteiro menor ou igual a op2, caso este 
	** nao seja um valor missing, colocando o resultado na 
	** pilha de dados.
	*/

	case OFLOOR:

#ifdef DEBUG
		D1print("FLOOR\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = floor(op2);

		break;

	/* 
	** Indica, entre dois numeros que nao sejam missing, 
	** qual e'  o maior.
	*/

	case OMAX:

#ifdef DEBUG
		D1print("MAX\n");
#endif

		op1 = pilha[--idpilha];

		if (vmissing(op1, op2))
			pilha[idpilha++] = (op1 >= op2)? op1 : op2;

		break;

	/* 
	** Indica, entre dois numeros que nao sejam missing, 
	** qual e'  o menor.
	*/

	case OMIN:

#ifdef DEBUG
		D1print("FLOOR\n");
#endif

		op1 = pilha[--idpilha];

		if (vmissing(op1, op2))
			pilha[idpilha++] = (op1 <= op2)? op1 : op2;

		break;

	/*
	** Alterar a semente do gerador de numeros aleatorios.
	*/

	case OSRAND:

#ifdef DEBUG
		D1print("SRAND");
#endif

		randseed = op2;

		if (vmissing((double) 0.0, op2)) {
			if (op2 < 0.0)
				op2 = -op2;

			if (op2 > 1.0) {
				i = (int) log10(op2) + 1;
				op2 = op2 / pow(10., (double) i);
			}

			irand = (RND_T) (op2 * MAXRAND);
		}

		else
			irand = (RND_T) (time(&hora) % (time_t) MAXRAND);

#ifdef DEBUG
		D1print("\t\tsemente: %ld\n", irand);
#endif

		srand(irand);

		break;

	/* 
	** Gera numeros aleatorios a partir de um no., desde que 
	** ele nao seja missing.
	*/

	case ORAND:

#ifdef DEBUG
		D1print("RAND");
#endif

		if (op2 != randseed && randseed == 1.0) {
			randseed = op2;

			if (vmissing((double) 0.0, op2)) {
				if (op2 < 0.0)
					op2 = -op2;

				if (op2 > 1.0) {
					i = (int) log10(op2) + 1;
					op2 = op2 / pow(10., (double) i);
				}

				irand = (RND_T) (op2 * MAXRAND);
			}

			else
				irand = (RND_T) (time(&hora)%(time_t)MAXRAND);

#ifdef DEBUG
			D1print("\t\tsemente: %ld", irand);
#endif

			srand(irand);
		}

#ifdef DEBUG
		D1print("\n");
#endif

		pilha[idpilha++] = ((double) rand()) / (MAXRAND + 1.);

		break;

	/*
	** Converte um numerico que nao seja missing para data.
	*/

	case OFTOD:

#ifdef DEBUG
		D1print("FTOD\t%s\n", ops2);
#endif

                if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = op2;

		break;

	/*
	** Converte uma data que nao seja missing para numerico.
	*/

	case ODTOF:

#ifdef DEBUG
		D1print("DTOF\t%s\n", ops2);
#endif

                if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = op2;

		break;

	/*
	** Converte um numero que nao seja missing para string.
	*/

	case OFTOA:

		if (op2 >= MISSING)
			buffer[0] = '\0';
		else 
			ftoa(op2, 15, buffer);

		ops1 = (char *) malloc((unsigned) (strlen(buffer)+1));

		if (ops1 == (char *) NULL)

/* Padroniz. msg erro - Adriana 09-01-95
			fprintf(stderr,"RAND: Memoria esgotada\n");
*** */
		  {
			fprintf(stderr, PegaMsg (1039) );
			termina (1);
		  }
			
		strcpy(ops1, buffer);

#ifdef DEBUG
		D1print("FTOA\t%s\n", ops1);
#endif

		spilha[idspilha++] = ops1;

		break;

	/*
	** Converte uma data que nao seja missing para string.
	*/

	case ODTOA:

		if (op2 >= MISSING)
			buffer[0] = '\0';
		else 
			dtoa(op2, buffer);

		ops1 = (char *) malloc((unsigned) (strlen(buffer)+1));

		if (ops1 == (char *) NULL)

/* Padroniz. msg erro - Adriana  09/01/95
			fprintf(stderr,"RAND: Memoria esgotada\n");
*** */
		  {
			fprintf(stderr, PegaMsg (1039));
			termina (1);
		  }
			
		strcpy(ops1, buffer);

#ifdef DEBUG
		D1print("DTOA\t%s\n", ops1);
#endif

		spilha[idspilha++] = ops1;

		break;


	/*
	** Converte um numero em string no formato monetario
	*/

	case OMNT: /* monet */
		{
		int 	resto  = 0;
		int	ponto  = 0;
		int	qbytes = 0;
		int	i,j;

		/* string final */
		ops1 = (char *) malloc((unsigned) (RDRBUF+16));

		/* converte valor para string */
		sprintf(buffer,"%lfRDRBUF:2",op2); 

		if (ops1 == (char *) NULL ) {
			fprintf(stderr, PegaMsg (1039) );
			termina (1);
		}

		if (op2 >= MISSING) 
			strcpy(ops1, "0,00");
		else  {
			if (op2 < 0 ) {
				qbytes=1;
				j=1;
				i=1;
				ops1[0]='-';
			}
			else {
				qbytes = 0;
				j=0;
				i=0;
			}
			/* verifica quantos bytes ainda existem ate' o ponto decimal*/
			for (; buffer[qbytes]!='.';++qbytes);
			qbytes = qbytes - j;	/* se houve sinal, retira-o aqui */
			/* verifica quantos digitos inteiros existem 
				antes do primeiro ponto para o formato monet */
			resto = (int) (qbytes % 3); 
			if (resto > 0 ) {
				while (resto > 0) {
					ops1[i]=buffer[j];
					--qbytes; --resto;
					++i; ++j;
				}
				if (qbytes > 0) {
					ops1[i]='.'; ++i;
				}
			}
			while (qbytes > 0 ) {
				for (ponto=0; ponto < 3; ponto++) {
					ops1[i]=buffer[j];
					++i; ++j; --qbytes;
				}
				if (qbytes > 0) {
					ops1[i]='.'; ++i;
				}
			}
			ops1[i]=',';
			ops1[i+1]=buffer[j+1]; /* pula o ponto decimal */
			ops1[i+2]=buffer[j+2]; /* segundo digito decimal */
			ops1[i+3]='\0';
		}
#ifdef DEBUG
		D1print("MNT\t%s\n", ops1);
#endif
		spilha[idspilha++] = ops1;
		break;
		}

	/*
	** Verifica se um arquivo "string" existe
	*/

	case OEXTARQ:
		ops2 = spilha[--idspilha];

#ifdef DEBUG
			D1print("EXTARQ\t.\n");
#endif
		/*
		** Verifica se o arquivo existe
		*/

#if defined(MSDOS) & !defined(DJGPP)
		if ( !_access(ops2, 00) )
#else
		if ( !access(ops2, 00) )
#endif
			    pilha[idpilha++] = (double) 1;
		else
			    pilha[idpilha++] = (double) 0;

#ifdef DEBUG
			D1print("EXTARQ\t%lg\n", op2);
#endif

		free(ops2);
		ops2 = (char *) NULL;

		break;
	/*
	** Obtem o tamanho de um string.
	*/

	case OSTRLEN: /* tamstr */

		ops2 = spilha[--idspilha];

#ifdef DEBUG
		D1print("STRLEN\t%s\n", ops2);
#endif

		pilha[idpilha++] = (double) strlen(ops2);

		/* TESTE */

		free(ops2);

		break;

	/*
	** Extrai o dia de uma data aaaammdd.
	*/

	case ODIA:


		if (op2 >= MISSING)
		    pilha[idpilha++] = MISSING;
		else 
		    pilha[idpilha++] = (double) ((long) op2 % 100L);

		break;

	/*
	** Extrai o mes de uma data aaaammdd.
	*/

	case OMES:

		if (op2 >= MISSING)
		    pilha[idpilha++] = MISSING;
		else 
		    pilha[idpilha++] = (double) (((long) op2 / 100L) % 100L);

		break;

	/*
	** Extrai o ano de uma data aaaammdd.
	*/

	case OANO:

		if (op2 >= MISSING)
		    pilha[idpilha++] = MISSING;
		else 
		    pilha[idpilha++] = (double) ((long) op2 / 10000L);

		break;

	/*
	** Informa o no. de dias corridos a partir de uma data.
	*/

	case ONDAY:

		/*
		** Verifica se e' MISSING.
		*/

		if (op2 >= MISSING) {
			pilha[idpilha++] = MISSING;

			break;
		}

		/*
		** Pega o ano ou a data inteira da pilha de dados.
		*/

		if (op2 > 9999.0) {
			temp  = (double) ((long) (op2 / 100.0)) ;
			op1  = op2 - temp*100.0;
			temp1  = (double) ((long) (temp / 100.0)) ;
			op3  = temp - temp1*100.0;
			op2  = temp1;
		}

		/*
		** Pega o mes e o dia que estao na pilha de dados.
		*/

		else {
			op3 = pilha[--idpilha];
			op1 = pilha[--idpilha];
		}

#ifdef DEBUG
		D1print("NDAY\t%f \t%f \t%f\n", op1, op3, op2);
#endif

		/*
		**  Testa se o valor do dia e' missing.
		*/

		if (op1 <= 0.0) {
			pilha[idpilha++] = MISSING;
			goto l1;
		}

		/*
		**  Testa se o valor do mes e' missing.
		*/

		if (op3 <= 0.0 || op3 > 12.0) {
			pilha[idpilha++] = MISSING;
			goto l1;
		}

		/*
		**  Testa se o valor do ano e' missing.
		*/

		if (op2 < 0.0) {
			pilha[idpilha++] = MISSING;
			goto l1;
		}

		switch( (int) op3) {

			/*
			**  Meses com 31 dias - testa se e' dia invalido.
			*/

			case 1:
			case 3:
			case 5:
			case 7:
			case 8:
			case 10:
			case 12:

				if (op1 > 31.0) { 
					pilha[idpilha++] = MISSING;
					goto l1;
				}

				break;

			/*
			**  Fevereiro.
			*/

			case 2:

				/*
				**  Fevereiro em ano comum - testa se 
				**  e' dia invalido.
				*/

				if ( ((int) op1%4) != 0)
					if (op1 > 28.0) {
						pilha[idpilha++] = MISSING;
						goto l1;
					}

				/*
				**  Fevereiro com 29 dias - testa se 
				**  e' dia invalido.
				*/

				if ( ((int) op1%400) == 0)
					if (op1 > 29.0) {
						pilha[idpilha++] = MISSING;
						goto l1;
					}

				if ( ((int) op1%100) != 0)
					if (op1 > 29.0) {
						pilha[idpilha++] = MISSING;
						goto l1;
					}

				break;

			/*
			**  Meses com 30 dias - testa se e' dia invalido.
			*/

			case 4:
			case 6:
			case 9:
			case 11:

				if (op1 > 30.0)  {
					pilha[idpilha++] = MISSING;

					goto l1;
				}

				break;
		}

		/*
		** Calcula o numero de dias corridos empilhando o resultado.
		*/

		nday = N_DIA_MES[(int) (op3-1.0)] + op2*365.0 + 
		       (double) (int) (op2 / 4.0) + op1 + 1.0 -
		       (double) (int) (op2/100.0) + (double) (int) (op2/400.0);

		/*
		** Ano bissexto com mes sendo janeiro ou fevereiro.
		*/

		if ( ((int) op2 % 400) == 0) {
			if (op3 <= 2.0)
				nday = nday - 1.0;
		}

		if ( ((int) op2 % 100) == 0)
			nday = nday - 1.0;

		pilha[idpilha++] = nday;

		break;

		l1:

/* Padroniz. msg erro - Adriana
		fprintf(stderr,"NDIA: Data invalida\n");
*** */

		fprintf(stderr, PegaMsg (1128) );
		termina (1);

		/*** ????? nao precisa pois e' goto ahhhgggg!!!
		 *** break;
		 ***/

	/*
	** Calcula a probabilidade N de um numero, desde que ele nao
	** seja missing.
	*/

	case ONPROB:

#ifdef DEBUG
		D1print("NPROB\n");
#endif

		if (vmissing((double) 0.0, op2))
			pilha[idpilha++] = nprob(op2);

		break;

	/*
	** Calcula a probabilidade N inversa de um numero, desde 
	** que ele nao seja missing.
	*/

	case ONINV:

#ifdef DEBUG
		D1print("NINV\n");
#endif

		if (op2 < 0.0 || op2 > 1.0) {
			pilha[idpilha++] = MISSING;
		}

		else {
			if (vmissing((double) 0.0, op2))
				pilha[idpilha++] = ninv(op2);
		}

		break;

	/*
	** Calcula a probabilidade T de um numero, desde 
	** que ele nao seja missing.
	*/

	case OTPROB:

		op1 = pilha[--idpilha];

#ifdef DEBUG
		D1print("TPROB\n");
#endif

		if (op2 <= 0.0) {
			pilha[idpilha++] = MISSING;
		}

		else {
			if (vmissing((double) 0.0, op2) && 
			    vmissing((double) 0.0, op1))
				pilha[idpilha++] = tprob(op1,op2);
		}

		break;

	/*
	** Calcula a probabilidade T inversa de um numero, desde 
	** que ele nao seja missing.
	*/

	case OTINV:

		op1 = pilha[--idpilha];

#ifdef DEBUG
		D1print("TINV\n");
#endif

		if (op2 <= 0.0 || op1 < 0.0 || op1 > 1.0) {
			pilha[idpilha++] = MISSING;
		}

		else {
			if (vmissing((double) 0.0, op2) && 
			    vmissing((double) 0.0, op1))
				pilha[idpilha++] = tinv(op1,op2);
		}

		break;

	/*
	** Calcula a probabilidade X de um numero, desde 
	** que ele nao seja missing.
	*/

	case OXPROB:

		op1 = pilha[--idpilha];

#ifdef DEBUG
		D1print("XPROB\n");
#endif

		if (op2 <= 0.0 || op1 < 0.0) {
			pilha[idpilha++] = MISSING;
		}

		else {
			if (vmissing((double) 0.0, op2) && 
			    vmissing((double) 0.0, op1))
				pilha[idpilha++] = xprob(op1,op2);
		}

		break;

	/*
	** Calcula a probabilidade X inversa de um numero, desde 
	** que ele nao seja missing.
	*/

	case OXINV:

		op1 = pilha[--idpilha];

#ifdef DEBUG
		D1print("XINV\n");
#endif

		if (op2 <= 0.0 || op1 < 0.0 || op1 > 1.0) {
			pilha[idpilha++] = MISSING;
		}

		else {
			if (vmissing((double) 0.0, op2) && 
			    vmissing((double) 0.0, op1))
				pilha[idpilha++] = xinv(op1,op2);
		}

		break;

	/*
	** Calcula a probabilidade F de um numero, desde 
	** que ele nao seja missing.
	*/

	case OFPROB:

		op3 = pilha[--idpilha];
		op1 = pilha[--idpilha];

#ifdef DEBUG
		D1print("FPROB\n");
#endif

		if (op2 <= 0.0 || op3 <= 0.0 || op1 < 0.0) {
			pilha[idpilha++] = MISSING;
		}

		else {
			if (vmissing((double) 0.0, op2) && 
			    vmissing((double) 0.0, op1) &&
			    vmissing((double) 0.0, op3))
				pilha[idpilha++] = fprob(op1,op3,op2);
		}

		break;

	/*
	** Calcula a probabilidade F inversa de um numero, desde 
	** que ele nao seja missing.
	*/

	case OFINV:

		op3 = pilha[--idpilha];
		op1 = pilha[--idpilha];

#ifdef DEBUG
		D1print("FINV\n");
#endif

		if (op2 <= 0.0 || op3 <= 0.0 || op1 < 0.0 || op1 > 1.0) {
			pilha[idpilha++] = MISSING;
		}

		else {
			if (vmissing((double) 0.0, op2) && 
			    vmissing((double) 0.0, op1) &&
			    vmissing((double) 0.0, op3))
				pilha[idpilha++] = finv(op1,op3,op2);
		}

		break;

	/*
	** Retorna a posicao de um sub-string em um string.
	*/

	case OSCAN: /* verstr */

		ops2 = spilha[--idspilha];
		ops1 = spilha[--idspilha];

#ifdef DEBUG
		D1print("SCAN\t%s %s\n", ops1, ops2);
#endif

		pilha[idpilha++] = (double) scan(ops1, ops2);

		free(ops2);
		ops2 = (char *) NULL;
		free(ops1);
		ops1 = (char *) NULL;

		break;
	}
}
