/***********************************************************************************
 Genese - Linguagem de programao para a gerao e manipulao de arquivo .dat

 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
 termos da Licena Pblica Geral GNU conforme publicada pela Free Software
 Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
 posterior.

 Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
 programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
 Temple Street, Suite 330, Boston, MA 02111-1307 USA.

 ***********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

	Identifica as funcoes que manipulam strings, executando-as

 Desenvolvido por:

      Oscar Ivan Palma Pacheco
      Carla Geovana do N. Macario
	Maria Fernanda Moura
	Adriana Delfino dos Santos
	

 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include <stdlib.h>

#ifndef VAXVMS
#include <malloc.h>
#endif

#ifdef SUN
extern time_t time();
#endif

#include "socasm.h"
#include "pcsgbd.h"
#include "genese.h"

#ifdef	MEM_DEBUG
#include	"mshell.h"
#endif


/* ----------------------------------------------------------------------------
**	
**
** Identifica funcoes que manipulam strings, executando-as
**
**
** Documentado por:
**
**		Carla Geovana do N. Macario
**		Oscar Ivan Palma Pacheco
**
**
** Data da ultima alteracao:
**
**		97/01/28 10:57:42
**
** ----------------------------------------------------------------------------
*/

#ifdef DEBUG
extern int	f_1;
#endif

extern	double	pilha[];	/* pilha de dados  */
extern	int	idpilha;	/* indice da pilha de dados  */
extern	char	*spilha[];	/* pilha de strings */
extern	int	idspilha;	/* indice da pilha de strings */
extern unsigned int inst;	/* instrucao do vetor de instrucoes a ser
				   executada  */

extern	double	op1, op2, op3;
extern	char	*ops1, *ops2, *ops3, *ops4;
extern	char	buffer[];

/*
** Tabela com os nomes dos dias da semana.
*/

char	*dia[7] = {	"Domingo",
			"Segunda",
			"Terca",
			"Quarta",
			"Quinta",
			"Sexta",
			"Sabado",
};

/*
** Tabela com os nomes dos meses do ano.
*/

char	*mes[12] = {	"Janeiro",
			"Fevereiro",
			"Marco",
			"Abril",
			"Maio",
			"Junho",
			"Julho",
			"Agosto",
			"Setembro",
			"Outubro",
			"Novembro",
			"Dezembro",
};



void funcs()
{

	long hora;
	struct tm *dthr;
	int i;
	int j;
	int lim;
	char *ptr;

	/* 
	** Pega o parametro da funcao a ser executada e faz um switch
	** caso seja um data ou hora por extenso.
	*/

	i = OPER1;
	if (i != OADATE && i != OATIME) {
		ops2 = spilha[--idspilha];
	}

	switch (i) {

	/*
	** Obtem a data por extenso, colocando o resultado na pilha de
	** strings.
	*/

	case OADATE:

		time(&hora);
		dthr = localtime(&hora);
		sprintf(buffer, "%s, %d de %s de %d", dia[dthr -> tm_wday],
			dthr -> tm_mday, mes[dthr -> tm_mon],
			dthr -> tm_year + 1900);

		if ((ops1 = (char *) malloc((unsigned) (strlen(buffer)+1))) ==
				(char *) NULL) {

			printf( PegaMsg (1039));
			termina (1);
		}

		strcpy(ops1, buffer);

#ifdef DEBUG
		D1print("ADATE\t%s\n", ops1);
#endif

		spilha[idspilha++] = ops1;

		break;

	/*
	** Obtem a hora por extenso, colocando o resultado na pilha de
	** strings.
	*/

	case OATIME:	/* ahora */

		time(&hora);
		dthr = localtime(&hora);
		sprintf(buffer, "%dh%dmin%ds", dthr -> tm_hour,
			dthr -> tm_min, dthr -> tm_sec);

		if ((ops1 = (char *) malloc((unsigned) (strlen(buffer)+1))) ==
				(char *) NULL)


		{
			printf ( PegaMsg (1039));
			termina (1);
		}

		strcpy(ops1, buffer);

#ifdef DEBUG
		D1print("ATIME\t%s\n", ops1);
#endif

		spilha[idspilha++] = ops1;

		break;

	/*
	** Converte um string para um numero float, colocando o resultado
	** na pilha de dados.
	*/

	case OATOF:

		ptr = ops2;

		while (*ptr == ' ')
			++ptr;

		/*
		** Testa se a string e' vazia.
		*/

		if (strlen(ptr) == 0) {
			op2 = MISSING;

#ifdef DEBUG
			D1print("ATOF\t.\n");
#endif

		}

		/*
		** Converte a string para float e libera sua area.
		*/

		else {
			op2 = atof(ptr);

#ifdef DEBUG
			D1print("ATOF\t%lg\n", op2);
#endif

		}

		pilha[idpilha++] = op2;
		free(ops2);
		ops2 = (char *) NULL;

		break;

	/*
	** Converte um string para uma data, colocando o resultado
	** na pilha de dados.
	*/

	case OATOD:

		ptr = ops2;

		while (*ptr == ' ')
			++ptr;

		/*
		** Testa se a string e' vazia.
		*/

		if (strlen(ptr) == 0) {
			op2 = MISSING;

#ifdef DEBUG
			D1print("ATOD\t.\n");
#endif

		}

		/*
		** Converte a string para data e libera sua area.
		*/

		else {
			op2 = atod(ptr);

#ifdef DEBUG
			D1print("ATOD\t%lg\n", op2);
#endif

		}

		pilha[idpilha++] = op2;
		free(ops2);
		ops2 = (char *) NULL;

		break;


	/*
	** Concatena duas strings, colocando o resultado na pilha de strings.
	*/

	case OSTRCAT:

		ops1 = spilha[--idspilha];


		if ( strlen (ops1) + strlen (ops2) > NOVO_MAXSTR )
		  {
		    printf (PegaMsg (1146), NOVO_MAXSTR );
		    termina (1);
		  }

		strcpy(buffer, ops1);
		strcat(buffer, ops2);
		free(ops2);
		free(ops1);
		ops1 = (char *) NULL;
		ops2 = (char *) NULL;
		spilha[idspilha + 1] = (char *) NULL;
		spilha[idspilha] = (char *) NULL;

		if ((ops1 = (char *) malloc((unsigned) (strlen(buffer)+1))) ==
				(char *) NULL){

			printf ( PegaMsg (1039));
			termina (1);
		}


		strcpy(ops1, buffer);

#ifdef DEBUG
		D1print("STRCAT\t%s\n", ops1);
#endif

		spilha[idspilha++] = ops1;

		break;

	/*
	** Funcao mid$ do BASIC.
	*/

	case OSTRMID:

		op2 = pilha[--idpilha];
		op1 = pilha[--idpilha];

		if ( (op1 < 0.0) || (op2 < 0.0) ) {
			/* Valor de Parametro Negativo ou Zero. Funcao : <%s> */
			printf (PegaMsg(1158),"substr"); 
		        termina (1);
		}

		if (!((op1 > 0.0) && (op1 <= strlen(ops2)) && (op2 >= 0))){
		    /* Valor dos Parametros Incompativel com o Tamanho da 
		    Cadeia de Caracteres. Funcao : <%s>" */
		    buffer[0] = '\0';
		}
		else {
		    j = (int) (op1 - 0.9);		/* inicio */
		    lim = (int) (op2 + 0.1);	/* fim */
		    buffer[0] = '\0';


		    if ( strlen (ops2) > NOVO_MAXSTR )
		      {
			printf (PegaMsg (1146), NOVO_MAXSTR);
			termina (1);
		      }

		    if (j < (int) strlen(ops2)) {
			for (i = 0; i < lim; ++i) {
				buffer[i] = ops2[j++];

				if (buffer[i] == '\0')
					break;
			}

			buffer[i] = '\0';
		    }
		}

		free(ops2);
		ops2 = (char *) NULL;
		spilha[idspilha] = (char *) NULL;

		if ((ops1 = (char *) malloc((unsigned) (strlen(buffer)+1))) ==
				(char *) NULL){

			printf( PegaMsg (1039));
			termina (1);
		}

		strcpy(ops1, buffer);

#ifdef DEBUG
		D1print("STRMID\t%s\n", ops1);
#endif

		spilha[idspilha++] = ops1;

		break;

	/*
	** Repete n vezes um caracter em uma dada string.
	*/

	case OASTRRP:

		op1 = pilha[--idpilha];


		if ( ( ( (strlen (ops2) ) * op1) + 1) > NOVO_MAXSTR )
		  {
		    printf (PegaMsg (1146), NOVO_MAXSTR);
		    termina (1);
		  }

		if ( ((int) (op1))  > 0 ) {

		    if ((ops1=
			(char *)malloc((unsigned)((int)op1*strlen(ops2)+1))) 
				== (char *) NULL){
				   printf( PegaMsg (1039));
				   termina (1);
		    }
		    repete(ops2, (int) op1, ops1);
		}
		else {
		    if ( ((int) (op1))  == 0 ) {
			ops1 = (char *) malloc((unsigned) 1);
			ops1 = '\0'; /* missing value */
		    }
		    else {
			/* Valor de Parametro Negativo ou Zero. Funcao : <%s> */
			printf (PegaMsg(1158),"repstr"); 
		        termina(1);
		    }
		}
			
		free(ops2);
		ops2 = (char *) NULL;

#ifdef DEBUG
		D1print("STRREP\t%s %f\n", ops1, op1);
#endif

		spilha[idspilha++] = ops1;

		break;

	/*
	** Retorna uma string entre determinados separadores.
	*/

	case OAWRDK: /* Funcao selstr */

		op1 = pilha[--idpilha];		/* numero */
		ops1 = spilha[--idspilha];	/* string */
		/* ops2 --->>> separador */

		if ( ((int) (op1) ) <= 0 ) {
			/* Valor de Parametro Negativo ou Zero. Funcao : <%s> */
			printf (PegaMsg(1158),"selstr"); 
			termina (1);
		}
		else {
		    if (strlen(ops1) <= 0 ) {
			spilha[idspilha++] = '\0'; /* missing */
		    }
		    else {
			if (strlen(ops2) <= 0 ) {
			    spilha[idspilha++] = '\0'; /* missing */
			}
			else {
			    spilha[idspilha++] = word_k(ops1, (int) op1, ops2);
			}
		    }
		}

		free(ops1);
		free(ops2);
		ops1 = (char *) NULL;
		ops2 = (char *) NULL;
		spilha[idspilha] = (char *) NULL;

#ifdef DEBUG
		D1print("GETWORD\t%s %f %s\n", ops1, op1, ops2);
#endif

		break;

	/*
	** Retira os brancos de um string.
	*/

	case OATRIM:


		if (((strlen(ops2))  + 1) > NOVO_MAXSTR)
		  {
		    printf ( PegaMsg (1146), NOVO_MAXSTR);
		    termina (1);
		  }


		if ((ops1 = (char *) malloc((unsigned) (strlen(ops2)+1))) ==
				(char *) NULL){
	  	   printf( PegaMsg (1039));
		   termina (1);
	       }

		if (strlen(ops2) > 0) 
		    trim(ops2, ops1);
		else 
		    ops1 [0] ='\0'; /* missing */

		spilha[idspilha++] = ops1;

		free(ops2);
		ops2 = (char *) NULL;


#ifdef DEBUG
		D1print("STRTRIM\t%s \n", ops2);
#endif

		break;

	/*
	** Troca os caracteres c de um string por outro, colocando na pilha
	** o resultado.
	*/

	case OASSTR: /* trcstr */

		/* 
		** Pega o caracter a ser trocado e o que ira lhe substituir.
		*/

		ops1 = spilha[--idspilha];
		ops3 = spilha[--idspilha];


		if ( (strlen (ops3) + 1 ) > NOVO_MAXSTR )
		  {
		    printf (PegaMsg (1146), NOVO_MAXSTR);
		    termina(1);
		  }

		if ((ops4 = (char *) malloc((unsigned) (strlen(ops3)+1))) ==
				(char *) NULL){
		  printf ( PegaMsg (1039));
		  termina (1);
	        }

		strcpy(ops4,"");

		/*
			ops1 = caracter a ser substituido
			ops2 = caracter que o substituira'
			ops3 = string original
			ops4 = string resultante 
		*/

		if (strlen(ops3) >= 1) {
			if ((strlen(ops1) >= 1) && (strlen(ops2) >=1)) {
			    substr(ops3, ops1, ops2, ops4);
			}
			else {
				strcpy(ops4,ops3); /* so' recebeu missing */
			}
		}
		else {
			strcpy(ops4,""); /* a string e' missing */
		}

		spilha[idspilha++] = ops4;

		free(ops1);
		free(ops2);
		free(ops3);

		ops1 = (char *) NULL;
		ops2 = (char *) NULL;
		ops3 = (char *) NULL;

		spilha[idspilha] = (char *) NULL;
		spilha[idspilha + 1] = (char *) NULL;
		spilha[idspilha + 2] = (char *) NULL;
#ifdef DEBUG
		D1print("substr\t%s %s %s\n", ops3, ops1, ops2);
#endif

		break;

	/*
	** Transforma todos os caracteres da string em maiusculos.
	*/

	case OAUPPER:  /* mastr */


		if ( (strlen (ops2) + 1 ) > NOVO_MAXSTR )
		  {
		    printf ( PegaMsg (1146), NOVO_MAXSTR);
		    termina  (1);
		  }


		if ((ops1 = (char *) malloc((unsigned) (strlen(ops2)+1))) ==
				(char *) NULL){
	 	  printf ( PegaMsg (1039));
		  termina (1);
		}

		/* 
			ops2 = string original
			ops1 = string resultante 
		*/

		supper(ops2, ops1);

		spilha[idspilha++] = ops1;

		free(ops2);
		ops2 = (char *) NULL;

#ifdef DEBUG
		D1print("SUPPER\t%s \n", ops2);
#endif

		break;

	/*
	** Transforma todos os caracteres da string em minusculos.
	*/

	case OALOWER: /* mistr */


		if ( (strlen (ops2) + 1) > NOVO_MAXSTR )
		  {
		    printf (PegaMsg (1146), NOVO_MAXSTR);
		    termina (1);
		  }


		if ((ops1 = (char *) malloc((unsigned) (strlen(ops2)+1))) ==
				(char *) NULL){
		  printf( PegaMsg (1039));
		  termina (1);
		}

		slower(ops2, ops1);

		spilha[idspilha++] = ops1;

		free(ops2);
		ops2 = (char *) NULL;

#ifdef DEBUG
		D1print("TOLOWER\t%s \n", ops2);
#endif

		break;

	/*
	** Posiciona a string `a esquerda
	*/

	case OAESQ:	/* esq */

		op1 = pilha[--idpilha];
/*
		  printf ("valor de op1: %l\n", op1);
*/

		if ( op1 < strlen (ops2) ) {
			  printf (PegaMsg (1151));
			  termina (1);
		}

		if ( ( strlen (ops2) > NOVO_MAXSTR ) ||
			     ( ( op1 + 1) > NOVO_MAXSTR ) )
		  {
		    printf (PegaMsg (1146), NOVO_MAXSTR);
		    termina (1);
		  }


		j = strlen(ops2);

		if ((ops1 = (char *) malloc((unsigned) op1+1)) ==
				(char *) NULL) {
		  printf (PegaMsg (1039));
		  termina (1);
		}


		strcpy(ops1,"");
		strcat(ops1, ops2);

		for ( i = j; i < op1; ++i )
			strcat(ops1, " ");
 	
		spilha[idspilha++] = ops1;

		free(ops2);
		ops2 = (char *) NULL;

#ifdef DEBUG
		D1print("ESQ\t%s \n", ops2);
#endif
		break;

	/*
	** Posiciona a string `a direita
	*/

	case OADIR:	/* dir */

		op1 = pilha[--idpilha];

		if ( op1 < strlen (ops2) ) {
		  printf (PegaMsg (1151));
		  termina (1);
		}

		if ( ( strlen (ops2) > NOVO_MAXSTR ) ||
		     ( ( op1 + 1) > NOVO_MAXSTR ) )
		  {
		    printf ( PegaMsg (1146), NOVO_MAXSTR);
		    termina (1);
		  }

		j = strlen(ops2);

		if ((ops1 = (char *) malloc((unsigned) op1+1)) ==
				(char *) NULL){
		  printf(PegaMsg (1039));
		  termina (1);
		}

		strcpy(ops1,"");
		for ( i = 0; i < op1 - j; ++i )
			strcat(ops1, " ");

		if ( j > 0 ) {
			strcat(ops1, ops2);
		}
 	
		spilha[idspilha++] = ops1;

		free(ops2);
		ops2 = (char *) NULL;

#ifdef DEBUG
		D1print("DIR\t%s \n", ops2);
#endif
		break;

	/*
	** Centraliza a string 
	*/

	case OACENT:	/* cent */

		{
		int	m;


		op1 = pilha[--idpilha];

		if ( op1 < strlen (ops2) ) {
		  printf (PegaMsg (1151));
		  termina (1);
		}

		if ( ( strlen (ops2) > NOVO_MAXSTR ) ||
		     ( ( op1 + 1) > NOVO_MAXSTR ) )
		  {
		    printf (PegaMsg (1146), NOVO_MAXSTR);
		    termina (1);
		  }

		m = 0;
		j = strlen(ops2);
		m = (int) ((int) op1 - j)/2;

		if ((ops1 = (char *) malloc((unsigned) op1+1)) ==
				(char *) NULL) {
			printf(PegaMsg (1039));
			termina (1);
		}


		strcpy(ops1,"");   /* Faltava inicializar a string - 
						Fernanda 19/07 */
		for ( i = 0; i < m; ++i )
			strcat(ops1, " ");

		strcat(ops1, ops2);

		for ( i = j + m; i < op1; ++i )
			strcat(ops1, " ");
 	
		spilha[idspilha++] = ops1;
		free(ops2);
		ops2 = (char *) NULL;

#ifdef DEBUG
		D1print("CEN\t%s \n", ops2);
#endif
		}
		break;

	/*
	** Justifica uma string
	*/

	case OAJUST:
		{
		int	numblk, numcont, resto, m;	
		char	*inicio1, *inicio2;	


		op1 = pilha[--idpilha];

		if ( op1 < strlen (ops2) ) {
		  printf (PegaMsg (1151));
		  termina (1);
		}

		if ( ( strlen (ops2) > NOVO_MAXSTR ) ||
		     ( ( op1 + 1) > NOVO_MAXSTR ) )
		  {
		    printf ( PegaMsg (1146), NOVO_MAXSTR);
		    termina (1);
		  }


		numblk = numcont = resto = m = 0;

		if ((ops1 = (char *) malloc((unsigned) op1+1)) ==
				(char *) NULL){

			printf(PegaMsg (1039));
			termina (1);
		}
		j = strlen(ops2);
		inicio2 = ops2;

		for (inicio2= ops2; *ops2 != '\0'; ++ops2)
			if ( *ops2 == ' ' )
				++numblk;
			else
				++numcont;
		ops2 = inicio2;
		
		/* Correcao Fernanda - 13/07/95 */
		if (numblk == 0 ) {
			/* Nao faz nada */
			strcpy(ops1,ops2);
		}
		else { /* o numero de brancos e' menor que o espaco restante */
		    m = ((int)  op1 - numcont - numblk)/numblk;

		    if ( m > 0 ) {
			    resto = ( (int) op1 - numcont) - (numblk * m);

			    for (inicio1 = ops1; *ops2 != '\0'; ++ops2) {
				if ( *ops2 == ' ' ) {
				    if ( numblk > resto )
					for ( i = 0; i < m; ++i ) {
						*ops1 = ' ';
						++ops1;
					--numblk;
					}
				    else
					    for ( i = 0; i <= m; ++i ) {
						*ops1 = ' ';
						++ops1;
					    }
				}
				else {	 
				    *ops1 = *ops2;
				    ++ops1;
				}
			    }
		    }
		    else { /* o numero de brancos e' maior que o espaco restante */
			    resto =  (int) op1 - numcont - numblk;

			    for (inicio1 = ops1; *ops2 != '\0'; ++ops2) {
				if ( *ops2 == ' ' && resto > 0) {
						*ops1 = ' ';
						++ops1; 
						--resto;
				}
			        *ops1 = *ops2;
			        ++ops1;
			    }
		    }

		    *ops1 = '\0';
		    ops2 = inicio2;
		    ops1 = inicio1;
		}
					
		spilha[idspilha++] = ops1;
		free(ops2);
		ops2 = (char *) NULL;

#ifdef DEBUG
		D1print("JUST\t%s \n", ops2);
#endif
		}
		break;

	/* 
	** Compara se duas strings sao diferentes.
	*/

	case ONEQ:

#ifdef DEBUG
		D1print("NEQ\n");
#endif

		ops1 = spilha[--idspilha];
		pilha[idpilha++] = (double) (strcmp(ops1, ops2) != 0);

		free(ops2);
		free(ops1);
		ops2 = (char *) NULL;
		ops1 = (char *) NULL;
		spilha[idspilha] = (char *) NULL;
		spilha[idspilha + 1] = (char *) NULL;

		break;

	/* 
	** Compara se duas strings sao iguais.
	*/

	case OEQ:

#ifdef DEBUG
		D1print("EQ\n");
#endif

		ops1 = spilha[--idspilha];
		pilha[idpilha++] = (double) (strcmp(ops1, ops2) == 0);
		free(ops2);
		free(ops1);
		ops2 = (char *) NULL;
		ops1 = (char *) NULL;
		spilha[idspilha] = (char *) NULL;
		spilha[idspilha + 1] = (char *) NULL;

		break;

	/* 
	** Compara se uma string e' menor que outra.
	*/

	case OLT:

#ifdef DEBUG
		D1print("LT\n");
#endif

		ops1 = spilha[--idspilha];
		pilha[idpilha++] = (double) (strcmp(ops1, ops2) < 0);
		free(ops1);
		free(ops2);
		ops1 = (char *) NULL;
		ops2 = (char *) NULL;
		spilha[idspilha] = (char *) NULL;
		spilha[idspilha + 1] = (char *) NULL;

		break;

	/* 
	** Compara se uma string e' menor ou igual a outra.
	*/

	case OLE:

#ifdef DEBUG
		D1print("LE\n");
#endif

		ops1 = spilha[--idspilha];
		pilha[idpilha++] = (double) (strcmp(ops1, ops2) <= 0);
		free(ops2);
		free(ops1);
		ops2 = (char *) NULL;
		ops1 = (char *) NULL;
		spilha[idspilha] = (char *) NULL;
		spilha[idspilha + 1] = (char *) NULL;

		break;

	/* 
	** Compara se uma string e' maior que outra.
	*/

	case OGT:

#ifdef DEBUG
		D1print("GT\n");
#endif

		ops1 = spilha[--idspilha];
		pilha[idpilha++] = (double) (strcmp(ops1, ops2) > 0);
		free(ops2);
		free(ops1);
		ops2 = (char *) NULL;
		ops1 = (char *) NULL;
		spilha[idspilha] = (char *) NULL;
		spilha[idspilha + 1] = (char *) NULL;

		break;

	/* 
	** Compara se uma string e' maior ou igual a outra.
	*/

	case OGE:

#ifdef DEBUG
		D1print("GE\n");
#endif

		ops1 = spilha[--idspilha];
		pilha[idpilha++] = (double) (strcmp(ops1, ops2) >= 0);
		free(ops2);
		free(ops1);
		ops2 = (char *) NULL;
		ops1 = (char *) NULL;
		spilha[idspilha] = (char *) NULL;
		spilha[idspilha + 1] = (char *) NULL;

		break;
	}
}

