/***********************************************************************************
 Genese - Linguagem de programao para a gerao e manipulao de arquivo .dat

 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Este programa  software livre; voc pode redistribu-lo e/ou modific-lo sob os
 termos da Licena Pblica Geral GNU conforme publicada pela Free Software
 Foundation; tanto a verso 2 da Licena, como (a seu critrio) qualquer verso
 posterior.

 Este programa  distribudo na expectativa de que seja til, porm, SEM NENHUMA
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral do GNU para mais detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral do GNU junto com este
 programa; se no, escreva para a Free Software Foundation, Inc., no endereo 59
 Temple Street, Suite 330, Boston, MA 02111-1307 USA.

 ***********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

	Contm as funes para criao e manipulao de uma pilha de dados

 Desenvolvido por:

	Carlos Eduardo Assis Paniago
	

 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/

/*
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// PILHA.C		1.2			92/03/26 13:21:27
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
// Funcao:
//	Implementa os operadores usados pelo 
//	tipo T_Pilha 
//
// Obs:
//	O tipo pilha implementado aqui foi baseado na
//	pilha definida em "Aho, A.V. & Hopcroft, J.E & Ullman, J.D.
//	Data Structures and Algorithms, Addison-Wesley, 1983"
//
//
// Sintaxe:
//
//	#include	"pilha.h"
//	
//	T_Pilha	Pil_Create();
//	T_Pilha	Pil_Destroy(T_Pilha);
//	T_Pilha	P_Push(T_Pilha, char);
//	T_Pilha	P_Pop(T_Pilha);
//	T_Pilha	P_MakeNull(T_Pilha);
//	char	P_Top(T_Pilha);
//	int	P_Depth(T_Pilha);
//	int	P_Empty(T_Pilha);
//	int	Pil_Empty(T_Pilha, int);
//
// Autor:
//	Macario, C.G.
//
// Historico:
//	12-9-88		Primeira Versao		C.G.Macario
// ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
*/

#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>

#if defined (MSDOS) | defined (DJGPP)
#   include <malloc.h>
#else
#   include <stdlib.h>
#endif

#include	"pilha.h"

static fl_open = 0;

/* Cria o Tipo T_Pilha, alocando area de memoria para ele */
T_Pilha	Pil_Create(n)
int n;
{
int	i;
	T_Pilha	P;
	P = (T_Pilha) calloc(1, sizeof(struct _T_Pilha));
	for (i=0; i< 1024; i++)
	    P->Item[i] = '\0'; /* (char **) calloc(n, sizeof(char *)); */
	P->Tamanho = n;
	P->Depth   = 0;
	return(P);
}
	
/* Cria o Tipo T_Pilha, alocando sua area */
T_Pilha	Pil_Destroy(P)
T_Pilha	P;
{
	if (P != (T_Pilha) NULL) {
		/* free((char *) P->Item); */
		free((char *) P);
		P = (T_Pilha) NULL;
	}
	return(P);
}	
 
/* Coloca um elemento na T_Pilha */
T_Pilha	P_Push(P, D)
T_Pilha	P;
char	D;
{
	if ( P->Depth >= P->Tamanho)
		printf("Erro! Estouro da pilha!!\n");
	else {
		P->Item[P->Depth++] = D;
	}
	return(P);
}

/* Retira um elemento na T_Pilha */
T_Pilha	P_Pop(P)
T_Pilha	P;
{
	 (P->Depth) ? --P->Depth : 0;
	return(P);
}	

/* Retorna o elemento que esta no topo da T_Pilha */
char	P_Top(P)
T_Pilha	P;
{
	if (P == (T_Pilha) NULL)
		return( (char) NULL );
	else
		return((P->Depth > 0) ? P->Item[(P->Depth) - 1] : (char) NULL);
}

/* Retorna o tamanho da T_Pilha */
int	P_Depth(P)
T_Pilha	P;
{
	return ( (P == (T_Pilha) NULL) ? 0 : (P->Depth) );
}

/* Retorna TRUE se a T_Pilha estiver vazia e FALSE caso contrario */
int	P_Empty(P)
T_Pilha	P;
{
	if (P == (T_Pilha) NULL)
		return(-1);
	else 
		return((P->Depth == 0) ? (int) -1 : (int) 0);
}

/* Armazena o estado do Tipo ( T_Pilha ) em um arquivo */
int	Pil_Print(P, numarq)
T_Pilha	P;
int	numarq;
{
	FILE	*fd;
	int	i;
	char	arq[128];

	if (P == (T_Pilha) NULL) {
		return(0);
	}

	sprintf(arq, "comp%d", numarq);

	if (fl_open) {
		fd = fopen(arq, "a+");
	}
	else {
		fd = fopen(arq, "w");
		fl_open = 1;
	}
	if (fd == (FILE *) NULL) {
		printf("Erro na abertura de arquivo (T_Pilha)\n");
		return(0);
	}
	fprintf(fd,"\n\nT_Pilha\n");
	for (i=0; i < P->Depth; ++i) {
		fprintf(fd,"%s\n", P->Item[i]);
	}

	fclose(fd);
	return(1);
}
