
/***********************************************************************************
 LIBSOC - Biblioteca padro para o SOftware Cientfico

 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Esta biblioteca  software livre; voc pode redistribu-la e/ou modific-la sob 
 os termos da Licena Pblica Geral Menor do GNU conforme publicada pela Free 
 Software Foundation; tanto a verso 2.1 da Licena, ou (a seu critrio) 
 qualquer verso posterior.

 Esta biblioteca  distribuda na expectativa de que seja til, porm, SEM NENHUMA 
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA 
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral Menor do GNU para mais 
 detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral Menor do GNU junto com 
 esta biblioteca; se no, escreva para a Free Software Foundation, Inc., no 
 endereo 59 Temple Street, Suite 330, Boston, MA 02111-1307 USA. 

 ***********************************************************************************

 Parte integrante do SOC - SOftware Cientfico 

 Descrio:
 
     Rotina para abrir arquivos de dados. Se o '.dat' esta' vazio, escreve 
     mensagem de erro 1003, Arquivo <arq> vazio e termina o processamento.
     Retorna o ponteiro para a base de dados aberta ou codigo de erro de
     abertura da base.


 Desenvolvido por:

		Oscar Ivan Palma Pacheco
		Carlos Fernando Assis Paniago
		Adriana Delfino dos Santos
		Carla Geovana do Nascimento Macrio


 Informaes do CVS:
  		$Source:
		$Revision:
		$Date:

***********************************************************************************/

/*
** --------------------------------------------------------------------------
**
** Rotina para abrir arquivos de dados. Nao faz leitura do ".dat" para 
** verificar se esta' vazio. Retorna o ponteiro de referencia da base
** de dados aberta ou codigo de erro.
**
** Forma de usar:
**
**		fdind = abre_arq_ponteiro(nome, modo);
**
** Variaveis envolvidas:
**
**		ARQ	*fdind;	ponteiro de referencia da base de dados
**		char	*nome;	nome da base de dados
**		int	modo;	0: leitura; 1: leitura e gravacao
**
** Valor retornado:
**
**			parq : Ponteiro de referencia da base de dados
**			  -1 : Nao conseguiu alocar memoria
**			  -2 : Erro na abertura do arquivo ".dat"
**			  -3 : Erro na abertura do arquivo ".hdr"
**			  -4 : Erro na leitura das variaveis
**
**			parq->nvar < 0 indica arquivo nao atualizado.
**
** 
** -------------------------------------------------------------------------
*/

#include <stdio.h>

#ifdef VAXVMS
# include <stdlib.h>
#else
# include <string.h>
# include <malloc.h>
#endif

#ifdef MSDOS
# include <stdlib.h>
# include <conio.h>
#endif

#ifdef SUN
# include <stdlib.h>
#endif

#include <math.h>

#include "pcsgbd.h"

#ifdef MEM_DEBUG
#include "mshell.h"
#endif

static char	**obs;

char	*__bufobs [OBSSIZE];


#ifdef MSDOS
# define MODO_0			"rb"
# define MODO_1			"rb+"
#else
# define MODO_0			"r"
# define MODO_1			"r+"
#endif

int ler0campo(fd, buffer)
FILE *fd;
char *buffer;
{
	int i;
	int c;
	int estado;

	estado = 0;

	if ((c = GetNext(fd)) == '\"')
		estado = 1;
	else
		UnGetNext(c, fd);

	i = 0;

	while (1) {
		c = GetNext(fd);

		if ((estado == 1 && c == '\"') || ( estado == 0
			&& (c == ';' || c == ',' || c == '\t' || c == '\n'))
			|| c == EOF) {
			if (estado == 1 && c == '\"') {
				c = GetNext(fd);
				if (c != ';' && c != ',' && c != '\t'
						&& c != '\n') {
					buffer[i++] = '\"';
					UnGetNext(c, fd);
					continue;
				}
			}
			buffer[i] = '\0';
			break;
		}
		else
			buffer[i++] = (char) c;

	}

	return c;
}

long ler0obs(fd, ivar, n, obslida, new_obs)
FILE	*fd;
VAR	*ivar;
int	n;
long	obslida;
char	*new_obs[];
{
	int	i, ch;
	char	*obser;

	/*
	** Ler observacao.
	*/

	obser = (char *) &new_obs[MAXVAR];
	new_obs[0] = (char *) NULL;

	for (i = 0; i < n; ++i) {

		new_obs[i] = obser;

		if((ch = ler0campo(fd, obser)) == EOF) {
			if (i != 0) {

/* *** Padroniz. msg erro - Adriana 30/11/94 
				fprintf(stderr,"\nObservacao %ld incompleta.\n",
*** */

				fprintf(stderr,
					PegaMsg (1052),
					obslida + 1,
					-2);
				return -2L;
			}

			return -1L;
		}

		if (ch == '\n' && i < n - 1) {

/* *** Padroniz. msg erro - Adriana 30/11/94 
			fprintf(stderr,"\nObservacao %ld incompleta.\n",
*** */

			fprintf(stderr,
				PegaMsg (1052),
				obslida + 1,
				-1);

			return -2L;
		}

		if (ivar -> tipo == 'c') {
			if (!strcmp(obser, "."))
				*obser = '\0';
		}
		else {
			if (!strlen(obser)) {
				*obser = '.';
				*(obser+1) = '\0';
			}
		}

		obser += (strlen(obser) + 1);
		++ivar;
	}

	new_obs[i] = (char *) NULL;

	/*
	** Atualizar numero da observacao lida.
	*/
	
	return (obslida + 1);

}

int atualarq(parq, nome)
ARQ	*parq;
char	*nome;
{
	int	i;
	int	n;
	int	t_arq;
	VAR	*ivar;
	FILE	*fd;
	long	oblida;
	char	comm[128];
	char	nomedat[128];
	char	nomehdr[128];
	int	ret;

	ret = 0;

	fprintf(stderr,
	"Atualizando base_ntia %s para a versao DEZEMBRO/90 ou posterior.\n",
		nome);

	n = parq -> nvar;

	ivar = parq -> pvar;

	t_arq = (ivar+n)->mmin.len;

	fclose(parq -> fdhdr);
	fclose(parq -> fddat);

	strcpy(nomedat, nome);
	strcpy(nomehdr, nome);
	strcat(nomedat, ".dat");
	strcat(nomehdr, ".hdr");

#ifdef MSDOS
	sprintf(comm, "copy %s tmpatu.dat > nul", nomedat);
#else
#ifdef VAXVMS
	sprintf(comm, "copy %s tmpatu.dat", nomedat);
#else
	sprintf(comm, "cp %s tmpatu.dat", nomedat);
#endif
#endif

	system(comm);

	if ((parq -> fddat = fopen(nomedat, "w+")) == (FILE *) NULL) {

/* *** Padroniz. msg erro - Adriana 15/12/94 
		printf("Nao consegui abrir <%s>\n", nomedat);
*** */

		printf(	PegaMsg (1034), nomedat);
		setOSstdin();
		exit(2);
	}

	if ((parq -> fdhdr = fopen(nomehdr, "r+")) == (FILE *) NULL) {

/* *** Padroniz. msg erro - Adriana 30/11/94 
		printf("Nao consegui abrir <%s>\n", nomehdr);
*** */

		printf(	PegaMsg (1034), nomehdr);
		setOSstdin();
		exit(2);
	}

	if ((fd = fopen("tmpatu.dat", "r")) == (FILE *) NULL) {

/* *** Padroniz. msg erro - Adriana 30/11/94 
		printf("Nao consegui abrir <tmpatu.dat>\n");
*** */

		printf(	PegaMsg (5020), "tmpatu.dat");
		setOSstdin();
		exit(2);
	}

	Set_Buff(parq -> fddat, (unsigned) 65000);
	Set_Buff(fd, (unsigned) 65000);

	for (i = 0; i < n; ++i) {
		if (ivar -> tipo == 'c') {
			ivar -> mmin.len = MAX_INT;
			ivar -> mmax.len = (int) MININT;
		}
		else {
			ivar -> mmin.valor = MAXREAL;
			ivar -> mmax.valor = MINREAL;
		}
		++ivar;
	}

	oblida = 0L;

	if ((obs = (char **) malloc((PATIPO) (OBSSIZE*sizeof(char *))))
			== (char **) NULL){

/* *** Padroniz. msg erro - Adriana 30/11/94 
		fprintf(stderr,"abre_arq: atualarq: Falta memoria para obs.\n");
*** */

		fprintf(stderr,	PegaMsg (1039) );
		setOSstdin();
		exit(25);
	}

	while (1) {
		switch(t_arq) {

		case 0:
			oblida = ler0obs(fd, ivar, n, oblida, obs);
			break;

		default:

/* *** Padroniz. msg erro - Adriana 30/11/94 
			fprintf(stderr,
				"ERRO: Versao invalida do arquivo NTIA.\n");
*** */

			fprintf(stderr, PegaMsg (1057) );
			setOSstdin();
			exit(5);
		}

		if (oblida < 0L)
			break;

		fprintf(stderr, "\r%ld", oblida);

		ivar = parq -> pvar;

		for (i = 0; i < n; ++i)
			minmax(ivar++, obs[i]);

		grav_obs(parq, obs);
	}

	free((ALTIPO) obs);

	if (oblida == -1L)
		parq -> novo = 1;	/* prepara para alterar */
	else {
		parq -> novo = 0;	/* nao altera		*/
		ret = -1;
	}

	if (fech_arq(parq) != 0) {

/* *** Padroniz. msg erro - Adriana 30/11/94 
		printf("\rNao consegui fechar <%s>\n", nome);
*** */

		printf(	PegaMsg (5005), nome);
		setOSstdin();
		exit(2);
	}

	if (oblida == -1L) {
		fprintf(stderr, "O arquivo_ntia %s foi atualizado.\n",
			nome);
	}
	else {

#ifdef MSDOS
		sprintf(comm, "copy tmpatu.dat %s > nul", nomedat);
#else
#ifdef VAXVMS
		sprintf(comm, "copy tmpatu.dat %s", nomedat);
#else
		sprintf(comm, "cp tmpatu.dat %s", nomedat);
#endif
#endif

		system(comm);


/* *** Padroniz. msg erro - Luciana 24/01/95 
		fprintf(stderr, "\rArquivo_ntia %s nao pode ser atualizado.  \n",
			nome);
*** */

		fprintf(stderr, PegaMsg (5054), nome);

	}

	Reset_Buff(fd);

	unlink("tmpatu.dat");

	return ret;
}

ARQ *abre_arq_ponteiro(nome, modo)
char	*nome;
int	modo;
{
	char	nomearq[128];
	ARQ	*parq;
	VAR	*pvar, *ivar;
	char	modoabre[6];
	char	buffer[128];
	int	quebra=0;

denovo:

#ifdef DEBUG
	printf("Abertura da base de dados %s.\n", nome);
#endif

	/*
	** Alocar area para arquivos.
	*/

#ifdef DEBUG
	printf("Alocando memoria.\n");
#endif

	if ((parq = (ARQ *) malloc((PATIPO) sizeof (ARQ))) == (ARQ *) NULL)
		return ((ARQ *) -1);
	if ((pvar = (VAR *) malloc(((PATIPO) sizeof (VAR)) * MAXVAR))
			== (VAR *) NULL) {
		free((ALTIPO) parq);
		return ((ARQ *) -1);
	}
	parq -> pvar = pvar;

	/*
	** Abrir arquivos.
	*/

	if (modo == 0)
		strcpy(modoabre, MODO_0);
	else
		strcpy(modoabre, MODO_1);

	strcpy(nomearq, nome);
	strcat(nomearq, ".dat");

#ifdef DEBUG
	printf("Abrindo arquivo <%s>\n", nomearq);
#endif

	if ((parq -> fddat = fopen(nomearq, modoabre)) == (FILE *) NULL) {
		free((ALTIPO) parq);
		free((ALTIPO) pvar);
		return ((ARQ *) -2);
	}

	strcpy(nomearq, nome);
	strcat(nomearq, ".hdr");

#ifdef DEBUG
	printf("Abrindo arquivo <%s>\n", nomearq);
#endif

	if ((parq -> fdhdr = fopen(nomearq, modoabre)) == (FILE *) NULL) {
		fclose(parq -> fddat);
		free((ALTIPO) parq);
		free((ALTIPO) pvar);
		return ((ARQ *) -3);
	}

	/*
	** Leitura da informacao a respeito das variaveis.
	*/

	strcpy(parq -> nome, nome);
	parq -> novo = 0;
	ivar = pvar;
	parq -> nvar = 0;
	parq -> proxobs = 0L;
	fseek(parq -> fdhdr, 0L, 0);

#ifdef DEBUG
	printf("Listando informacao da base %s\n", nome);
#endif

	while (1) {
		if (quebra = (lercampo(parq -> fdhdr, ivar -> nome)==EOF))
					    break;

		if (quebra= (lercampo(parq -> fdhdr, &(ivar -> tipo))==EOF))
					    break;

		if (quebra= (lercampo(parq -> fdhdr, buffer)==EOF))
					    break;

		if (ivar -> tipo == 'c')
			ivar -> mmin.len = atoi(buffer);
		else
			ivar -> mmin.valor = atof(buffer);

		if (quebra=(lercampo(parq -> fdhdr, buffer)==EOF))
			break;

		if (ivar -> tipo == 'c')
			ivar -> mmax.len = atoi(buffer);
		else
			if (strcmpi(buffer, ".") == 0)
				ivar -> mmax.valor = MAXREAL;
			else
				ivar -> mmax.valor = atof(buffer);

		if (strlen(ivar -> nome) == 0)
			break;

#ifdef DEBUG
		printf("Nomevar: %s\n", ivar -> nome);
		printf("tipo   : %c\n", ivar -> tipo);
		if (ivar -> tipo == 'c') {
			printf("minlen : %d\n", ivar -> mmin.len);
			printf("maxlen : %d\n", ivar -> mmax.len);
		}

		else if (ivar -> tipo == 'd') {
			printf("mindat : %ld\n", (long) (ivar -> mmin.valor));
			printf("maxdat : %ld\n", (long) (ivar -> mmax.valor));
		}

		else {
			printf("minval : %f\n", ivar -> mmin.valor);
			printf("maxval : %f\n", ivar -> mmax.valor);
		}
#endif

		parq -> nvar++;
		ivar++;
	}
	if (quebra != 0) {	/* final inesperado do hdr */
		fclose(parq -> fdhdr);
		fclose(parq -> fddat);
		free((ALTIPO) parq);
		free((ALTIPO) pvar);
		return ((ARQ *) -4);
	}

	if (ivar->mmin.len != TIPOARQ) {
		if (atualarq(parq, nome) == 0)
			goto denovo;
		else {
			setOSstdin();
			exit(5);
		}
	}

	if (modo == 0) {
		fclose(parq -> fdhdr);
		parq -> fdhdr = (FILE *) NULL;
	}

	Set_Buff(parq -> fddat, (unsigned) TAMFILEBUF);

	return parq;
}






/* ----------------------------------------------------------------------------
**
**
** Rotina para abrir arquivos de dados. Se o '.dat' esta' vazio, escreve 
** mensagem de erro 1003, Arquivo <arq> vazio e termina o processamento.
** Retorna o ponteiro para a base de dados aberta ou codigo de erro de
** abertura da base.
**
** Forma de usar:
**
**		fdind = abre_arq(nome, modo);
**
** Variaveis envolvidas:
**
**		ARQ	*fdind;	ponteiro de referencia da base de dados
**		char	*nome;	nome da base de dados
**		int	modo;	0: leitura; 1: leitura e gravacao
**
** Valor retornado:
**
**			parq : Ponteiro de referencia da base de dados
**			  -1 : Nao conseguiu alocar memoria
**			  -2 : Erro na abertura do arquivo ".dat"
**			  -3 : Erro na abertura do arquivo ".hdr"
**			  -4 : Erro na leitura das variaveis
**
**			parq->nvar < 0 indica arquivo nao atualizado.
**
**
**
** Data da ultima alteracao:
**
** -------------------------------------------------------------------------
*/

ARQ *abre_arq(nome, modo, flgvazio)
char	*nome;
int	modo;
int	flgvazio;
{

#ifndef SEEK_SET
# define SEEK_SET       0       /* inicio do arquivo */
#endif
 
	char	nomearq[128];
	ARQ	*parq;
	VAR	*pvar, *ivar;
	char	modoabre[6];
	char	buffer[128];
	int	quebra=0;


denovo:

#ifdef DEBUG
	printf("Abertura da base de dados %s.\n", nome);
#endif

	/*
	** Alocar area para arquivos.
	*/

#ifdef DEBUG
	printf("Alocando memoria.\n");
#endif

	if ((parq = (ARQ *) malloc((PATIPO) sizeof (ARQ))) == (ARQ *) NULL)
		return ((ARQ *) -1);
	if ((pvar = (VAR *) malloc(((PATIPO) sizeof (VAR)) * MAXVAR))
			== (VAR *) NULL) {
		free((ALTIPO) parq);
		return ((ARQ *) -1);
	}
	parq -> pvar = pvar;

	/*
	** Abrir arquivos.
	*/

	if (modo == 0)
		strcpy(modoabre, MODO_0);
	else
		strcpy(modoabre, MODO_1);

	strcpy(nomearq, nome);
	strcat(nomearq, ".dat");

#ifdef DEBUG
	printf("Abrindo arquivo <%s>\n", nomearq);
#endif

	if ((parq -> fddat = fopen(nomearq, modoabre)) == (FILE *) NULL) {
		free((ALTIPO) parq);
		free((ALTIPO) pvar);
		return ((ARQ *) -2);
	}


	strcpy(nomearq, nome);
	strcat(nomearq, ".hdr");

#ifdef DEBUG
	printf("Abrindo arquivo <%s>\n", nomearq);
#endif

	if ((parq -> fdhdr = fopen(nomearq, modoabre)) == (FILE *) NULL) {
		fclose(parq -> fddat);
		free((ALTIPO) parq);
		free((ALTIPO) pvar);
		return ((ARQ *) -3);
	}

	/*
	** Leitura da informacao a respeito das variaveis.
	*/

	strcpy(parq -> nome, nome);
	parq -> novo = 0;
	ivar = pvar;
	parq -> nvar = 0;
	parq -> proxobs = 0L;
	fseek(parq -> fdhdr, 0L, 0);

#ifdef DEBUG
	printf("Listando informacao da base %s\n", nome);
#endif

	while (1) {
		if (quebra = (lercampo(parq -> fdhdr, ivar -> nome)==EOF))
					    break;

		if (quebra= (lercampo(parq -> fdhdr, &(ivar -> tipo))==EOF))
					    break;

		if (quebra= (lercampo(parq -> fdhdr, buffer)==EOF))
					    break;

		if (ivar -> tipo == 'c')
			ivar -> mmin.len = atoi(buffer);
		else
			ivar -> mmin.valor = atof(buffer);

		if (quebra=(lercampo(parq -> fdhdr, buffer)==EOF))
			break;

		if (ivar -> tipo == 'c')
			ivar -> mmax.len = atoi(buffer);
		else
			if (strcmpi(buffer, ".") == 0)
				ivar -> mmax.valor = MAXREAL;
			else
				ivar -> mmax.valor = atof(buffer);

		if (strlen(ivar -> nome) == 0)
			break;

#ifdef DEBUG
		printf("Nomevar: %s\n", ivar -> nome);
		printf("tipo   : %c\n", ivar -> tipo);
		if (ivar -> tipo == 'c') {
			printf("minlen : %d\n", ivar -> mmin.len);
			printf("maxlen : %d\n", ivar -> mmax.len);
		}

		else if (ivar -> tipo == 'd') {
			printf("mindat : %ld\n", (long) (ivar -> mmin.valor));
			printf("maxdat : %ld\n", (long) (ivar -> mmax.valor));
		}

		else {
			printf("minval : %f\n", ivar -> mmin.valor);
			printf("maxval : %f\n", ivar -> mmax.valor);
		}
#endif

		parq -> nvar++;
		ivar++;
	}
	if (quebra != 0) {	/* final inesperado do hdr */
		fclose(parq -> fdhdr);
		fclose(parq -> fddat);
		free((ALTIPO) parq);
		free((ALTIPO) pvar);
		return ((ARQ *) -4);
	}

	if (ivar->mmin.len != TIPOARQ) {
		if (atualarq(parq, nome) == 0)
			goto denovo;
		else {
			setOSstdin();
			exit(5);
		}
	}

	if (modo == 0) {
		fclose(parq -> fdhdr);
		parq -> fdhdr = (FILE *) NULL;
	}

	Set_Buff(parq -> fddat, (unsigned) TAMFILEBUF);

	if (ler_obs (parq, __bufobs) == EOF )
	{
	    if ( flgvazio ) {	
		printf(PegaMsg (1003), parq->nome);
		fclose (parq -> fddat);
		free((ALTIPO) parq);
		free((ALTIPO) pvar);
		setOSstdin();
		exit (2);
	    }
	}
	else
	  rew_arq (parq);

	return parq;
}

