/***********************************************************************************
 LIBSOC - Biblioteca padro para o SOftware Cientfico

 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Esta biblioteca  software livre; voc pode redistribu-la e/ou modific-la sob 
 os termos da Licena Pblica Geral Menor do GNU conforme publicada pela Free 
 Software Foundation; tanto a verso 2.1 da Licena, ou (a seu critrio) 
 qualquer verso posterior.

 Esta biblioteca  distribuda na expectativa de que seja til, porm, SEM NENHUMA 
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA 
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral Menor do GNU para mais 
 detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral Menor do GNU junto com 
 esta biblioteca; se no, escreva para a Free Software Foundation, Inc., no 
 endereo 59 Temple Street, Suite 330, Boston, MA 02111-1307 USA. 

 ***********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

    Calcula a probabilidade acumulada para uma Beta incompleta. 

DEsenvovlido por:

		Dalton Francisco Andrade
       	Oscar Ivan Palma Pacheco
		Slvio Roberto Medeiros Evangelista


 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/

/*
**------------------------------------------------------------------------
**         CDBeta  -- Cumulative Beta Distribution
**------------------------------------------------------------------------
**
** Function:	CDBeta
**
** Purpose:	Evaluates CPDF of Incomplete Beta Function
**
** Calling Sequence:
**
**	P	= CDBeta(X, Alpha, Beta: REAL;
**			Dprec, Maxitr	: INTEGER;
**			VAR Cprec	: REAL;
**			VAR Iter	: INTEGER;
**			VAR Ifault	: INTEGER  ) : REAL;
**
**	X	--- Upper percentage point of PDF
**	Alpha	--- First shape parameter
**	Beta	--- Second shape parameter
**	Dprec	--- Number of digits of precision required
**	Maxitr	--- Maximum number of iterations
**	Cprec	--- Actual resulting precision
**	Iter	--- Iterations actually used
**	Ifault	--- error indicator
**			= 0:	no error
**			= 1:	argument error
**			= 4:	erro na t_student aproximar para 1.0
**
**	P	--- Resultant probability
**
**	Calls:
**
**	ALGama
**
**	Method:
**
**		The continued fraction expansion as given by
**		Abramowitz and Stegun (1964) is used.  This
**		method works well unless the minimum of (Alpha, Beta)
**		exceeds about 70000.
**
**		An error in the input arguments results in a returned
**		probability of -1.
**
**
**------------------------------------------------------------------------
*/

#include <stdio.h>
#include <math.h>

#include "pcsgbd.h"
#include "const.h"

double CDBeta(X, Alpha, Beta, Dprec, MaxIter, Cprec, Iter,Ifault)
double	X;
double	Alpha;
double	Beta;
int	Dprec;
int	MaxIter;
double	*Cprec;
int	*Iter;
int	*Ifault;
{
	double	AUX;
	double	Epsz, A, B, C, F, Fx, Apb, Zm, Alo, Ahi, Blo;
	double	Bhi,Bod,Bev,Zm1,D1,Aev,Aod;
	double	val_cd;

	int	QSwap, Qconv;	/* booleanas */

	/*
	** Initialize
	*/

	if (Dprec > MaxPrec)
		Dprec = MaxPrec;
	else {
		if (Dprec <= 0)
			Dprec = 1;
	}

	*Cprec = (double) Dprec;

	Epsz = pow(10.0, (double)-Dprec);

	A = Alpha;
	B = Beta;
	QSwap = 0;
	val_cd = -1.0;

	/*
	** Check arguments
	*/

	/*
	** Error if: X <= 0  A <= 0  B <= 0
	*/

	*Ifault = 1;

	if(X <= 0.0)
		return(val_cd);

	if((A <= 0.0) || (B <= 0.0))
		return(val_cd);

	val_cd = 1.0;
	*Ifault = 0;

	/*
	** If X >= 1, return 1.0 as prob
	*/

	if(X >= 1.0)
		return(val_cd);

	/*
	** If X > A / ( A + B ) then swap
	** A, B for more efficient eval.
	*/

	if(X > (A / (A + B))) {
		X = 1.0 - X;
		A = Beta;
		B = Alpha;
		QSwap = 1;
	}

	/*
	** Check for extreme values
	*/

	if (X != A && X != B && A != ((B * X) / (1.0 - X)) &&
			fabs(A - (X * (A + B))) > Epsz) {

		C = ALGama(A + B) + A * log(X) +
			B * log(1.0 - X) - ALGama(A) - ALGama(B) -
			log(A - X * (A + B));

		if((C < -36.0) && QSwap)
			return(val_cd);

		val_cd = 0.0;
		if(C < -180.0)
			return(val_cd);
	}

	/*
	** Set up continued fraction expansion
	** evaluation.
	*/

	Apb = A + B;
	Zm = 0.0;
	Alo = 0.0;
	Bod = 1.0;
	Bev = 1.0;
	Bhi = 1.0;
	Blo = 1.0;

	AUX = ALGama(Apb) + A * log(X) +
		B * log(1.0 - X) - ALGama(A + 1.0) -
		ALGama(B);

	if (AUX < MinExp) {
		*Ifault = 4;
		return(-1.0);
	}
	else
		Ahi = exp(AUX);

	F = Ahi;

	*Iter = 0;

	/*
	** Continued fraction loop begins here.
        ** Evaluation continues until maximum
        ** iterations are exceeded, or
        ** convergence achieved.
	*/

	Qconv = 0;

	do {
		Fx = F;
		Zm1 = Zm;
		Zm = Zm + 1.0;
		D1 = A + Zm + Zm1;
		Aev = -(A + Zm1) * (Apb + Zm1) * X / D1 / (D1 - 1.0);
		Aod = Zm * (B - Zm) * X / D1 / (D1 + 1.0);
		Alo = Bev * Ahi + Aev * Alo;
		Blo = Bev * Bhi + Aev * Blo;
		Ahi = Bod * Alo + Aod * Ahi;
		Bhi = Bod * Blo + Aod * Bhi;

		if (fabs(Bhi) < Rsmall)
			Bhi = 0.0;

		if(Bhi != 0.0) {
			F = Ahi / Bhi;
			Qconv = (fabs((F - Fx) / F) < Epsz);
		}

		*Iter = *Iter + 1;

	} while ((*Iter <= MaxIter) && (!Qconv));

	/*
	** Arrive here when convergence
	** achieved, or maximum iterations
	** exceeded.
	*/

	if (QSwap)
		val_cd = 1.0 - F;
	else
		val_cd = F;

	/*
	** Calculate precision of result
	*/

	if (fabs(F - Fx) != 0.0)
		*Cprec = -log10(fabs(F - Fx));
	else
		*Cprec = (double) MaxPrec;

	return(val_cd);

}	/* CDBeta */

