/***********************************************************************************
 LIBSOC - Biblioteca padro para o SOftware Cientfico

 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Esta biblioteca  software livre; voc pode redistribu-la e/ou modific-la sob 
 os termos da Licena Pblica Geral Menor do GNU conforme publicada pela Free 
 Software Foundation; tanto a verso 2.1 da Licena, ou (a seu critrio) 
 qualquer verso posterior.

 Esta biblioteca  distribuda na expectativa de que seja til, porm, SEM NENHUMA 
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA 
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral Menor do GNU para mais 
 detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral Menor do GNU junto com 
 esta biblioteca; se no, escreva para a Free Software Foundation, Inc., no 
 endereo 59 Temple Street, Suite 330, Boston, MA 02111-1307 USA. 

 ***********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

    Rotinas para gerenciar uma base de dados no formato SOC (.hdr + .dat): 
    copiar base, renomear base, posicionar no incio da base, ler um campo
    na base e gravar um campo na base.

 Desenvolvido por:

		Oscar Ivan Palma Pacheco
		Adriana Delfino dos Santos
		Carla Geovana do Nascimento Macrio
		Marcos Lordello Chaim
		
 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/


/*
** --------------------------------------------------------------------------
** Rotina para copiar base de dados.
**
** Forma de usar:
**
**		cod = copia_arq(de, para);
**
** Variaveis envolvidas:
**
**		int	cod;	Codigo de retorno
**		char	*de;	nome da base de dados original
**		char	*para;	nome da nova base de dados
**
** Valor retornado:
**
**			   0 : Retorno nornal
**			  -1 : Erro ao tentar copiar o arquivo .hdr
**			  -2 : Erro ao tentar copiar o arquivo .dat
**			  -3 : Erro ao tentar copiar ambos arquivos
**			  -4 : Erro no system() para .hdr
**			  -5 : Erro no system() para .dat
**
** ----------------------------------------------------------------------------
**
** Rotina para renomear base de dados.
**
** Forma de usar:
**
**		cod = rename_arq(de, para);
**
** Variaveis envolvidas:
**
**		int	cod;	Codigo de retorno
**		char	*de;	nome da base de dados original
**		char	*para;	nome da nova base de dados
**
** Valor retornado:
**
**			   0 : Retorno nornal
**			  -1 : Arquivo .hdr nao existe
**			  -2 : Arquivo .dat nao existe
**			  -3 : Ambos arquivos nao existem
**			  -4 : Erro ao renomear arquivo .hdr
**			  -5 : Erro ao renomear arquivo .dat
**
** ----------------------------------------------------------------------------
**
** Rotina para ler um campo num arquivo.
**
** Forma de usar:
**
**		cod = lercampo(fd, s);
**
** Variaveis envolvidas:
**
**		int	cod	 Valor retornado.
**		FILE	*fd	 Ponteiro de referencia do arquivo
**		char	*s	 Buffer de leitura fornecido pelo cliente.
**
** Valor retornado:
**
**			EOF  : Fim de arquivo
**			';'  : Fim de campo
**			'\n' : Fim do ultimo campo
**
** --------------------------------------------------------------------------
**
** Rotina para gravar um campo num arquivos de dados.
**
** Forma de usar:
**
**		gravacampo(fd, s, sep)
**
** Variaveis envolvidas:
**
**		FILE	*fd;	Ponteiro para referencia do arquivo
**		char	*s;	Cadeia de caracteres a ser gravada
**		char	sep;	Separador
**
** Valor retornado:
**
**			  0 : Normal
**			EOF : Nao conseguiu gravar no arquivo.
**
** --------------------------------------------------------------------------
**
** Rotina para posicionar no inicio dos arquivos.
**
** Forma de usar:
**
**		rew_arq(parq);
**
** Variaveis envolvidas:
**
**		ARQ	*parq;	Ponteiro de referencia da base de dados
**
** Valor retornado:
**
**			Codigo de retorno do fseek.
**
** --------------------------------------------------------------------------
*/

#include	<stdio.h>
#include	<malloc.h>
#include	<stdlib.h>

#ifdef MSDOS
#include	<process.h>
#include	<conio.h>
#include	<io.h>
#endif

#ifdef RS6000
#include <sys/io.h>
#endif

#ifndef VAXVMS
# include	<string.h>
#endif

#ifdef SUN
#include	<unistd.h>
#endif

#ifdef ULTRIX
#include	<unistd.h>
#endif

#include	"pcsgbd.h"

#ifdef MEM_DEBUG
#include 	"mshell.h"
#endif

#ifdef MSDOS
# define MAXFILES		50
# define RWTT			unsigned
#else
# ifdef SCO
#  define MAXFILES		100
#  define RWTT			unsigned
# else
#  define MAXFILES		100
#  define RWTT			int
# endif
#endif

static char	*buffer[MAXFILES];

static int	indb[MAXFILES];
static int	tamb[MAXFILES];
static char	PNflag[MAXFILES];
static long	OffSet[MAXFILES];

static RWTT	TamBuf[MAXFILES] = {		0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
#ifndef MSDOS
						0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
#endif
						0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
						0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

extern	int	__tamobs;


int copia_arquivo (char *de, char *para);

char 	*pega_env(arq, ext)
char	*arq;
char	*ext;
{

	static char fileaux[TAMSTR];
	char *parea;

	if(strrchr(arq, '/') != (char *) NULL)
		return(arq);

#ifdef MSDOS
	if(strrchr(arq, '\\') != (char *) NULL)
		return(arq);
#endif

	strcpy(fileaux, arq);
	strcat(fileaux, ext);

	if(!(access(fileaux,0)))
		return(arq);

        parea = getenv("NTIABASES");

        if(parea != (char *) NULL) {
		strcpy(fileaux, parea);
#ifndef VAXVMS
#ifdef  MSDOS
                strcat(fileaux, "\\");
#else
                strcat(fileaux, "/");
#endif
#endif
		strcat(fileaux, arq);

		return(fileaux);
        }

	return(arq);
}

int	copia_arq(de, para)
char	*de;
char	*para;
{
	int	ret;
	char	buffde[256];
	char	buffpara[256];

	ret = 0;

	strcpy(buffde, de);
	strcat(buffde, ".dat");

	if (access(buffde, F_OK))
		ret -= 2;

	strcpy(buffde, de);
	strcat(buffde, ".hdr");

	if (access(buffde, F_OK))
		ret -= 1;


	strcpy(buffde, de);
	strcat(buffde, ".hdr");
	strcpy(buffpara, para);
	strcat(buffpara, ".hdr");

	if (!copia_arquivo(buffde, buffpara))
		ret =  -4;

	strcpy(buffde, de);
	strcat(buffde, ".dat");
	strcpy(buffpara, para);
	strcat(buffpara, ".dat");

	if (!copia_arquivo(buffde, buffpara))
		ret =  -5;

	return ret;
}

int	rename_arq(de, para)
char	*de;
char	*para;
{
	int	ret;
	char	file1[128];
	char	file2[128];

	ret = 0;

	strcpy(file1, de);
	strcat(file1, ".dat");

	if (access(file1, F_OK))
		ret -= 2;

	strcpy(file2, de);
	strcat(file2, ".hdr");

	if (access(file2, F_OK))
		ret -= 1;

	if (ret == 0) {
		apaga_arq(para);

		strcpy(file2, para);
		strcat(file2, ".dat");

		if (rename(file1, file2) != 0)
			ret = -4;

		strcpy(file1, de);
		strcat(file1, ".hdr");
		strcpy(file2, para);
		strcat(file2, ".hdr");

		if (rename(file1, file2) != 0)
			ret = -5;
	}

	return ret;
}

int ToNextObs(fd)
ARQ *fd;
{
	int i;

	i = fileno(fd -> fddat);

	if (i < MAXFILES && TamBuf[i]) {
		while (1) {
			if (indb[i] >= tamb[i]) {
				if (tamb[i] && (unsigned) tamb[i] < TamBuf[i])
					return EOF;

				OffSet[i] += ((long) tamb[i]);

				tamb[i] = read(i, buffer[i], TamBuf[i]);

				if (tamb[i] <= 0)
					return EOF;

				indb[i] = 0;
			}

			if (buffer[i][indb[i]] == '\n')
				break;

			++indb[i];
		}

		++indb[i];
	}
	else
		return 1;

	return 0;
}

int ToPrevObs(fd)
ARQ *fd;
{
	int i;

	i = fileno(fd -> fddat);

	if (i < MAXFILES && TamBuf[i]) {
		if (indb[i] == 0) {
			if (OffSet[i] == 0)
				return EOF;

			OffSet[i] -= TamBuf[i];

			lseek(i, OffSet[i], SEEK_SET);

			tamb[i] = read(i, buffer[i], TamBuf[i]);

			if (tamb[i] <= 0)
				return EOF;

			indb[i] = tamb[i] - 1;
		}

		else
			--indb[i];

		while (1) {
			if (indb[i] == 0) {
				if (OffSet[i] == 0)
					return EOF;

				OffSet[i] -= TamBuf[i];

				lseek(i, OffSet[i], SEEK_SET);

				tamb[i] = read(i, buffer[i], TamBuf[i]);

				if (tamb[i] <= 0)
					return EOF;

				indb[i] = tamb[i] - 1;
			}

			else
				--indb[i];

			if (buffer[i][indb[i]] == '\n')
				break;
		}

		++indb[i];
	}
	else
		return 1;

	return 0;
}

int Set_OffSet(fd, new_OS)
ARQ *fd;
long new_OS;
{
	int i;
	long New_FOS;

	i = fileno(fd -> fddat);

	if (i < MAXFILES && TamBuf[i]) {
		if (PNflag[i] && indb[i]) {
			if (ftell(fd -> fddat) != OffSet[i])
				lseek(i, OffSet[i], SEEK_SET);

			tamb[i] = write(i, buffer[i], (RWTT) indb[i]);

			if (tamb[i] < indb[i])
				return EOF;

			indb[i] = 0;

			PNflag[i] = 0;
		}

		New_FOS = (new_OS / TamBuf[i]) * TamBuf[i];

		if (New_FOS != OffSet[i] || tamb[i] == 0) {
			OffSet[i] = New_FOS;

			lseek(i, OffSet[i], SEEK_SET);

			tamb[i] = read(i, buffer[i], TamBuf[i]);

			if (tamb[i] <= 0)
				return EOF;
		}

		indb[i] = (int) (new_OS - OffSet[i]);
	}

	else
		return 1;

	return 0;
}

long Curr_OffSet(fd)
ARQ *fd;
{
	int	i;
	long	ret;

	i = fileno(fd -> fddat);

	if (i < MAXFILES && TamBuf[i]) {
		ret = OffSet[i] + (long) indb[i];
	}
	else
		ret = -1L;

	return ret;
}

int UnGetNext(c, fd)
int c;
FILE *fd;
{
	int i, ret;

	i = fileno(fd);

	if (i < MAXFILES && TamBuf[i]) {
		if (c != EOF)
			--indb[i];

		ret = c;
	}
	else
		ret = ungetc(c, fd);

	return ret;
}

int GetNext(fd)
FILE *fd;
{
	int i, ret;

	i = fileno(fd);

	PNflag[i] = 0;

	do {
		if (i < MAXFILES && TamBuf[i]) {
			if (indb[i] >= tamb[i]) {
				if (tamb[i] && (unsigned) tamb[i] < TamBuf[i])
					return EOF;

				OffSet[i] += ((long) tamb[i]);

				tamb[i] = read(i, buffer[i], TamBuf[i]);

				if (tamb[i] <= 0)
					return EOF;

				indb[i] = 0;
			}

			ret = (int) buffer[i][indb[i]];

			++indb[i];
		}
		else
			ret = getc(fd);
	} while (ret == 0x0d);

	return ret;
}

int PutNext(c, fd)
int c;
FILE *fd;
{
	int i, ret;

	i = fileno(fd);

	PNflag[i] = 1;

	if (i < MAXFILES && TamBuf[i]) {
#ifdef MSDOS
		if (c == '\n') {
			if ((unsigned) indb[i] >= TamBuf[i]) {
				if (ftell(fd) != OffSet[i])
					lseek(i, OffSet[i], SEEK_SET);

				tamb[i] = write(i, buffer[i], TamBuf[i]);

				if (tamb[i] < (int) TamBuf[i])
					return EOF;

				OffSet[i] += ((long) tamb[i]);

				indb[i] = 0;
			}

			buffer[i][indb[i]] = 0x0d;

			++indb[i];
		}
#endif

		if ((unsigned) indb[i] >= TamBuf[i]) {
			if (ftell(fd) != OffSet[i])
				lseek(i, OffSet[i], SEEK_SET);

			tamb[i] = write(i, buffer[i], TamBuf[i]);

			if (tamb[i] < (int) TamBuf[i])
				return EOF;

			OffSet[i] += ((long) tamb[i]);

			indb[i] = 0;
		}

		buffer[i][indb[i]] = (char) c;
		ret = c;

		++indb[i];
	}
	else
		ret = fputc(c, fd);
	
	return ret;
}

int Reset_Buff(fd)
FILE *fd;
{
	int i;

	if ((i = fileno(fd)) < MAXFILES) {
		if (TamBuf[i]) {
			if (PNflag[i] && indb[i]) {
				if (ftell(fd) != OffSet[i])
					lseek(i, OffSet[i], SEEK_SET);

				tamb[i] = write(i, buffer[i], (RWTT) indb[i]);

				if (tamb[i] < indb[i])
					return EOF;

				indb[i] = 0;
			}

			TamBuf[i] = 0;
			tamb[i] = 0;
			free((ALTIPO) buffer[i]);
		}
	}

	return 0;
}

void Set_Buff(fd, t)
FILE *fd;
unsigned t;
{
	int i;

	i = fileno(fd);

	if (i >= MAXFILES)
		return;

	while (1) {
		if ((buffer[i] = (char *) malloc((PATIPO) t)) != (char *) NULL)
			break;

		t /= 2;
	}

	TamBuf[i] = t;
	tamb[i] = 0;
	indb[i] = 0;
	PNflag[i] = 0;
	OffSet[i] = 0L;

}

int lercampo(fd, s)
FILE	*fd;
char	*s;
{
	int	c;

	while (1) {
		c = GetNext(fd);

		if (c == ';' || c == EOF || c == '\n')
			break;

		if (c == '\\') {
			c = GetNext(fd);

			switch (c) {
			case '\\':
			case ';':
				break;

			case 'n':
				c = '\n';

				break;

			case 't':
				c = '\t';

				break;

			default:

/* *** Padroniz. msg erro - Adriana 30/11/94 
				fprintf(stderr,
					"Erro: Sequencia de escape invalida");
				fprintf(stderr, " na base de dados.\n");
*** */

				fprintf(stderr, PegaMsg (1016) );

				break;
			}
		}

		*s++ = (char) c;
		if ( __tamobs != 0 ) {
		--__tamobs;
			if ( __tamobs < 0 ) {
				fprintf(stderr, "ADVERTENCIA!!!\n");
				fprintf(stderr, "\t\tAtingido o tamanho maximo da observacao.\n");
				fprintf(stderr, "\t\tSeu conteudo sera' truncado.\n\n");
				break;
			}
		}
	}

	*s = '\0';

	return c;

}

int gravacampo(fd, s, sep)
FILE *fd;
char *s;
char sep;
{
	int c;

	while ((c = *s++)) {
		if (c == '\\' || c == ';' || c == '\n' || c == '\t') {
			if (PutNext('\\', fd) == EOF)
				return EOF;

			if (c == '\n')
				c = 'n';

			else if (c == '\t')
				c = 't';
		}

		if (PutNext(c, fd) == EOF)
			return EOF;

		if ( __tamobs != 0 ) {
		--__tamobs;
			if ( __tamobs < 0 ) {
				fprintf(stderr, "ADVERTENCIA!!!\n");
				fprintf(stderr, "\t\tAtingido o tamanho maximo da observacao.\n");
				fprintf(stderr, "\t\tSeu conteudo sera' truncado.\n\n");
				break;
			}
		}
	}

	if (PutNext(sep, fd) == EOF)
		return EOF;

	return 0;
}

#ifndef SEEK_SET
# define SEEK_SET	0	/* inicio do arquivo */
#endif

rew_arq(parq)
ARQ	*parq;
{
	int i;
	int r;

	if ((i = fileno(parq -> fddat)) < MAXFILES) {
		if (TamBuf[i]) {
			if (PNflag[i] && indb[i]) {
				if (ftell(parq -> fddat) != OffSet[i])
					lseek(i, OffSet[i], SEEK_SET);

				tamb[i] = write(i, buffer[i], (RWTT) indb[i]);

				if (tamb[i] < indb[i])
					return EOF;

				indb[i] = 0;
			}

			indb[i] = 0;
			tamb[i] = 0;
			PNflag[i] = 0;
			OffSet[i] = 0L;
		}
	}

	parq -> proxobs = 0;

	r = fseek(parq -> fddat, 0L, SEEK_SET);

	return r;
}



int copia_arquivo  (char *de, char *para)
{
  /* copia arquivo "de" no arquivo "para".
     Retorna 1, se sucesso; 0, caso contrario.
   */

  	FILE 	*f1, *f2;
	int	c;

  	f1 = fopen(de,"rb");
  	if (f1 == NULL) 
		return 0;
	
  	f2 = fopen(para,"wb");
  	if (f2 == NULL) 
		return 0;
	
  	while ((c = fgetc(f1)) != EOF) {
	    if (fputc(c,f2) == EOF) {
	    /* erro !! */
	    
	    return 0;
	    }
      }

      fclose(f1);
      fclose(f2);
      return 1;
}
