/***********************************************************************************
 LIBSOC - Biblioteca padro para o SOftware Cientfico

 Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

 Esta biblioteca  software livre; voc pode redistribu-la e/ou modific-la sob 
 os termos da Licena Pblica Geral Menor do GNU conforme publicada pela Free 
 Software Foundation; tanto a verso 2.1 da Licena, ou (a seu critrio) 
 qualquer verso posterior.

 Esta biblioteca  distribuda na expectativa de que seja til, porm, SEM NENHUMA 
 GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA 
 FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral Menor do GNU para mais 
 detalhes.

 Voc deve ter recebido uma cpia da Licena Pblica Geral Menor do GNU junto com 
 esta biblioteca; se no, escreva para a Free Software Foundation, Inc., no 
 endereo 59 Temple Street, Suite 330, Boston, MA 02111-1307 USA. 

 ***********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

    Manipula o arquivo .CTL associado a base.  A cada alteracao num registro da base 
    que esta' sendo manipulada, deve atualizar o arquivo .CTL associado a base.
    
 Desenvolvido por:

        	Mrio Nascimento
		Oscar Ivan Palma Pacheco
		
 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/


/*
** ----------------------------------------------------------------------------
**
** A cada alteracao num registro da base que esta' sendo manipulada, 
** deve atualizar o arquivo .CTL associada a base. O lay-out do arquivo .CTL 
** e' o seguinte:
**         Num_Reg ; Num_Reg_Ind ; Tam_Base_Bytes \n
**         Cod ; Num_Reg ; Reg \n ...
**         Cod ; Num_Reg ; Reg \n
**         <EOF>
**
** Onde: Num_Reg e' o numero de registros da base; Num_Reg_Ind e' o numero de
**       registros indexados da base; Tam_Base_Bytes e' o tamanho da base em
**       bytes; Cod = D se for desindexar o registro alterado ou R se for 
**       indexar ; Num_Reg e' o numero do registro alterado;  Reg e' uma copia
**       do registro alterado.
**
** Autor : MARIO Nascimento
**
** Obs.: Ha' uma variavel global (static int) chamada EstaCriandoCTL 
**
** ----------------------------------------------------------------------------
*/

#include <stdio.h>

#include        "pcsgbd.h"
#include        "filectl.h"

void    CriaHeaderCTL(nome_base)     
char    nome_base[];		/* nome da base em uso */
{
	/*
	** apenas inicializa o header que o RECINFO vai atualizar.
	*/

	char    filectl[TAMSTR];	/* nome do arquivo .CTL */
	FILE    *fpctl;			/* (FILE *) p/ o arquivo.CTL */

	sprintf(filectl,"%s.ctl",nome_base);

	if (access(filectl, 00) != 0) {

		/*
		** inicia o arquivo .CTL
		*/

		if ((fpctl = fopen(filectl,"w")) == (FILE *) NULL)
			ErroCTL("Nao consegui abrir arquivo '.CTL'");

		EstaCriandoCTL = 1;	/* Variavel Global */

		fprintf(fpctl,"0;0;0\n");

		fclose(fpctl);

	} else {
		/*
		** pode ser que exista mas tem '0;0;0', i.e., nao
		** foi indexado ainda, entao nao coloca como 'R'
		*/

		if (ReadTamanhoBaseCTL(nome_base) == 0L && 
		    ReadNumRegsBaseCTL(nome_base) == 0L && 
		    ReadNumRegIndxBaseCTL(nome_base) == 0L) 

			EstaCriandoCTL = 1;	/* Variavel Global */
		else
			EstaCriandoCTL = 0;	/* Variavel Global */

		return; /* se o arquivo ja' existe so' faz o 'append' */

	}
}


void    AppendCTL_XT(numr,cod,reg,nome_base) 
unsigned numr;				/* numero do registro na base */
char    cod;				/* INSERE_CTL (R) ou DELETA_CTL (D) */
char    *reg[];				/* campos do registro */
char    nome_base[];			/* nome da base em uso */
{
	/*
	** adiciona o registro da base a ser index/desind SE E SO' SE
	** a alteracao for proveniente de um ^T ou ^X
	*/

	int     c;			/* contador geral */
	char    filectl[TAMSTR];	/* nome do arquivo .ctl */
	FILE    *fpctl;			/* (FILE *) p/ o arquivo .CTL */
	char	*ptr;

	if (EstaCriandoCTL)
		return;

	/*
	** abre o arquivo .CTL p/ fazer uma "append"
	*/

	sprintf(filectl,"%s.ctl",nome_base);

	if ((fpctl = fopen(filectl,"a")) == (FILE *) NULL)
		ErroCTL("Nao consegui abrir arquivo '.CTL'");

	/*
	** se for 0 entao esta' no 1o. reg
	*/

	if (!numr)
		numr = 1;
	
	fprintf(fpctl,"%d;%c;\"",numr-1,cod);	/* num reg + cod de operacao */

	for (c = 0; reg[c]; ) {
		ptr = reg[c];

		while(*ptr!='\0') {
			if(*ptr==';')
				putc('\\',fpctl);

			fputc(*ptr,fpctl);
			ptr++;
		}

		/*
		** fprintf(fpctl,"%s",reg[c]); registro (cpo a cpo)
		*/

		if (reg[++c])			/* ja' incrementa  'c' e se */
			fprintf(fpctl,";");	/* ha' proximo coloca o ';' */
	}

	fprintf(fpctl,"\"\n");

	fclose(fpctl);

}


void    AppendCTL_DU(nome_base,numr,offset,cod)
char    nome_base[];		/* nome da base em uso */
unsigned numr;			/* numero do registro na base adicionado de 1 */
long    *offset;		/* byte onde comeca cada registro */
char    cod;			/* 'R' ou 'D'   */
{
	/*
	** adiciona o registro da base a ser index/desind SE E SO' SE 
	** a alteracao for proveniente de um ^D
	*/

	int     c;			/* contador geral */
	char    filectl[TAMSTR];	/* nome do arquivo .CTL */
	char    filedat[TAMSTR];	/* nome do arquivo .DAT */
	FILE    *fpctl;			/* (FILE *) p/ o arquivo .CTL */
	FILE    *fpdat;			/* (FILE *) p/ o arquivo .DAT */

	if (EstaCriandoCTL)
		return;

	/*
	** abre o arquivo .CTL p/ fazer uma "append"
	*/

	sprintf(filectl,"%s.ctl",nome_base);

	if ((fpctl = fopen(filectl,"a")) == (FILE *) NULL)
		ErroCTL("Nao consegui abrir arquivo '.ctl'");

	fprintf(fpctl,"%d;%c;\"",numr-1,cod);	/* num reg + 'D' ou 'R' */

	/*
	** agora tenho que ler diretamente da base p/ poder gravar no .CTL
	*/

	sprintf(filedat,"%s.dat",nome_base);

	if ((fpdat = fopen(filedat,"r")) == (FILE *) NULL)
		ErroCTL("Nao consegui abrir arquivo '.dat'");

	/*
	** posiciona no inicio do registro
	*/

	fseek(fpdat, offset[numr - 1], 0);

	/*
	** le e grava cada campo desse registro enquanto nao for
	** fim do registro atual (== \n).
	*/

	while ((c = getc(fpdat)) != '\n')
		putc(c,fpctl);
 
	fclose(fpdat);
 
	fprintf(fpctl,"\"\n");
	fclose(fpctl);

}


unsigned ReadTamanhoBaseCTL(nome_base)
char     nome_base[];
{
	/*
	** retorna o numero de registros que esta no .CTL
	*/

	char    header[TAMSTR];		/* 1a. linha do .CTL */
	char    num[TAMSTR];		/* numero lido do header */
	FILE    *fpctl;
	int     c, c2;			/* contador de uso geral */

	fpctl = OpenFileCTLRead(nome_base);

	if (fgets(header, TAMSTR, fpctl) == NULL)
		ErroCTL("Problemas na leitura do .CTL");

	for(c = 0; (num[c] = header[c]) != ';'; c++)
		;

	num[c] = '\0';			/* tira o ';' */

	/*
	** pula o 1o. ';', inicia o 'c2' e anda uma posicao em header
	*/

	c2 = 0;

	for(c++; (num[c2] = header[c]) != ';'; c++, c2++)
		;

	/*
	** pula o 2o. ';', re-inicia o 'c2' e anda uma posica em header
	*/

	c2 = 0;

	for(c++; (num[c2] = header[c]) != '\n'; c++, c2++)
		;

	num[c2] = '\0';			/* tira o '\n' */

	CloseFileCTL(fpctl);

	return((unsigned) atoi(num));
}


unsigned ReadNumRegIndxBaseCTL(nome_base)
char     nome_base[];
{
	/*
	** retorna o numero de registros que esta no .CTL
	*/

	char    header[TAMSTR];		/* 1a. linha do .CTL */
	char    num[TAMSTR];		/* numero lido do header */
	FILE    *fpctl;
	int     c, c2;			/* contadores de uso geral */

	fpctl = OpenFileCTLRead(nome_base);

	if (fgets(header, TAMSTR, fpctl) == NULL)
		ErroCTL("Problemas na leitura do .CTL");

	for(c = 0; (num[c] = header[c]) != ';'; c++)
		;

	/*
	** pula o 1o. ';', inicia o 'c2' e anda uma posicao em header
	*/

	c2 = 0;

	for(c++; (num[c2] = header[c]) != ';'; c++, c2++)
		;

	num[c2] = '\0';			/* tira o ';' */

	CloseFileCTL(fpctl);

	return((unsigned) atoi(num));
}


unsigned ReadNumRegsBaseCTL(nome_base)
char     nome_base[];
{
	/*
	** retorna o numero de registros que esta no .CTL
	*/

	char    header[TAMSTR];		/* 1a. linha do .CTL */
	char    num[TAMSTR];		/* numero lido do header */
	FILE    *fpctl;
	int     c;			/* contador de uso geral */

	fpctl = OpenFileCTLRead(nome_base);

	if (fgets(header, TAMSTR, fpctl) == NULL)
		ErroCTL("Problemas na leitura do .CTL");

	for(c = 0; (num[c] = header[c]) != ';'; c++)
		;

	num[c] = '\0';			/* tira o ';' */

	CloseFileCTL(fpctl);

	return((unsigned) atoi(num));
}


void    WriteTamanhoBaseCTL(nome_base,num)
char    nome_base[];
unsigned num;
{
	/*
	** escreve o novo tamanho da base no .CTL
	*/

	FILE    *fptmp;
	char    filetmp[TAMSTR];
	char    filectl[TAMSTR];

	strcpy(filetmp, GetFileNameTMP());

	fptmp = OpenFileTMPWrite(filetmp);

	fprintf(fptmp,"%ld;%ld;%ld\n",ReadNumRegsBaseCTL(nome_base),
				      ReadNumRegIndxBaseCTL(nome_base),
				      num);

	CopyFileCTL(fptmp,nome_base);

	CloseFileTMP(fptmp);

	sprintf(filectl,"%s.ctl",nome_base);

	unlink(filectl);

	rename(filetmp,filectl);
}


void    WriteNumRegsBaseCTL(nome_base,num)
char    nome_base[];
unsigned num;
{
	/*
	** escreve o novo numero de registros da base no .CTL
	*/

	FILE    *fptmp;
	char    filetmp[TAMSTR];
	char    filectl[TAMSTR];

	strcpy(filetmp, GetFileNameTMP());

	fptmp = OpenFileTMPWrite(filetmp);

	fprintf(fptmp,"%ld;%ld;%ld\n",num,
				      ReadNumRegIndxBaseCTL(nome_base),
				      ReadTamanhoBaseCTL(nome_base));

	CopyFileCTL(fptmp,nome_base);

	CloseFileTMP(fptmp);

	sprintf(filectl,"%s.ctl",nome_base);

	unlink(filectl);

	rename(filetmp,filectl);
}


void    WriteNumRegIndxBaseCTL(nome_base,num)
char    nome_base[];
unsigned num;
{
	/*
	** escreve o novo tamanho da base no .CTL
	*/

	FILE    *fptmp;
	char    filetmp[TAMSTR];
	char    filectl[TAMSTR];

	strcpy(filetmp, GetFileNameTMP());

	fptmp = OpenFileTMPWrite(filetmp);

	fprintf(fptmp,"%ld;%ld;%ld\n",ReadNumRegsBaseCTL(nome_base),
				      num,
				      ReadTamanhoBaseCTL(nome_base));

	CopyFileCTL(fptmp,nome_base);

	CloseFileTMP(fptmp);

	sprintf(filectl,"%s.ctl",nome_base);

	unlink(filectl);

	rename(filetmp,filectl);
}


FILE    *OpenFileCTLRead(nome_base)
char     nome_base[];
{
	/*
	** abre o arquivo .CTL p/ leitura e retorna um (FILE *) p/ o mesmo
	*/

	char    filectl[TAMSTR];
	FILE    *fpctl;

	/*
	** Obtendo o nome da base juntamente com o caminho completo.
	*/

	strcpy(filectl, nome_base);
	strcat(filectl, ".ctl");

	if ((fpctl = fopen(filectl,"r")) == (FILE *) NULL)
		ErroCTL("Nao consegui abrir arquivo '.CTL'");

	else
		return(fpctl);
}


void    CloseFileCTL(fpctl)
FILE    *fpctl;
{
	if (fclose(fpctl) == EOF)
		ErroCTL("Nao consegui fechar arquivo '.CTL'");
}


char    *GetFileNameTMP()
{
#ifndef MSDOS
	char    nome[TAMSTR];
#endif

#ifdef  MSDOS
	return("_!CTL!_.TMP");
#else
	sprintf(nome,"_!CTL!_%d.TMP",getpid());
	return(nome);
#endif
}


FILE    *OpenFileTMPWrite(filetmp)
char    filetmp[];
{
	/*
	** abre um arquivo temporario p/ escrita
	*/

	FILE    *fptmp;

	if ((fptmp = fopen(filetmp,"w")) == (FILE *) NULL)
		ErroCTL("Nao consegui abrir arquivo '.CTL' temporario");

	else
		return(fptmp);
	
}


void    CloseFileTMP(fptmp)
FILE    *fptmp;
{
	if (fclose(fptmp) == EOF)
		ErroCTL("Nao consegui fechar arquivo '.CTL' temporario");
}


void    CopyFileCTL(fptmp,nome_base)
FILE    *fptmp;
char    nome_base[];
{
	char    buffer[KBYTE];		/* p/ fread e fwrite */
	FILE    *fpctl;
	unsigned nbytes;		/* num. de bytes lidos */

	fpctl = OpenFileCTLRead(nome_base);

	/*
	** leio o header do .CTL e o desprezo
	*/

	if (fgets(buffer, TAMSTR, fpctl) == NULL)
		ErroCTL("Problemas na leitura do .CTL");

	/*
	** faco um looping p/ leitura do .CTL e escrita do .TMP
	*/

	while(1) {

		nbytes = fread(buffer, sizeof(char), KBYTE, fpctl);

		if (nbytes)
			fwrite(buffer, sizeof(char), nbytes, fptmp);

		else
			break;

	} 

	CloseFileCTL(fpctl);
}


void    ErroCTL(msg)
char    msg[];
{
	fprintf(stderr,"\n%s\n");
	exit(5);
}

