/***********************************************************************************
LIBSOC - Biblioteca padro para o SOftware Cientfico

Direitos Autorais Reservados (c) 2004 Embrapa Informtica Agropecuria

Esta biblioteca  software livre; voc pode redistribu-la e/ou modific-la sob
os termos da Licena Pblica Geral Menor do GNU conforme publicada pela Free
Software Foundation; tanto a verso 2.1 da Licena, ou (a seu critrio)
qualquer verso posterior.

Esta biblioteca  distribuda na expectativa de que seja til, porm, SEM NENHUMA
GARANTIA; nem mesmo a garantia implcita de COMERCIABILIDADE OU ADEQUAO A UMA
FINALIDADE ESPECFICA. Consulte a Licena Pblica Geral Menor do GNU para mais
detalhes.

Voc deve ter recebido uma cpia da Licena Pblica Geral Menor do GNU junto com
esta biblioteca; se no, escreva para a Free Software Foundation, Inc., no
endereo 59 Temple Street, Suite 330, Boston, MA 02111-1307 USA.

 ***********************************************************************************

 Parte integrante do SOC - SOftware Cientfico

 Descrio:

    ninv -- calcula ainversa da Normal, com baixa acuracidade. 

Desenvolvido por:


		Dalton Francisco Andrade
		Maria Fernanda Moura
		Oscar Ivan Palma Pacheco
		Slvio Roberto Medeiros Evangelista
		Snia Ternes


 Informaes do CVS:
        $Source:
        $Revision:
        $Date:

***********************************************************************************/

#include <stdio.h>
#include <math.h>
#include "pcsgbd.h"


/*--------------------------------------------------------------*/
/* ninv -- Find Percentage Point of Normal Distribution         */
/*--------------------------------------------------------------*/
/*                                                              */
/*Function:  ninv                                               */
/*                                                              */
/*Purpose:   Finds percentage point of normal distribution      */
/*           (low accuracy)                                     */
/*                                                              */
/*Calling Sequence:                                             */
/*                                                              */
/*     Perc  := ninv( P : REAL ) : REAL;                        */
/*                                                              */
/*          P      --- input probability                        */
/*          Perc   --- resultant percentage point               */
/*                                                              */
/*Calls:  None                                                  */
/*                                                              */
/*Remark:  This method provides about 6.5 decimal digits of     */
/*         accuracy.                                            */
/*                                                              */
/*--------------------------------------------------------------*/

#ifdef MSDOS
#define PRECISAO 1.e-15
#else
#define PRECISAO 1.e-08
#endif

/* Structured */ 

static double PN[]  = {
  		       0.0,
		      -0.322232431088  ,
		      -1.0             ,
		      -0.342242088547  ,
		      -0.0204231210245 ,
		      -0.453642210148E-4 
		      };

static double QN[] = {
		      0.0,
		      0.0993484626060 ,
		      0.588581570495  ,
		      0.531103462366  ,
		      0.103537752850  ,
		      0.38560700634E-2  
		      };

double ninv( P ) 
double P;
{

	double Lim = 1.0e-20;

	double Y,Pr,Nv;

	/* Inicio ninv */

	if ( P  <=  PRECISAO)		return(8.21);
	if ( P  >=  (1.0 - PRECISAO) )	return(-8.21);

	P = 1.0 - P;

	if( P > 0.5 ) 
		Pr = 1.0 - P;
	else
		Pr = P;

	if (Pr < Lim)
		Nv = 8.21;

	else if (Pr == 0.5)
		Nv = 0.0;

	else {
		Y      = sqrt ( log( 1.0 / Pr / Pr ) );

         	Nv  = Y + ((((Y * PN[ 5 ] + PN[ 4 ]) * Y + PN[ 3 ] ) * Y
                                       + PN[ 2 ]) * Y + PN[ 1 ] ) /
                       ((((Y * QN[ 5 ] + QN[ 4 ]) * Y + QN[ 3 ] ) * Y
                                       + QN[ 2 ]) * Y + QN[ 1 ] );
		if( P < 0.5 ) 	Nv = -1.0*Nv;
		else 		Nv = Nv;
	}

	return(Nv);
}

